package xic;

import java.util.HashSet;
import java.util.Set;

import xic.codegen.Instr.Register;


public class Configuration {
	public enum Architecture {
		I386,
		X86_64
	};
	
	public enum Target {
		LINUX,
		DARWIN,
		WINDOWS
	};

	public enum AssemblySyntax {
		ATT_SYNTAX,
		INTEL_SYNTAX
	}

	public static Architecture ARCH = Architecture.X86_64;
	public static Target TARGET = Target.LINUX;
	public static Set<String> REG_NAMES;
	public static int WORD_SIZE;
	public static int LOG_WORD_SIZE;
	public static Register[] AVAILABLE_REGISTERS;
	public static String[] PARAMETER_REGISTERS;
	public static String CALLEE_SAVE_REGS[];
	public static String[] CALLER_SAVE_REGS;
	public static String RETVAL_REGISTER;
	public static String IP_REGISTER;
	public static int STACK_ALIGNMENT;
	public static int SHADOW_ZONE_SIZE;
	public static boolean IP_RELATIVE_ADDRESSING;
	public static AssemblySyntax SYNTAX = AssemblySyntax.ATT_SYNTAX;

	static {
		setTarget(detectTarget());
		setArchitecture(Architecture.X86_64);
	};
	
	public static void setArchitecture(Architecture arch) {
		ARCH = arch;
		configure();
	}

	public static void setTarget(Target trg) {
		TARGET = trg;
		configure();
	}

	private static void configure() {
		if (ARCH == Architecture.I386) {
			WORD_SIZE = 4;
			LOG_WORD_SIZE = 2;
			SHADOW_ZONE_SIZE = 0;
			
			if (TARGET == Target.DARWIN)
				STACK_ALIGNMENT = 16;
			else
				STACK_ALIGNMENT = 4;

			REG_NAMES = new HashSet<String>();
			REG_NAMES.add("al"); REG_NAMES.add("ax"); REG_NAMES.add("eax");
			REG_NAMES.add("bl"); REG_NAMES.add("bx"); REG_NAMES.add("ebx");
			REG_NAMES.add("cl"); REG_NAMES.add("cx"); REG_NAMES.add("ecx");
			REG_NAMES.add("dl"); REG_NAMES.add("dx"); REG_NAMES.add("edx");
			REG_NAMES.add("bp"); REG_NAMES.add("ebp");
			REG_NAMES.add("sp"); REG_NAMES.add("esp");
			REG_NAMES.add("si"); REG_NAMES.add("esi");
			REG_NAMES.add("di"); REG_NAMES.add("edi");
	
			/* Usable registers */
			AVAILABLE_REGISTERS = new Register[] { 
				/* The order is intentional - pick callee-save registers last */
				new Register(Register.Type.A),
				new Register(Register.Type.C),
				new Register(Register.Type.D),
				new Register(Register.Type.B),
				new Register(Register.Type.DI),
				new Register(Register.Type.SI),
				new Register(Register.Type.BP)
			};
	
			CALLEE_SAVE_REGS = new String[] { "ebx", "esi", "edi", "ebp" };
			CALLER_SAVE_REGS = new String[] { "eax", "ecx", "edx" };
			RETVAL_REGISTER = "eax";
			IP_REGISTER = "eip";
			IP_RELATIVE_ADDRESSING = false;

			/* cdecl calling conventions -- pass everything using the stack */
			PARAMETER_REGISTERS = new String[] {};
		} else if (ARCH == Architecture.X86_64) {
			WORD_SIZE = 8;
			LOG_WORD_SIZE = 3;
			STACK_ALIGNMENT = 16;

			REG_NAMES = new HashSet<String>();
			REG_NAMES.add("al");   REG_NAMES.add("ax");   REG_NAMES.add("eax");  REG_NAMES.add("rax");
			REG_NAMES.add("bl");   REG_NAMES.add("bx");   REG_NAMES.add("ebx");  REG_NAMES.add("rbx");
			REG_NAMES.add("cl");   REG_NAMES.add("cx");   REG_NAMES.add("ecx");  REG_NAMES.add("rcx");
			REG_NAMES.add("dl");   REG_NAMES.add("dx");   REG_NAMES.add("edx");  REG_NAMES.add("rdx");
			REG_NAMES.add("bpl");  REG_NAMES.add("bp");   REG_NAMES.add("ebp");  REG_NAMES.add("rbp");
			REG_NAMES.add("spl");  REG_NAMES.add("sp");   REG_NAMES.add("esp");  REG_NAMES.add("rsp");
			REG_NAMES.add("sil");  REG_NAMES.add("si");   REG_NAMES.add("esi");  REG_NAMES.add("rsi");
			REG_NAMES.add("dil");  REG_NAMES.add("di");   REG_NAMES.add("edi");  REG_NAMES.add("rdi");
			REG_NAMES.add("r8b");  REG_NAMES.add("r8w");  REG_NAMES.add("r8d");  REG_NAMES.add("r8");
			REG_NAMES.add("r9b");  REG_NAMES.add("r9w");  REG_NAMES.add("r9d");  REG_NAMES.add("r9");
			REG_NAMES.add("r10b"); REG_NAMES.add("r10w"); REG_NAMES.add("r10d"); REG_NAMES.add("r10");
			REG_NAMES.add("r11b"); REG_NAMES.add("r11w"); REG_NAMES.add("r11d"); REG_NAMES.add("r11");
			REG_NAMES.add("r12b"); REG_NAMES.add("r12w"); REG_NAMES.add("r12d"); REG_NAMES.add("r12");
			REG_NAMES.add("r13b"); REG_NAMES.add("r13w"); REG_NAMES.add("r13d"); REG_NAMES.add("r13");
			REG_NAMES.add("r14b"); REG_NAMES.add("r14w"); REG_NAMES.add("r14d"); REG_NAMES.add("r14");
			REG_NAMES.add("r15b"); REG_NAMES.add("r15w"); REG_NAMES.add("r15d"); REG_NAMES.add("r15");
			REG_NAMES.add("rip");

			RETVAL_REGISTER = "rax";
			IP_REGISTER = "rip";
			IP_RELATIVE_ADDRESSING = true;

			if (TARGET == Target.LINUX || TARGET == Target.DARWIN) {
				/* Usable registers */
				AVAILABLE_REGISTERS = new Register[] { 
					/* The order is intentional - pick callee-save registers last */
					new Register(Register.Type.A),
					new Register(Register.Type.C),
					new Register(Register.Type.D),
					new Register(Register.Type.DI),
					new Register(Register.Type.SI),
					new Register(Register.Type.R8),
					new Register(Register.Type.R9),
					new Register(Register.Type.R10),
					new Register(Register.Type.R11),
					
					/* Discourage the use of callee-save registers */
					new Register(Register.Type.B),
					new Register(Register.Type.R14),
					new Register(Register.Type.R15),
					
					/* Discourage the use of RBP, R12 and R13, since they cause
					 * longer instruction encodings */ 
					new Register(Register.Type.R12),
					new Register(Register.Type.R13),
					new Register(Register.Type.BP)
				};
	
				/* System V calling conventions */
				PARAMETER_REGISTERS = new String[] {
					"rdi", "rsi", "rdx", "rcx", "r8", "r9"
				};
	
				CALLEE_SAVE_REGS = new String[] { "rbx", "rbp", "r12", "r13", "r14", "r15" };
				CALLER_SAVE_REGS = new String[] { "rax", "rcx", "rdx", "rsi", "rdi", "r8", "r9", "r10", "r11" };
				SHADOW_ZONE_SIZE = 0;
			} else if (TARGET == Target.WINDOWS) {
				/* Microsoft calling conventions */

				/* Usable registers */
				AVAILABLE_REGISTERS = new Register[] { 
					/* The order is intentional - pick callee-save registers last */
					new Register(Register.Type.A),
					new Register(Register.Type.C),
					new Register(Register.Type.D),
					new Register(Register.Type.R8),
					new Register(Register.Type.R9),
					new Register(Register.Type.R10),
					new Register(Register.Type.R11),
					
					/* Discourage the use of callee-save registers */
					new Register(Register.Type.B),
					new Register(Register.Type.R14),
					new Register(Register.Type.R15),
					new Register(Register.Type.DI),
					new Register(Register.Type.SI),
					
					/* Discourage the use of RBP, R12 and R13, since they cause
					 * longer instruction encodings */ 
					new Register(Register.Type.R12),
					new Register(Register.Type.R13),
					new Register(Register.Type.BP)
				};
				
				PARAMETER_REGISTERS = new String[] {
					"rcx", "rdx", "r8", "r9"
				};

				CALLEE_SAVE_REGS = new String[] { "r12", "r13", "r14", "r15", "rdi", "rsi", "rbx", "rbp" };
				CALLER_SAVE_REGS = new String[] { "rax", "rcx", "rdx", "r8", "r9", "r10", "r11" };
				SHADOW_ZONE_SIZE = 4 * WORD_SIZE;
			}
		} else {
			throw new RuntimeException("Unknown architecture!");
		}
	}
	
	public static void setSyntax(AssemblySyntax syntax) {
		SYNTAX = syntax;
	}
	
	public static Target detectTarget() {
		String osName = System.getProperty("os.name").toLowerCase();
		if (osName.contains("windows"))
			return Target.WINDOWS;
		else if (osName.contains("mac os"))
			return Target.DARWIN;
		else if (osName.contains("linux"))
			return Target.LINUX;
		else
			throw new InternalCompilerException("Unknown/unsupported OS!");
	}

	public static int registerIndex(String name) {
		for (int i=0; i<AVAILABLE_REGISTERS.length; ++i)
			if (name.equals(AVAILABLE_REGISTERS[i].toString()))
				return i;
		throw new InternalCompilerException("Register " + name + " not found!");
	}
}
