package xic.analysis;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

/**
 * Abstract dataflow analysis framework
 */
public abstract class DataFlowAnalysis {
	/* Strategy for visiting nodes */
	public static final int WORKLIST_NAIVE = 1;
	public static final int WORKLIST_REV_POSTORDER = 2;

	protected enum Direction {
		FORWARD, BACKWARD
	};

	public static abstract class FlowInfo {
		public Value entry;
		public Value exit;
	};
	
	protected Direction dir;
	protected Map<CFGNode, FlowInfo> flowInfo;
	protected Value top;
	protected Set<CFGNode> nodes;
	protected int algorithm;
	protected boolean debug;

	public DataFlowAnalysis(Set<CFGNode> nodes) {
		this.nodes = nodes;
		this.debug = false;
		this.algorithm = WORKLIST_REV_POSTORDER;
	}
	
	public void setDebug(boolean debug) {
		this.debug = debug;
	}

	public void setAlgorithm(int algorithm) {
		this.algorithm = algorithm;
	}
	
	public Direction getDirection() {
		return dir;
	}

	protected void initFlowInfo() {
		if (flowInfo == null) {
			flowInfo = new HashMap<CFGNode, FlowInfo>();
			for (CFGNode node : nodes) 
				flowInfo.put(node, createFlowInfo(node));
		}
	}

	/**
	 * Abstract transfer function. To be defined by each
	 * particular data flow analysis.
	 */
	protected abstract void F(FlowInfo f);

	/**
	 * Used to initially create the FlowInfo instance associated with a CFG node
	 */
	protected abstract FlowInfo createFlowInfo(CFGNode node);

	/**
	 * Return the <code>FlowInfo</code> instance associated with a CFG node
	 */
	public FlowInfo getFlowInfo(CFGNode node) {
		return flowInfo.get(node);
	}

	/**
	 * Run the work-list algorithm (naive queue-based ordering)
	 */
	private void solveWorklistNaive() {
		int iterations = 0;
		Queue<CFGNode> queue = new ArrayDeque<CFGNode>(nodes);
		initFlowInfo();

		for (CFGNode n : nodes)
			flowInfo.get(n).exit = top.copy();

		while (!queue.isEmpty()) {
			++iterations;

			CFGNode node = queue.remove();
			FlowInfo fi = flowInfo.get(node);

			/* Compute entry value */
			Value entry = null;

			if(dir == Direction.FORWARD) {
				if(node.inEdges().size() > 0) { 
					for(CFGNode np : node.inEdges()) {
						if (entry == null)
							entry = flowInfo.get(np).exit.copy();
						else
							entry.meet(flowInfo.get(np).exit);
					}
				} else {
					entry = top.copy();
					entry.setNullMeet();
				}
			} else {
				if(node.outEdges().size() > 0) {
					for(CFGNode np : node.outEdges()) {
						if (entry == null)
							entry = flowInfo.get(np).exit.copy();
						else
							entry.meet(flowInfo.get(np).exit);
					}
				} else {
					entry = top.copy();
					entry.setNullMeet();
				}
			} 
			fi.entry = entry;
			Value oldExit = fi.exit.copy();

			/* Apply the transfer function */
			F(fi);
			
			if (debug) {
				System.out.println("=======================================");
				System.out.println("It. #" + iterations + ": processing node '" 
							+ node + "'");
				System.out.println("entry      = " + fi.entry);
				System.out.println("exit (old) = " + oldExit);
				System.out.println("exit (new) = " + fi.exit);
			}

			if (oldExit.equals(fi.exit)) {
				if (debug)
					System.out.print("\n");
				continue;
			} else if (debug) {
				System.out.println(" -> Scheduling " + 
						((dir == Direction.FORWARD) ? node.outEdges() : node.inEdges()) + " for update\n");
			}


			/* Push the successors/predecessors onto the queue */
			if(dir == Direction.FORWARD)
				queue.addAll(node.outEdges());
			else
				queue.addAll(node.inEdges());

		}
		if (debug)
			System.out.println(getClass().getSimpleName() 
					+ ": solved dataflow equations after " + iterations + " iterations");
	}

	/**
	 * Run the work-list algorithm with reverse postorder traversal ordering
	 */
	private void solveWorklistPostOrder() {
		int iterations = 0;
		List<CFGNode> currentList = new ArrayList<CFGNode>();
		List<CFGNode> laterList = new ArrayList<CFGNode>(nodes);
		initFlowInfo();

		for (CFGNode n : nodes)
			flowInfo.get(n).exit = top.copy();

		while (true) {
			if (currentList.isEmpty()) {
				if (laterList.isEmpty()) {
					break;
				} else {
					List<CFGNode> tmp = currentList;
					currentList = laterList;
					laterList = tmp;
					dfsf(currentList);
				}
			}
			++iterations;

			CFGNode node = currentList.remove(currentList.size()-1);
			FlowInfo fi = flowInfo.get(node);

			/* Compute entry value */
			Value entry = null;

			if(dir == Direction.FORWARD) {
				if(node.inEdges().size() > 0) { 
					for(CFGNode np : node.inEdges()) {
						if (entry == null)
							entry = flowInfo.get(np).exit.copy();
						else
							entry.meet(flowInfo.get(np).exit);
					}
				} else {
					entry = top.copy();
					entry.setNullMeet();
				}
			} else {
				if(node.outEdges().size() > 0) {
					for(CFGNode np : node.outEdges()) {
						if (entry == null)
							entry = flowInfo.get(np).exit.copy();
						else
							entry.meet(flowInfo.get(np).exit);
					}
				} else {
					entry = top.copy();
					entry.setNullMeet();
				}
			} 
			fi.entry = entry;
			Value oldExit = fi.exit.copy();

			/* Apply the transfer function */
			F(fi);

			if (debug) {
				System.out.println("=======================================");
				System.out.println("It. #" + iterations + ": processing node '" 
							+ node + "'");
				System.out.println("entry      = " + fi.entry);
				System.out.println("exit (old) = " + oldExit);
				System.out.println("exit (new) = " + fi.exit);
			}

			if (oldExit.equals(fi.exit)) {
				if (debug)
					System.out.print("\n");
				continue;
			} else if (debug) {
				System.out.println(" -> Scheduling " + 
						((dir == Direction.FORWARD) ? node.outEdges() : node.inEdges()) + " for update\n");
			}

			/* Push the successors/predecessors onto the queue */
			if(dir == Direction.FORWARD)
				laterList.addAll(node.outEdges());
			else
				laterList.addAll(node.inEdges());

		}
		if (debug)
			System.out.println(getClass().getSimpleName() 
					+ ": solved dataflow equations after " + iterations + " iterations");
	}
	
	public void solve() {
		if (algorithm == WORKLIST_NAIVE)
			solveWorklistNaive();
		else
			solveWorklistPostOrder();
	}

	
	/**
	 * Compute the reverse postorder traversal sequence of a list of nodes
	 */
	private void dfsf(List<CFGNode> input) {
		Set<CFGNode> notSeen = new HashSet<CFGNode>(input);
		
		int i = input.size();
		while (notSeen.size() > 0) {
			CFGNode node = notSeen.iterator().next();
			i = dfsfTraverse(i, notSeen, node);
		}

		Collections.sort(input);
	}

	private int dfsfTraverse(int i, Set<CFGNode> notSeen, CFGNode node) {
		notSeen.remove(node);
		
		if (dir == Direction.FORWARD) {
			for (CFGNode n : node.outEdges()) {
				if (!notSeen.contains(n))
					continue;
				i = dfsfTraverse(i, notSeen, n);
			}
		} else {
			for (CFGNode n : node.inEdges()) {
				if (!notSeen.contains(n))
					continue;
				i = dfsfTraverse(i, notSeen, n);
			}
		}
		node.setTraversalIndex(i);
		return i-1;
	}
}
