package xic.analysis;

import java.io.IOException;
import java.util.Map;
import java.util.Set;

import xic.ErrorSink;
import xic.InternalCompilerException;
import xic.Statistics;
import xic.analysis.BoundingIntervals.Interval;
import xic.ast.ASTNode;
import xic.ast.ErrorReporter;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRStmt;
import xic.visitors.IRConstantFolder;
import xic.visitors.NodeVisitor;


public class IntervalAnalysis {
	private static class IntervalRewritingVisitor extends NodeVisitor {
		private Map<String, Interval> intervals, arrayLengths;
		private int numRewrites = 0;

		public IntervalRewritingVisitor() {
		}

		public void update(Map<String, Interval> intervals, Map<String, Interval> arrayLengths) {
			this.intervals = intervals;
			this.arrayLengths = arrayLengths;
			this.numRewrites = 0;
		}

		public int getNumRewrites() {
			return numRewrites;
		}

		protected ASTNode leave(ASTNode parent, ASTNode nOld, ASTNode nNew,
				NodeVisitor vNew) {
			if (nOld instanceof IRExpr && !(nOld instanceof IRConst)) {
				Interval interval = BoundingIntervals.eval((IRExpr) nOld, intervals, arrayLengths);
				if (interval.isSingleElement() && interval.l().isNumber()) {
					numRewrites++;
					return new IRConst((int) interval.l().value());
				}
			} else if (nOld instanceof IRMove
					&& !(((IRMove) nOld).target() instanceof IRMem)
					&& (((IRMove) nOld).target() != ((IRMove) nNew).target())) {
				/* Do not allow to replace a temp in MOVE(temp, ...) */
				numRewrites--;
				IRMove oldMove = (IRMove) nOld;
				IRMove newMove = (IRMove) nNew;
				return new IRMove((IRExpr) oldMove.target().deepCopy(), newMove.expr());
			}
			return nNew;
		}
	}

	public static void transform(Statistics stats, int pass, Map<String, CFGNode> cfgs) {
		ErrorSink errorSink = new ErrorSink();
		IRConstantFolder cfVisitor = new IRConstantFolder(errorSink);
		IntervalRewritingVisitor rewritingVisitor = new IntervalRewritingVisitor();

		for (String funcName : cfgs.keySet()) {
			CFGNode head = cfgs.get(funcName);
			Set<CFGNode> nodeSet = head.flatten();
			BoundingIntervals bi = new BoundingIntervals(nodeSet);
			bi.solve();
			int numReplacements = 0;
			for (CFGNode node:nodeSet) {
				if (node.value() instanceof IRStmt) {
					rewritingVisitor.update(bi.in(node).intervals(), bi.in(node).arrayLengths());

					IRNode backup = ((IRNode) node.value());
					if (!(backup instanceof IRCJump))
						continue;
					IRNode rewrittenNode = (IRNode) rewritingVisitor.visit((IRStmt) node.value());					
					IRNode newNode = (IRNode) cfVisitor.visit(rewrittenNode);
					numReplacements += rewritingVisitor.getNumRewrites();

					if (rewritingVisitor.getNumRewrites() > 0) {
						System.out.println("In function " + funcName + ": rewrote " + backup);
						System.out.println("==> " + newNode);
					}

						ConstPropagation.cleanupIfCJump(node, backup, newNode);

					node.setValue(newNode);
				}
			}

			if (numReplacements > 0) {
				CFGGenerator.removeOrphans(head);
				stats.log("intr", "Pass " + pass + ": '" + funcName
						+ "': simplified " + numReplacements + " expressions");
			}
		}

		if (errorSink.hasErrors()) {
			try {
				ErrorReporter.printErrors(errorSink);
				throw new InternalCompilerException("Constant folding produced errors!");
			} catch (IOException e) {
				throw new InternalCompilerException("Unexpected I/O error!", e);
			}
		}
	}
}
