package xic.analysis;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.Statistics;
import xic.ir.IRCJump;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;
import xic.ir.StructHashSet;
import xic.ir.StructHashSet.StructWrapper;
import xic.visitors.RewritingVisitor;


public class PartialRedundancy {
	private AnticipatedExpressions ante;
	private EarliestExpressions early;
	private Delayedness del;
	private Latestness latest;
	private Isolated iso;
	private Statistics stats;
	private Set<CFGNode> cfgNodes;
	private Map<CFGNode, StructHashSet<IRNode>> opt;
	private Map<CFGNode, StructHashSet<IRNode>> redn;
	private Map<StructWrapper,String> tempHash;
	private NameGenerator nameGen;
	
	private static class CriticalEdge {
		CFGNode from;
		CFGNode to;

		public CriticalEdge(CFGNode from, CFGNode to) {
			this.from = from;
			this.to = to;
		}
	};

	public PartialRedundancy(Statistics stats, NameGenerator nameGen) {
		this.stats = stats;
		this.nameGen = nameGen;
		tempHash = new HashMap<StructWrapper,String>();
		opt = new HashMap<CFGNode, StructHashSet<IRNode>>();
		redn = new HashMap<CFGNode, StructHashSet<IRNode>>();
	}

	public void run(Map<String, CFGNode> cfgs) {
		for (String name : cfgs.keySet()) {
			CFGNode cfgRoot = cfgs.get(name);
			fixCriticalEdges(nameGen, cfgRoot.flatten());
			analyze(name, cfgRoot);
			optimize(name);
		}
	}
	
	public static void fixCriticalEdges(NameGenerator nameGen, Set<CFGNode> nodes) {
		List<CriticalEdge> critEdges = new ArrayList<CriticalEdge>();
		for (CFGNode n : nodes) {
			if (n.outEdges().size() <= 1)
				continue;
			for (CFGNode n2 : n.outEdges()) {
				if (n2.inEdges().size() <= 1)
					continue;
				critEdges.add(new CriticalEdge(n, n2));
			}
		}

		for (CriticalEdge e : critEdges) {
			String targetName = e.to.value() instanceof String ?
					(String) e.to.value() : ((IRLabel) e.to.value()).name();
			String newLabelName = ".C"+nameGen.nextLabel().substring(2);
			CFGNode labelNode = new CFGNode(new IRLabel(newLabelName));
			CFGNode jumpNode = new CFGNode(new IRJump(new IRName(targetName)));
			e.from.removeEdgeTo(e.to);
			e.from.addEdgeTo(labelNode);
			labelNode.addEdgeTo(jumpNode);
			jumpNode.addEdgeTo(e.to);

			IRNode n = (IRNode) e.from.value();
			if (n instanceof IRJump) {
				e.from.setValue(new IRJump(new IRName(newLabelName)));
			} else if (n instanceof IRCJump) {
				IRCJump cJump = (IRCJump) e.from.value();
				if (targetName.equals(cJump.trueLabel()))
					e.from.setValue(new IRCJump(cJump.expr(), newLabelName, cJump.falseLabel()));
				else if (targetName.equals(cJump.falseLabel()))
					e.from.setValue(new IRCJump(cJump.expr(), cJump.trueLabel(), newLabelName));
				else
					throw new InternalCompilerException("Error while fixing CJump labels!");
			} else {
				throw new InternalCompilerException("Unexpected node type");
			}
		}
	}

	public void analyze(String funcName, CFGNode cfgRoot) {
		/* Perform the dataflow analyses required by PRE */
		cfgNodes = cfgRoot.flatten();
		
		stats.startTimer("pre/ante");
		ante = new AnticipatedExpressions(cfgNodes);
		ante.solve();
		stats.stopTimer("pre/ante");
		stats.startTimer("pre/earliest");
		early = new EarliestExpressions(ante, cfgNodes);
		early.solve();
		stats.stopTimer("pre/earliest");
		stats.startTimer("pre/delay");
		del = new Delayedness(cfgNodes, cfgRoot, ante, early);
		del.solve();
		stats.stopTimer("pre/delay");
		stats.startTimer("pre/latest");
		latest = new Latestness(cfgNodes, ante, del);
		stats.stopTimer("pre/latest");
		stats.startTimer("pre/iso");
		iso = new Isolated(cfgNodes, cfgRoot, ante, latest);
		iso.solve();
		stats.stopTimer("pre/iso");
		stats.startTimer("pre/opt");
		for (CFGNode node : cfgNodes) {
			StructHashSet<IRNode> optExprs = new StructHashSet<IRNode>(latest.in(node));
			optExprs.removeAll(iso.out(node));
			opt.put(node, optExprs);

			StructHashSet<IRNode> rednExprs = new StructHashSet<IRNode>(ante.eval(node));
			rednExprs.removeAll(latest.in(node));
			rednExprs.removeAll(iso.out(node));

			opt.put(node, optExprs);
			redn.put(node, rednExprs);
		}
		stats.stopTimer("pre/opt");
		stats.log("pre", "'" + funcName + "': total # of expressions : " + ante.allExprs.size());
		
	}
	
	public StructHashSet<IRNode> opt(CFGNode node) {
		return opt.get(node);
	}

	public StructHashSet<IRNode> redn(CFGNode node) {
		return redn.get(node);
	}

	public void optimize(String funcName) {
		Map<StructWrapper, IRExpr> redundant = new HashMap<StructWrapper, IRExpr>();
		List<IRExpr> optList = new ArrayList<IRExpr>();
		List<CFGNode> optNodes = new ArrayList<CFGNode>();
		tempHash.clear();
		int numOpt = 0;
		int numRedn = 0;

		for(CFGNode n : cfgNodes) {
			redundant.clear();
			optList.clear();
			optNodes.clear();
			
			/* don't save this expression if it won't be used */ 
			for(Iterator<IRNode> it= opt.get(n).iterator(); it.hasNext();) {
				IRNode expr = it.next();
				if (n.value() instanceof IRNode) {
					IRNode irnode = (IRNode)n.value();
					/* Always insert an optimal compute point if the expression is
					   used more than once in the statement */
					if(AvailableExpressions.countOccurrences((IRExpr)expr, irnode) > 1)
						continue;
				}
				/* Otherwise, check if it is ever redundant anywhere. If not,
				   we don't really need to save it to a TEMP */
				boolean used = false;
				for(StructHashSet<IRNode> rednval : redn.values()) 
					used |= rednval.contains(expr);
				if(!used)
					it.remove();
			}			
			
			/* Topological sort: subexpressions must come first! */
			for(IRNode expr : opt.get(n))
				optList.add((IRExpr)expr);
			sortExprs(optList);
			numOpt += optList.size();

			for (IRNode expr : optList) {
				StructWrapper wExpr = new StructWrapper(expr);
				String name;
				if(tempHash.containsKey(wExpr)) {
					name = tempHash.get(wExpr);
				} else {
					name = nameGen.nextTemp();
					tempHash.put(wExpr, name);
				}
				IRMove move = new IRMove(new IRTemp(name), 
						(IRExpr) rewrite((IRExpr) expr.deepCopy(), redundant));
				redundant.put(wExpr, new IRTemp(name));			

				/* insert new instruction */
				
				CFGNode newNode = new CFGNode(move);
				if (optNodes.size() > 0)
					optNodes.get(optNodes.size()-1).addEdgeTo(newNode);
				optNodes.add(newNode);
			}
			if (n.value() instanceof IRLabel) {
				if (n.outEdges().size() != 1) 
					throw new InternalCompilerException("IRLabel: unexpected # of outgoing edges");
				if (optNodes.size() > 0) {
					/* Insert computation after the label */
					CFGNode nextNode = n.outEdges().get(0);
					n.removeEdgeTo(nextNode);
					n.addEdgeTo(optNodes.get(0));
					optNodes.get(optNodes.size()-1).addEdgeTo(nextNode);
				}
			} else if (n.value() instanceof IRNode) {
				IRNode original = (IRNode)n.value();
				/* rewrite old instruction */		
				n.setValue(rewrite(original, redundant));
				if (n.inEdges().size() != 1)
					throw new InternalCompilerException("IRNode: unexpected # of incoming edges");
				if (optNodes.size() > 0) {
					/* Insert computation after the label */
					CFGNode prevNode = n.inEdges().get(0);
					prevNode.removeEdgeTo(n);
					prevNode.addEdgeTo(optNodes.get(0));
					optNodes.get(optNodes.size()-1).addEdgeTo(n);
				}
			}
		}

		for(CFGNode n : cfgNodes) {
			if(n.value() instanceof IRNode) {
				IRNode original = (IRNode)n.value();
				
				redundant.clear();
				/* create a rewrite map for the redundant expressions
				 * at this node.  This prevents rewrites of expressions
				 * that are optimal at this node, but may contain
				 * redundant subexpressions.
				 */
				for(IRNode expr : redn.get(n)) {
					StructWrapper wExpr = new StructWrapper(expr);
					if(!tempHash.containsKey(wExpr)) {
						throw new InternalCompilerException("unknown redundant expression: "+ expr);
					} else {
						redundant.put(wExpr, new IRTemp( tempHash.get(wExpr)));
						
					}								
				}

				numRedn += redundant.size();
				n.setValue(rewrite(original, redundant));
			}
		}
		if (numOpt > 0)
			stats.log("pre", "'" + funcName + "': inserted " + numOpt + " optimal compute points");
		if (numRedn > 0)
			stats.log("pre", "'" + funcName + "': removed " + numRedn + " redundant computations");
	}

	public static IRNode rewrite(IRNode irNode, Map<StructWrapper, IRExpr> tempHash) {
		return (IRNode) new RewritingVisitor(tempHash).visit(irNode);
	}

	/* To make sure we emit code in the right order, 
	 * sort expressions by subexpression.
	 * this is expensive, but sets should be small */
	public List<IRExpr> sortExprs(List<IRExpr> exprs) {
		if(exprs.size() <= 1) {
			return exprs;
		} else if(exprs.size() == 2) {
			if(isSubExpression(exprs.get(0),exprs.get(1)))
				return exprs;
			else {
				/* swap */
				IRExpr e1 = exprs.remove(0);
				exprs.add(e1);
				return exprs;
			}
		} else {
			/* Need to do a topological sort */
			ArrayDeque<StructWrapper> exprQueue = new ArrayDeque<StructWrapper>();
			HashMap<StructWrapper,HashSet<StructWrapper>> edges = 
				new HashMap<StructWrapper,HashSet<StructWrapper>>();
			HashMap<StructWrapper,HashSet<StructWrapper>> invEdges = 
				new HashMap<StructWrapper,HashSet<StructWrapper>>();
			List<IRExpr> sorted = new ArrayList<IRExpr>();
						
			for(IRExpr e : exprs) {
				HashSet<StructWrapper> subExprs = 
					new HashSet<StructWrapper>();
				StructWrapper wExpr = new StructWrapper(e);
				for(IRExpr sub : exprs)
					if(isSubExpression(sub,e)) {
						StructWrapper wSub = new StructWrapper(sub);
						subExprs.add(wSub);
						if(invEdges.containsKey(wSub))
							invEdges.get(wSub).add(wExpr);
						else {
							HashSet<StructWrapper> parents 
								= new HashSet<StructWrapper>();
							parents.add(wExpr);
							invEdges.put(wSub,parents);
						}
					}
				edges.put(wExpr, subExprs);
			}

			for(IRExpr e: exprs) {
				StructWrapper wExpr = new StructWrapper(e);
				if(!invEdges.containsKey(wExpr))
					exprQueue.add(wExpr);

			}

			while(!exprQueue.isEmpty()) {
				StructWrapper wExpr = exprQueue.remove();
				sorted.add((IRExpr)wExpr.value());
				//System.out.println("Added node " +wExpr.value());
				if(edges.containsKey(wExpr)) {
 					for(Iterator<StructWrapper> it 
							= edges.get(wExpr).iterator(); it.hasNext();) {
						StructWrapper wSubExpr = it.next();
						/* remove edge */
						it.remove();
						if(invEdges.containsKey(wSubExpr)) {
							invEdges.get(wSubExpr).remove(wExpr);

							/* if no more incoming edges, add to Q */
							if(invEdges.get(wSubExpr).size()==0)
								exprQueue.add(wSubExpr);
						} else
							exprQueue.add(wSubExpr);
					}
 					if(edges.get(wExpr).size()==0)
 						edges.remove(wExpr);
				}
			}
			if(edges.size()!=0)
				throw new InternalCompilerException("unknown error while sorting subexpressions");
			else
				return sorted;
		}
	}

	private boolean isSubExpression(IRExpr sub, IRExpr e) {
		StructHashSet<IRNode> exprs = AvailableExpressions.expressions(e);
		exprs.remove(e);
		return exprs.contains(sub);
	}

	public StructHashSet<IRNode> latestness(CFGNode node) {
		return latest.in(node);
	}
}
