package xic.analysis;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.Instr;


/**
 * Reaching definitions analysis - only implemented for assembly for now.
 */
public class ReachingDefinitions extends DataFlowAnalysis {
	/// Value class used during live variable analysis
	public static class RDValue implements Value {
		/* List of variable names */
		private BitSet defs;

		public RDValue(int size) {
			defs = new BitSet(size);
		}

		public RDValue(BitSet vars) {
			this.defs = (BitSet) vars.clone();
		}

		public Value copy() {
			return new RDValue(defs);
		}

		public void meet(Value v) {
			defs.or(((RDValue) v).defs);
		}
		
		public String toString() {
			return defs.toString();
		}

		public int hashCode() {
			return defs.hashCode();
		}
		
		public BitSet defs() {
			return defs;
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof RDValue))
				return false;
			return defs.equals(((RDValue) obj).defs);
		}

		public void setNullMeet() {
			defs.clear();
		}
	};

	/// Dataflow-related information associated with each node
	public class RDFlowInfo extends FlowInfo {
		public BitSet prsv, gen;

		public RDFlowInfo(CFGNode node) {
			entry = new RDValue(defs.size());
			exit = new RDValue(defs.size());
			gen = new BitSet(defs.size());
			prsv = new BitSet(defs.size());
			prsv.set(0, prsv.size());

			if (!defToIndex.containsKey(node))
				return;

			int defIndex = defToIndex.get(node);
			List<Integer> kill = varToDefIndex.get(defIndexToVar.get(defIndex));
			for (int i: kill)
				prsv.clear(i);
			gen.set(defIndex);
		}
	};

	private Map<String, List<Integer>> varToDefIndex;
	private Map<Integer, String> defIndexToVar;
	private Map<CFGNode, Integer> defToIndex;
	private List<CFGNode> defs;
	
	public ReachingDefinitions(Set<CFGNode> nodes) {
		super(nodes);
		initFlowInfo();

		dir = Direction.FORWARD;
		top = new RDValue(defs.size());
	}

	protected void initFlowInfo() {
		Set<String> use = new HashSet<String>();
		Set<String> def = new HashSet<String>();

		defToIndex = new HashMap<CFGNode, Integer>();
		defs = new ArrayList<CFGNode>();
		defIndexToVar = new HashMap<Integer, String>();
		varToDefIndex = new HashMap<String, List<Integer>>();

		for (CFGNode n : nodes) {
			Object o = n.value();
			if (o instanceof Instr) {
				((Instr) o).analyze(use, def);
				if (def.size() == 0)
					continue;
				if (def.size() > 1)
					throw new InternalCompilerException("Instruction defines more than 1 variable!");
				String defName = def.iterator().next();

				if (!varToDefIndex.containsKey(defName))
					varToDefIndex.put(defName, new ArrayList<Integer>());
				varToDefIndex.get(defName).add(defs.size());
				defIndexToVar.put(defs.size(), defName);
				defToIndex.put(n, defs.size());
				defs.add(n);
			}			
		}
		
		super.initFlowInfo();
	}

	
	public BitSet getReachingDefinitionsBefore(CFGNode node) {
		return ((RDValue) ((RDFlowInfo) getFlowInfo(node)).entry).defs;
	}

	public BitSet getReachingDefinitionsAfter(CFGNode node) {
		return ((RDValue) ((RDFlowInfo) getFlowInfo(node)).exit).defs;
	}

	public List<Integer> getDefinitionsForVariable(String name) {
		return varToDefIndex.get(name);
	}

	public List<CFGNode> getDefinitions() {
		return defs;
	}
	
	public CFGNode getDefinition(int i) {
		return defs.get(i);
	}
	
	public int getDefinitionIndex(CFGNode node) {
		if (!defToIndex.containsKey(node))
			return -1;
		return defToIndex.get(node);
	}

	public String getVarName(CFGNode def) {
		return defIndexToVar.get(defToIndex.get(def));
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new RDFlowInfo(node);
	}

	public void F(FlowInfo flowInfo) {
		RDFlowInfo f = (RDFlowInfo) flowInfo;
		BitSet exit = ((RDValue) f.exit).defs;
		BitSet entry = ((RDValue) f.entry).defs;
		
		exit.clear();
		exit.or(entry);
		exit.and(f.prsv);
		exit.or(f.gen);
	}
}