package xic.ast;

import java.util.ArrayList;
import java.util.List;

import xic.ErrorSink;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.xi.CompilationException;
import edu.cornell.cs.cs4120.xi.Position;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * Describes an array of elements whose type is given
 * by a TypeNode instance.
 */
public class ArrayType extends TypeNode {	
	/* Static types without position information */
	public static final ArrayType INT_ARRAY_TYPE = 
		new ArrayType(PrimitiveType.INTEGER_TYPE);
	public static final ArrayType ANYTYPE_ARRAY_TYPE=
		new ArrayType( UnitType.UNIT_TYPE);
		
	private TypeNode elementType = null;
	private Expr lengthExpr = null;

	public ArrayType(Position pos, TypeNode subType, Expr lengthExpr) {
		super(pos);
		this.elementType = subType;
		this.lengthExpr = lengthExpr;
	}

	public ArrayType(TypeNode subType) {
		this.elementType = subType;
	}

	public TypeNode elementType() {
		return elementType;
	}
	
	public boolean hasLengthExpr() {
		return lengthExpr != null;
	}

	public Expr lengthExpr() {
		return lengthExpr;
	}

	public Iterable<VisualizableTreeNode> children() {
		List<VisualizableTreeNode> result = new ArrayList<VisualizableTreeNode>();

		result.add(elementType);
		if (hasLengthExpr())
			result.add(lengthExpr);

		return result;
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	TypeNode elementType = (TypeNode) v.visit(this, this.elementType);
		Expr lengthExpr = (Expr) v.visit(this, this.lengthExpr);

		if (elementType != this.elementType || lengthExpr != this.lengthExpr)
			return new ArrayType(position, elementType, lengthExpr);

		return this;
	}
    
    public Copyable copy() {
		return new ArrayType(position, elementType, lengthExpr);
    }

	public String label() {
		return "array";
	}

	public boolean isSubtypeOf(TypeNode node) {
		if (node instanceof ArrayType) {
			TypeNode elType = ((ArrayType) node).elementType();
			if (elType instanceof UnitType)
				return true;
			if (containsClassType(elType) || containsClassType(elementType))
				return elementType.isIdenticalType(elType);
			return elementType.isSubtypeOf(elType);
		} else if (node instanceof UnitType) {
			return true;
		} else if (node instanceof TupleType) {
			TupleType tnode = (TupleType)node;
			if (tnode.elementTypes().size() == 1)
				return isSubtypeOf(tnode.elementTypes().get(0));				
			else
				return false;
		} else {
			return false;
		}
	}

	private boolean containsClassType(TypeNode type) {
		if (type instanceof ClassType) {
			return true;
		} else if (type instanceof ArrayType) {
			return containsClassType(((ArrayType) type).elementType);
		} else if (type instanceof TupleType) {
			boolean result = false;
			for (TypeNode t : ((TupleType) type).elementTypes())
				result |= containsClassType(t);
			return result;
		} else {
			return false;
		}
	}

	public String toString() {
		return elementType.toString() + "[]";
	}

	public int size() {
		return 1;
	}

	public boolean isIdenticalType(TypeNode node) {
		if (!(node instanceof ArrayType))
			return false;

		ArrayType other = (ArrayType) node;
		if (hasLengthExpr() && !other.hasLengthExpr()
		 || !hasLengthExpr() && other.hasLengthExpr())
			return false;

		if (hasLengthExpr() && lengthExpr.equals(other.lengthExpr))
			return false;
		
		return elementType.isIdenticalType(other.elementType);
	}

	/**
	 * Used to reverse the order of nested array types for 
	 * more intuitive indexing behavior.
	 */
	public static ArrayType nestArrayType(Position p, ArrayType t, Expr l) {
		ArrayType backup = t;
		while (t.elementType instanceof ArrayType)
			t = (ArrayType) t.elementType;
		t.elementType = new ArrayType(p, t.elementType, l);
		return backup;
	}
	
	public void match(TypeNode node, ErrorSink sink) {
		if (node instanceof ArrayType) {
			ArrayType at = (ArrayType) node;
			at.elementType.match(this.elementType,sink);
		}
		else 
			sink.addError(new CompilationException("Unmatched type, excepted: "
					+this.label()+", got: "+node.label(), node.position()));
	}
}
