package xic.ast;

import java.util.ArrayList;
import java.util.List;

import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.xi.Position;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * This class describes Iota_9 tuple declarations (as in <tt>(i:int, j:int)</tt>).
 * To simplify the number of possible cases, all forms of variable declarations
 * are treated as instances of a general nested scheme. 
 * 
 * The rules are as follows:
 * Tuple declarations are either bound to a name and a type, or they are unnamed and
 * consist of nested tuple declarations. A valid tree of tuple declarations must consist
 * of bound leaf nodes and unbound inner nodes. Types of leaf nodes are fixed, while
 * the types of inner nodes can easily be inferred from the tree structure. The 
 * underscore token is treated as a special leaf tuple declaration with name == null 
 * and the type set to an instance of <code>UnitType</tt>.
 * 
 * The expressiveness of this approach makes it possible to represent a wide range of
 * declarations, for example
 * - <tt>i:int</tt>
 * - <tt>(i: int, j:int>)</tt>
 * - <tt>(i: int, j:(int, int))</tt>
 * - <tt>(i: int, (j:int, _, k:int))</tt>
 */
public class TupleDecl extends ASTNode implements Typed {
	private String name;
	private TypeNode type;
	private List<TupleDecl> subTupleDecls;
	private DataSource dataSource;

	/**
	 * Create a name->type binding
	 */
	public TupleDecl(Position pos, String name, TypeNode type) {
		super(pos);
		this.name = name;
		this.type = type;
		subTupleDecls = null; // there will be no children
	}
	
	/**
	 * Create a tuple declaration with several variables of the same type
	 */
	public TupleDecl(Position pos, IdentifierList il, TypeNode type) {
		super(pos);
		this.name = null; // not bound to a name
		this.type = null;
		subTupleDecls = new ArrayList<TupleDecl>();
		for (String name : il)
			subTupleDecls.add(new TupleDecl(pos, name, type));
	}


	/**
	 * Create a tuple declaration consisting of child tuple declarations
	 */
	public TupleDecl() {
		super(null);
		this.name = null; // not bound to a name
		this.type = null;
		subTupleDecls = new ArrayList<TupleDecl>();
	}

	/**
	 * Create a tuple declaration consisting of child tuple declarations.
	 * Adds the first element, for convenience.
	 */
	public TupleDecl(TupleDecl firstChild) {
		this();
		appendTupleDecl(firstChild);
	}

	/**
	 * Copy constructor, makes a shallow copy while
	 * providing new position information.
	 */
	public TupleDecl(Position pos, TupleDecl decl) {
		super(pos);
		this.name = decl.name();
		this.type = decl.getType();
		this.subTupleDecls = decl.subTupleDecls();
	}

	public TupleDecl(Position pos, String name, TypeNode type,
			List<TupleDecl> subTupleDecls, DataSource dataSource) {
		super(pos);
		this.name = name;
		this.type = type;
		this.subTupleDecls = subTupleDecls;
		this.dataSource = dataSource;
	}

	public boolean isLeaf() {
		return subTupleDecls == null;
	}

	public boolean isUnderscore() {
		return (isLeaf() && name == null && type instanceof UnitType);		
	}

	public String name() {
		return name;
	}
	
    public TupleDecl peel() {
    	return subTupleDecls.get(0);
    }

    public TupleDecl remainder() {
		TupleDecl td = new TupleDecl();
		TupleType type = new TupleType();
		DataSource ds = new DataSource();
		for (int i=1; i<subTupleDecls.size(); ++i) {
			type.appendType(subTupleDecls.get(i).getType());
			td.appendTupleDecl(subTupleDecls.get(i));
			ds.add(subTupleDecls.get(i).getDataSource());
		}
		return (TupleDecl) td.setDataSource(ds).setType(type);
    }


	public void appendTupleDecl(TupleDecl child) {
		subTupleDecls.add(child);
		extendPositionTo(child.position());
	}
	
	public List<TupleDecl> subTupleDecls() {
		return subTupleDecls;
	}
	
	public void setSubTupleDecls(List<TupleDecl> decls) {
		this.subTupleDecls = decls;
	}

	public TypeNode getType() {
		return type;
	}

	public Typed setType(TypeNode type) {
		if (type == this.type)
			return this;

		TupleDecl e = (TupleDecl) copy();
		e.type = type;
		return e;
	}	

	public DataSource getDataSource() {
		return dataSource;
	}

	public Typed setDataSource(DataSource dataSource) {
		if (dataSource == this.dataSource)
			return this;

		TupleDecl e = (TupleDecl) copy();
		e.dataSource = dataSource;
		return e;
	}

	public String label() {
		if (isLeaf()) {
			if (!isUnderscore())
				return "tupleDecl[" + name + "]";
			else
				return "tupleDecl[undescore]";
		} else {
			return "tupleDecl";
		}
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	boolean modified = false;

    	TypeNode type = (TypeNode) v.visit(this, this.type);
    	if (type != this.type)
    		modified = true;

    	ArrayList<TupleDecl> results = null;
    	if (!isLeaf()) {
        	results = new ArrayList<TupleDecl>();
	    	for (TupleDecl declNode : subTupleDecls) {
	    		TupleDecl newDecl = (TupleDecl) v.visit(this, declNode);
	    		if (declNode != newDecl)
	    			modified = true;
	    		results.add(newDecl);
	    	}
    	}

		if (modified)
			return new TupleDecl(position, this.name, type, results, dataSource);

		return this;
	}

	public Iterable<VisualizableTreeNode> children() {
		List<VisualizableTreeNode> result = new ArrayList<VisualizableTreeNode>();
		if (isLeaf()) {
			if (!isUnderscore())
				result.add(type);
		} else {
			result.addAll(subTupleDecls);
		}
		return result;
	}

	public Copyable copy() {
		ArrayList<TupleDecl> subTupleDecls = (this.subTupleDecls == null) 
			? null: new ArrayList<TupleDecl>(this.subTupleDecls);
		return new TupleDecl(position, this.name, type, subTupleDecls, dataSource);
	}

	public List<String> getNames() {
		ArrayList<String> names = new ArrayList<String>();
		if (isLeaf())
			names.add(name);
		else {
			for (TupleDecl td : subTupleDecls)
				names.addAll(td.getNames());
		}
		return names;
	}
}