package xic.codegen;

import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRFuncDecl;
import xic.ir.IRMem;
import xic.ir.IRTemp;

/**
 * Register allocator interface:
 * 
 * The method regAlloc takes an instruction list and resolves 
 * all the temp names to registers or stacks. The result is
 * returned in the form of a new <code>InstrList</code> instance.
 * The isFunction parameter indicates whether a preamble / epilog
 * are required.
 */
public abstract class RegAlloc {
	/* Some special stack-related names that are used in the IR, 
	   pseudo-assembly & dataflow analyses */
	public static final String CALL_RV_SCRATCH      = "RV_SCRATCH";
	public static final String CALL_ARG_SCRATCH     = "ARG_SCRATCH";
	public static final String SPILL_SCRATCH        = "SPILL_SCRATCH";
	public static final String MEM_NAME             = "MEM";
	public static final String FP_NAME              = "FP";
	public static final String ABSTRACT_REG_PREFIX  = "REG";
	public static final String RV_NAME              = "RV";
	public static final String THIS_NAME            = "THIS";
	public static final String OFFSET_NAME          = "OFFSET";

	public abstract InstrList regAlloc(
		String funcName, IRFuncDecl irDecl, InstrList il, boolean isFunction);

	public static boolean isAStackPointer(String name) {
		return name != null && (name.equals(FP_NAME) 
				|| name.equals(CALL_RV_SCRATCH)
				|| name.equals(CALL_ARG_SCRATCH)
				|| name.equals(SPILL_SCRATCH));
	}

	public static boolean isAbstractRegisterName(String name) {
		return name.startsWith(ABSTRACT_REG_PREFIX);
	}

	public static int getAbstractRegisterIndex(String name) {
		return Integer.parseInt(name.substring(ABSTRACT_REG_PREFIX.length()));
	}

	/// Return a simple alias name for certain memory expressions
	public static String alias(IRMem mem) {
		IRExpr memExpr = mem.expr();
		String tempName = null;

		if (memExpr instanceof IRTemp) {
			tempName = ((IRTemp) memExpr).name();
		} else if (memExpr instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) memExpr;
			IRExpr left = binOp.left();
			IRExpr right = binOp.right();

			if (!(right instanceof IRConst) || !(left instanceof IRTemp))
				return MEM_NAME;

			tempName = ((IRTemp) left).name();
		}
		
		if (tempName != null &&
			(tempName.equals(CALL_RV_SCRATCH) 
				|| tempName.equals(CALL_ARG_SCRATCH)
				|| tempName.equals(FP_NAME)))
			return tempName;

		return MEM_NAME;
	}
}
