package xic.codegen.matchers;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import xic.Configuration;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;


/**
 * Helper class for detecting Intel-specific addressing modes in IR trees. Exhaustively
 * searches through the space of possible combinations
 */
public class AddrModeHelper {
	public static class AddrMode {
		private Tile base;
		private int scale;
		private Tile index;
		private long displacement;
		private String immediateName;
		
		protected AddrMode copy()  {
			AddrMode a = new AddrMode();
			a.base = base;
			a.scale = scale;
			a.index = index;
			a.displacement = displacement;
			a.immediateName = immediateName;
			return a;
		}

		/// Turn into an instruction operand
		public Instr.Op toOp(InstrList il, NameGenerator ng) {
			String base = hasBase() ? this.base.translate(il, ng) : null;
			String index = null;

			if (hasIndex()) {
				if (this.index == this.base)
					index = base;
				else
					index = this.index.translate(il, ng);
			}

			Instr.Op memOp = new Instr.Op(Instr.Op.Type.MEM);
			memOp.setBase(base);
			memOp.setIndex(index);
			memOp.setScale(scale);
			memOp.setImmediate(displacement);
			memOp.setImmediateName(immediateName);

			return memOp;
		}

		/// Estimate the cost created from having to compute operands in the addressing mode
		public int estimateCost() {
			int cost = 0;
			if (hasBase())
				cost += base.estimateCost();
			if (hasIndex())
				cost += index.estimateCost();
			return cost;
		}

		public Tile getBase() {
			return base;
		}
		
		public int getScale() {
			return scale;
		}
		
		public Tile getIndex() {
			return index;
		}
		
		public long getDisplacement() {
			return displacement;
		}

		public boolean hasBase() {
			return base != null;
		}
		
		public boolean hasIndex() {
			return index != null;
		}
	
		public boolean hasDisplacement() {
			return displacement != 0;
		}
		
		public String getImmediateName() {
			return immediateName;
		}
		
		public boolean hasImmediateName() {
			return immediateName != null;
		}

		public String toString() {
			StringBuffer buf = new StringBuffer();
			buf.append("[");
			boolean empty = true;

			if (hasBase()) {
				buf.append(base);
				empty = false;
			}
	
			if (hasIndex()) {
				if (!empty)
					buf.append('+');
				buf.append(index);
				buf.append('*').append(scale);
				empty = false;
			}

			if (hasImmediateName()) {
				if (!empty)
					buf.append('+');
				buf.append(immediateName);
				empty = false;
			}

			if (hasDisplacement()) {
				if (!empty)
					buf.append('+');
				buf.append(displacement);
			}
			buf.append(", cost=" + estimateCost() + "]");
			return buf.toString();
		}

		/// Merge two addressing modes if the don't conflict
		public AddrMode merge(AddrMode other) {
			if (other.index != null && index != null)
				return null; // conflict
			if (other.base != null && base != null)
				return null; // conflict
			if (other.immediateName != null && immediateName != null)
				return null; // conflict
			AddrMode mode = new AddrMode();
			mode.index = (index != null) ? index : other.index;
			mode.scale = (index != null) ? scale : other.scale;
			mode.immediateName = (immediateName != null) ? immediateName : other.immediateName;
			mode.base = (base != null) ? base : other.base;
			mode.displacement = displacement + other.displacement;
			return mode;
		}
	}

	private static final Tile FAKE_TILE_COST_0 = new Tile() {
		public int estimateCost() {
			return 0;
		}
		public String translate(InstrList il, NameGenerator ng) {
			return null;
		}
	};

	private static final Tile FAKE_TILE_COST_1 = new Tile() {
		public int estimateCost() {
			return 1;
		}
		public String translate(InstrList il, NameGenerator ng) {
			return null;
		}
	};


	private static int getPowerOfTwo(IRBinOp binOp, IRExpr expr, boolean constIsRight) {
		if (!(expr instanceof IRConst))
			return -1;

		long value = ((IRConst) expr).value();
		if (binOp.opType() == IRBinOp.OpType.MUL
				&& (value == 1 || value == 2 || value == 4 || value == 8))
			return (int) value;
		else if (binOp.opType() == IRBinOp.OpType.AR_LSHIFT && value <= 3 && constIsRight)
			return 1 << value;

		return -1;
	}

	private static boolean testPowerOfTwoPlus1(IRBinOp binOp, IRExpr expr) {
		if (!(expr instanceof IRConst))
			return false;
		long value = ((IRConst) expr).value() - 1;

		if (binOp.opType() == IRBinOp.OpType.MUL)
			return value == 1 || value == 2 || value == 4 || value == 8;
		return false;
	}

	/**
	 * Used below to check if an IR tree can be turned into
	 * an addressing mode with effective zero additional cost.
	 * This is useful to throw away useless (basically free) expressions
	 * in AvailableExpressions and some of the derived analyses.
	 */
	private static List<AddrMode> hasZeroCostAddressingMode(IRExpr node, int depth) {
		List<AddrMode> result = new ArrayList<AddrMode>();
		AddrMode mode;

		if (depth > 3)
			return result;

		mode = new AddrMode();
		mode.base = fakeTile(node);
		result.add(mode);

		mode = new AddrMode();
		mode.index = fakeTile(node);
		mode.scale = 1;
		result.add(mode);

		if (node instanceof IRConst) {
			mode = new AddrMode();
			mode.displacement = ((IRConst) node).value();
			result.add(mode);
		} else if (node instanceof IRName) {
			mode = new AddrMode();
			mode.immediateName = ((IRName) node).name();
			if (Configuration.IP_RELATIVE_ADDRESSING)
				mode.base = fakeTile(new IRTemp(Configuration.IP_REGISTER));
			result.add(mode);
		} else if (node instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) node;
			IRExpr left = binOp.left(), right = binOp.right();

			if (binOp.opType() == IRBinOp.OpType.ADD) {
				List<AddrMode> leftModes  = hasZeroCostAddressingMode(left, depth+1);
				List<AddrMode> rightModes = hasZeroCostAddressingMode(right, depth+1);

				/// O(n^2) search, n very small since we only do shallow traversals
				for (int i=0; i<leftModes.size(); ++i) {
					for (int j=0; j<rightModes.size(); ++j) {
						AddrMode addrMode = leftModes.get(i).merge(rightModes.get(j));
						if (addrMode != null)
							result.add(addrMode);
					}
				}
			} else if (binOp.opType() == IRBinOp.OpType.SUB && right instanceof IRConst) {
				List<AddrMode> leftModes  = hasZeroCostAddressingMode(left, depth+1);
				AddrMode rightMode = new AddrMode();
				rightMode.displacement = -((IRConst) right).value();
				for (int i=0; i<leftModes.size(); ++i) {
					AddrMode addrMode = leftModes.get(i).merge(rightMode);
					if (addrMode != null)
						result.add(addrMode);
				}
			}

			if (getPowerOfTwo(binOp, left, false) > 0) {
				mode = new AddrMode();
				mode.scale = getPowerOfTwo(binOp, left, false);
				mode.index = fakeTile(right);
				result.add(mode);
			} 
			
			if (getPowerOfTwo(binOp, right, true) > 0) {
				mode = new AddrMode();
				mode.scale = getPowerOfTwo(binOp, right, true);
				mode.index = fakeTile(left);
				result.add(mode);
			}

			if (testPowerOfTwoPlus1(binOp, left)) {
				mode = new AddrMode();
				mode.scale = (int) ((IRConst) left).value()-1;
				mode.index = fakeTile(right);
				mode.base = mode.index;
				result.add(mode);
			}

			if (testPowerOfTwoPlus1(binOp, right)) {
				mode = new AddrMode();
				mode.scale = (int) ((IRConst) right).value()-1;
				mode.index = fakeTile(left);
				mode.base = mode.index;
				result.add(mode);
			}
		}
		return result;
	}
	
	
	private static Tile fakeTile(IRExpr node) {
		if (node instanceof IRTemp)
			return FAKE_TILE_COST_0;
		else
			return FAKE_TILE_COST_1;
	}

	private static List<AddrMode> findAddressingModes(Tiler tiler, IRNode node, boolean allowDirect, int depth) {
		List<AddrMode> result = new ArrayList<AddrMode>();
		AddrMode mode;

		if (depth > 3)
			return result;

		if (allowDirect) {
			mode = new AddrMode();
			mode.base = tiler.tile(node);
			result.add(mode);

			mode = new AddrMode();
			mode.index = tiler.tile(node);
			mode.scale = 1;
			result.add(mode);
		}

		if (node instanceof IRConst) {
			mode = new AddrMode();
			mode.displacement = ((IRConst) node).value();
			result.add(mode);
		} else if (node instanceof IRName) {
			mode = new AddrMode();
			mode.immediateName = ((IRName) node).name();
			if (Configuration.IP_RELATIVE_ADDRESSING)
				mode.base = tiler.tile(new IRTemp(Configuration.IP_REGISTER));
			result.add(mode);
		} else if (node instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) node;
			IRExpr left = binOp.left(), right = binOp.right();

			if (binOp.opType() == IRBinOp.OpType.ADD) {
				List<AddrMode> leftModes  = findAddressingModes(tiler, left, true, depth+1);
				List<AddrMode> rightModes = findAddressingModes(tiler, right, true, depth+1);

				/// O(n^2) search, n very small since we only do shallow traversals
				for (int i=0; i<leftModes.size(); ++i) {
					for (int j=0; j<rightModes.size(); ++j) {
						AddrMode addrMode = leftModes.get(i).merge(rightModes.get(j));
						if (addrMode != null)
							result.add(addrMode);
					}
				}
			} else if (binOp.opType() == IRBinOp.OpType.SUB && right instanceof IRConst) {
				List<AddrMode> leftModes  = findAddressingModes(tiler, left, allowDirect, depth+1);
				AddrMode rightMode = new AddrMode();
				rightMode.displacement = -((IRConst) right).value();
				for (int i=0; i<leftModes.size(); ++i) {
					AddrMode addrMode = leftModes.get(i).merge(rightMode);
					if (addrMode != null)
						result.add(addrMode);
				}
			}

			if (getPowerOfTwo(binOp, left, false) > 0) {
				mode = new AddrMode();
				mode.scale = getPowerOfTwo(binOp, left, false);
				mode.index = tiler.tile(right);
				result.add(mode);
			} 
			
			if (getPowerOfTwo(binOp, right, true) > 0) {
				mode = new AddrMode();
				mode.scale = getPowerOfTwo(binOp, right, true);
				mode.index = tiler.tile(left);
				result.add(mode);
			}

			if (testPowerOfTwoPlus1(binOp, left)) {
				mode = new AddrMode();
				mode.scale = (int) ((IRConst) left).value()-1;
				mode.index = tiler.tile(right);
				mode.base = mode.index;
				result.add(mode);
			}

			if (testPowerOfTwoPlus1(binOp, right)) {
				mode = new AddrMode();
				mode.scale = (int) ((IRConst) right).value()-1;
				mode.index = tiler.tile(left);
				mode.base = mode.index;
				result.add(mode);
			}
		}
		return result;
	}

	/// Filter out invalid addressing modes
	public static List<AddrMode> filter(List<AddrMode> result) {
		for (Iterator<AddrMode> it = result.iterator(); it.hasNext(); ) {
			AddrMode mode = it.next();
			if (!mode.hasBase() && !mode.hasIndex() && !mode.hasImmediateName())
				it.remove();
			if (mode.hasIndex() && mode.getScale() == 1 && !mode.hasBase()) {
				mode.base = mode.index;
				mode.index = null;
			}
			if (Configuration.IP_RELATIVE_ADDRESSING && mode.hasImmediateName() && mode.hasBase() && mode.hasIndex()) {
				/* Things like [rip + rcx*9 + symbolName] are not allowed */
				it.remove();
			}
		}
		return result;
	}

	public static List<AddrMode> findMemAddressingModes(Tiler tiler, IRMem node) {
		return filter(findAddressingModes(tiler, node.expr(), true, 0));
	}

	/**
	 * Used below to check if an IR tree can be turned into
	 * an addressing mode with effective zero additional cost.
	 * This is useful to throw away useless (basically free) expressions
	 * in AvailableExpressions and some of the derived analyses.
	 */
	public static boolean hasZeroCostAddressingMode(IRExpr expr) {
		List<AddrMode> tiles = hasZeroCostAddressingMode(expr, 0);
		for (AddrMode mode: tiles)
			if (mode.estimateCost() == 0)
				return true;
		return false;
	}

	/**
	 * Special LEA addressing mode search -- prohibits operations 
	 * of the form
	 *  
	 *  lea t1, [t2]
	 *  
	 * to prevent infinite recursion.
	 */
	public static List<AddrMode> findLeaAddressingModes(Tiler tiler, IRBinOp node) {
		return filter(findAddressingModes(tiler, node, false, 0));
	}
}
