package xic.codegen.matchers;

import java.util.List;

import xic.InternalCompilerException;
import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;
import xic.ir.IRBinOp.OpType;


/**
 * Handle unary operations -- only NEG for now, since the 'NOT'
 * instruction does not implement a logical NOT.
 */
public class UnaryOpMatcher implements Matcher {
	protected static class UnaryOpTile extends AbstractTile {
		private OpType opType;
		private AddrMode addrMode;
		private boolean overwrite;

		public UnaryOpTile(OpType opType, Tile tile, boolean overwrite) {
			this.opType = opType;
			this.overwrite = overwrite;
			children.add(tile);
		}

		public UnaryOpTile(OpType opType, AddrMode addrMode, boolean overwrite) {
			this.opType = opType;
			this.addrMode = addrMode;
			this.overwrite = overwrite;
		}

		public int localCost() {
			return 1 + (addrMode == null ? 0 : addrMode.estimateCost());
		}

		public String translate(InstrList il, NameGenerator ng) {
			Instr.Op op;
			String mnemonic;

			switch (opType) {
				case SUB: mnemonic = "neg"; break;
				default:
					throw new InternalCompilerException("Invalid unary operation");
			}

			if (addrMode != null)
				op = addrMode.toOp(il, ng);
			else
				op = new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng));

			if (!overwrite) {
				String tempName = ng.nextTemp();
				il.add(new Instr("mov",
						new Instr.Op(Instr.Op.Type.REG, tempName), op));
				il.add(new Instr(mnemonic,
						new Instr.Op(Instr.Op.Type.REG, tempName)));
				return tempName;
			} else {
				il.add(new Instr(mnemonic, op));
				return null; // This is a statement
			}
		}
	}

	public Tile match(IRNode node, Tiler tiler) {
		Tile bestTile = null;
		int bestCost = Integer.MAX_VALUE;

		if (node instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) node;
			IRBinOp.OpType opType = binOp.opType();
			
			if (opType == IRBinOp.OpType.SUB
					&& binOp.left() instanceof IRConst
					&& ((IRConst) binOp.left()).value() == 0) {

				bestTile = new UnaryOpTile(binOp.opType(), tiler.tile(binOp.right()), false);

				if (binOp.right() instanceof IRMem) {
					IRMem mem = (IRMem) binOp.right();
					List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);
		
					for (AddrMode addrMode : modes) {
						Tile tile = new UnaryOpTile(binOp.opType(), addrMode, false);
						int cost = tile.estimateCost();
						if (cost < bestCost) {
							bestTile = tile;
							bestCost = cost;
						}
					}
				}
			}
		}
		
		if (!(node instanceof IRMove))
			return bestTile;
		IRMove move = (IRMove) node;

		if (!(move.expr() instanceof IRBinOp
			&& (move.target() instanceof IRMem || move.target() instanceof IRTemp)))
			return bestTile;

		IRExpr target = ((IRMove) node).target();
		IRBinOp binOp = (IRBinOp) move.expr();
		IRBinOp.OpType opType = binOp.opType();

		if (!(opType == IRBinOp.OpType.SUB
				&& binOp.left() instanceof IRConst
				&& ((IRConst) binOp.left()).value() == 0))
			return bestTile;

		if (target.equalsTree(binOp.right())) {
			if (binOp.right() instanceof IRMem) {
				List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) binOp.right());
				for (AddrMode addrMode : modes) {
					Tile tile = new UnaryOpTile(binOp.opType(), addrMode, true);
					int cost = tile.estimateCost();
					if (cost < bestCost) {
						bestTile = tile;
						bestCost = cost;
					}
				}
			} else if (binOp.right() instanceof IRTemp) {
				bestTile = new UnaryOpTile(binOp.opType(), tiler.tile(binOp.right()), true);
			}
		}

		return bestTile;
	}
}
