package xic.ir;

import java.util.HashSet;
import java.util.Set;

import xic.ast.ASTNode;
import xic.ast.Copyable;


public abstract class IRNode extends ASTNode {
	
	public IRNode() {
		super(null); // IR nodes do not use position information
	}

	/* Helper function for IR tree comparisons 
	  - in a separate function to preserve equals()/hasCode() semantics
	*/
	public abstract boolean equalsTree(Object object);
	public abstract int treeHashCode();

	public abstract void findReadTemps(Set<String> tempNames);
	public abstract void findWrittenTemps(Set<String> tempNames);
	
	public abstract boolean containsMemWrites();
	public abstract boolean containsMemReads();

	public abstract boolean containsCalls();
	
	/// Count the total number of IR nodes in the subtree rooted here
	public abstract int nodeCount();

	private boolean setsOverlap(Set<String> s1, Set<String> s2) {
		HashSet<String> s = new HashSet<String>(s1);
		s.retainAll(s2);
		return s.size() > 0;
	}

	/* Max number of arguments of any call in the subtree */
	public abstract int computeMaximumCallArguments();

	/* Max number of results of any call in the subtree */
	public abstract int computeMaximumCallResults();

	/* Simplistic commutativity analysis */
	public boolean commutesWith(IRNode node) {
		Set<String> readTemps1 = new HashSet<String>();
		Set<String> writtenTemps1 = new HashSet<String>();
		Set<String> readTemps2 = new HashSet<String>();
		Set<String> writtenTemps2 = new HashSet<String>();
		boolean calls1 = containsCalls();
		boolean calls2 = node.containsCalls();
		boolean memWrites1 = containsMemWrites();
		boolean memWrites2 = node.containsMemWrites();
		boolean memReads1 = containsMemReads();
		boolean memReads2 = node.containsMemReads();
		boolean debugCommutativityTests = false;
		boolean result = true;

		findReadTemps(readTemps1);
		findWrittenTemps(writtenTemps1);
		node.findReadTemps(readTemps2);
		node.findWrittenTemps(writtenTemps2);

		if ((calls1 && calls2) || (calls1 && (memReads2 || memWrites2)) 
				|| (calls2 && (memReads1 || memWrites1)))
			result = false;

		if (setsOverlap(writtenTemps1, readTemps2)
				|| setsOverlap(readTemps1, writtenTemps2))
			result = false;
		
		if ((memWrites1 && memWrites2) || 
				(memWrites1 && memReads2) ||
				(memReads1 && memWrites2))
			result = false;
		
		if (debugCommutativityTests) {
			System.out.println("Checking for commutativity of \n");
			System.out.println(IRPrinter.IRToString(this));
			System.out.println(" and ");
			System.out.println(IRPrinter.IRToString(node));
			System.out.println("MEM reads   : " + memReads1 + "/" + memReads2);
			System.out.println("MEM writes  : " + memWrites1 + "/" + memWrites2);
			System.out.println("TEMP reads  : " + readTemps1 + "/" + readTemps2);
			System.out.println("TEMP writes : " + writtenTemps1 + "/" + writtenTemps2);
			System.out.println("CALLs       : " + calls1 + "/" + calls2);
			System.out.println(" ====> "
					+ (result ? "Nodes commute" : "Nodes do NOT commute"));
		}

		return result;
	}
	
	/**
	 * Create a deep copy (as opposed to copy(), which makes a shallow one)
	 */
	public abstract Copyable deepCopy();

	public String toString() {
		return IRPrinter.IRToString(this);
	}
}
