package xic.ir;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import xic.Configuration;
import xic.InternalCompilerException;
import xic.codegen.RegAlloc;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * Simple IR simulator for test cases
 */
public class IRSimulator {
	private IRCompUnit compUnit;
	private HashMap<String, Long> nameToIndex;
	private HashMap<Long, IRNode> indexToInstr;
	private HashMap<IRNode, Long> instrToIndex;
	private long mem[], heapPtr;
	private ExprStack exprStack;
	private int debugLevel = 0;
	private BufferedReader inReader;
	private List<String> libraryFunctions, ctors;

	/**
	 * While traversing the IR tree, we require a stack in order to hold
	 * a number of single-word values (e.g. to evaluate binary expressions).
	 * This also keeps track of whether a value was created by a TEMP
	 * or MEM, or NAME reference, which is useful when executing moves. 
	 */
	private class ExprStack {
		public class StackItem {
			public static final int COMPUTED = 0;
			public static final int MEM      = 1;
			public static final int TEMP     = 2;
			public static final int NAME     = 3;

			public int type;
			public long value;
			public long addr;
			public String temp;
			public String name;
			
			public StackItem(long value) {
				type = COMPUTED;
				this.value = value;
			}

			public StackItem(long value, long addr) {
				type = MEM;
				this.value = value;
				this.addr = addr;
			}

			public StackItem(int type, long value, String string) {
				this.type = type;
				this.value = value;
				if (type == TEMP)
					this.temp = string;
				else
					this.name = string;
			}
		};

		private Stack<StackItem> stack;
		
		public ExprStack() {
			stack = new Stack<StackItem>();
		}

		public long popValue() {
			if (debugLevel > 1)
				System.out.println("Popping value " + stack.peek().value);
			return stack.pop().value;
		}

		public StackItem pop() {
			return stack.pop();
		}

		public void pushAddr(long value, long addr) {
			if (debugLevel > 1)
				System.out.println("Pushing MEM " + value + " ("+addr+")");
			stack.push(new StackItem(value, addr));
		}

		public void pushTemp(long value, String temp) {
			if (debugLevel > 1)
				System.out.println("Pushing TEMP " + value + " ("+temp+")");
			stack.push(new StackItem(StackItem.TEMP, value, temp));
		}
		
		public void pushName(long value, String name) {
			if (debugLevel > 1)
				System.out.println("Pushing NAME " + value + " ("+name+")");
			stack.push(new StackItem(StackItem.NAME, value, name));
		}
		
		public void pushValue(long value) {
			if (debugLevel > 1)
				System.out.println("Pushing value " + value);
			stack.push(new StackItem(value));
		}
	}

	/**
	 * Holds temporaries and the instruction pointer
	 * within an execution frame.
	 */
	private class ExecutionFrame {
		public HashMap<String, Long> temps;
		public IRNode ip; // instruction pointer
		public boolean entering;

		public ExecutionFrame(IRNode ip) {
			entering = true;
			this.ip = ip;
			temps = new HashMap<String, Long>();
		}

		public void put(String tempName, long value) {
			temps.put(tempName, value);
		}

		public long get(String tempName) {
			if (!temps.containsKey(tempName)) {
				/* Referencing a temp before having written to it - initialize
				   with random contents */
				put(tempName, (int) (Math.random() * Integer.MAX_VALUE));
			}
			return temps.get(tempName);
		}

		/**
		 * Advance the instruction pointer. Since we're dealing with a tree,
		 * this is like a DFS traversal, one step at a time.
		 */
		public boolean advance() {
			/* This is insanely inefficient, but will do for test cases */
			if (entering) {
				if (debugLevel > 1)
					System.out.println("Entering " + ip.label());
				Iterator<VisualizableTreeNode> it = ip.children().iterator();
				if (it.hasNext())
					ip = (IRNode) it.next();
				else
					entering = false;
			} else {
				if (debugLevel > 1)
					System.out.println("Leaving " + ip.label());
				IRNode backupIP = ip;
				leave(this);

				if (ip != backupIP) /* A jump was performed */
					return true;

				IRNode parent = getParent(ip);
				if (parent instanceof IRCompUnit) {
					ip = null; 
					return false;
				}
				Iterator<VisualizableTreeNode> it = parent.children().iterator();
				while (it.next() != ip)
					;
				if (it.hasNext()) {
					ip = (IRNode) it.next();
					entering = true;
				} else {
					ip = parent;
					if (!advance())
						return false;
				}
			}
			return true;
		}
		
		public void setIP(IRNode node) {
			if (debugLevel > 1)
				System.out.println("Jumping to " + node.label());
			ip = node;
			entering = true;
		}
	};

	public static class Trap extends RuntimeException {
		private static final long serialVersionUID = 8429929900405296472L;

		public Trap(String message) {
			super(message);
		}
	};
	
	public IRSimulator(IRCompUnit compUnit) {
		this(compUnit, 10240);
	}

	public IRSimulator(IRCompUnit compUnit, int heapSize) {
		this.compUnit = compUnit;
		nameToIndex = new HashMap<String, Long>();
		indexToInstr = new HashMap<Long, IRNode>();
		instrToIndex = new HashMap<IRNode, Long>();
		mem = new long[heapSize];
		for(int i=0; i<heapSize;i++)
			mem[i] = (long) (Math.random() * Integer.MAX_VALUE);
		exprStack = new ExprStack();
		ctors = new ArrayList<String>();
		heapPtr = 0;
		inReader = new BufferedReader(new InputStreamReader(System.in));

		buildInstructionMaps(0, compUnit);
		libraryFunctions = new ArrayList<String>();
		libraryFunctions.add("_Iprint_pai");
		libraryFunctions.add("_Iprintln_pai");
		libraryFunctions.add("_Ireadln_ai");
		libraryFunctions.add("_Igetchar_i");
		libraryFunctions.add("_Ieof_b");
		libraryFunctions.add("_I_outOfBounds_p");
		libraryFunctions.add("_I_alloc_i");
		libraryFunctions.add("_IparseInt_t2ibai");
		libraryFunctions.add("_IunparseInt_aii");
		libraryFunctions.add("_Iassert_pb");

		for (int i=0; i<ctors.size(); ++i)
			call(ctors.get(i), new long[] { });
	}

	private long buildInstructionMaps(long index, IRNode node) {
		indexToInstr.put(index, node);
		if (instrToIndex.containsKey(node))
			throw new InternalCompilerException("Error - encountered "
					+ "duplicate node " +node.label()+ " in the IR tree -- go fix the generator.");
		instrToIndex.put(node, index);
		if (node instanceof IRLabel) {
			nameToIndex.put(((IRLabel) node).name(), index);
	 	} else if (node instanceof IRFuncDecl) {
			nameToIndex.put(((IRFuncDecl) node).name(), index);
		} else if (node instanceof IRDataSeq) {
			IRDataSeq seq = (IRDataSeq) node;
			int count = seq.values().size();
			long ptr = malloc(count * Configuration.WORD_SIZE);
			List<IRExpr> values = seq.values();
			for (int i=0; i<count; ++i) {
				if (values.get(i) instanceof IRConst)
					store(ptr + i*Configuration.WORD_SIZE, ((IRConst) values.get(i)).value());
				else if (values.get(i) instanceof IRName) {
					String name = ((IRName) values.get(i)).name();
					if (!nameToIndex.containsKey(name))
						throw new InternalCompilerException("Unknown name '" + name + "'");
					store(ptr + i*Configuration.WORD_SIZE, nameToIndex.get(name));
				}
			}
			nameToIndex.put(seq.name(), ptr);
		} else if (node instanceof IRZData) {
			IRZData seq = (IRZData) node;
			int count = seq.size();
			long ptr = malloc(count * Configuration.WORD_SIZE);
			for (int i=0; i<count; ++i)
				store(ptr + i*Configuration.WORD_SIZE, 0);
			nameToIndex.put(seq.name(), ptr);
		} else if (node instanceof IRCtor) {
			ctors.add(((IRCtor) node).name());
		}
		index++;
		for (VisualizableTreeNode c : node.children())
			index = buildInstructionMaps(index, (IRNode) c);
		return index;
	}

	private IRNode getParent(IRNode node, IRNode searchRoot) {
		for (VisualizableTreeNode child : searchRoot.children()) {
			if (child == node) {
				return searchRoot;
			} else {
				IRNode result = getParent(node, (IRNode) child);
				if (result != null)
					return result;
			}	
		}
		return null;
	}
	
	public long malloc(long size) {
		if (size <= 0)
			throw new Trap("Invalid memory allocation");
		long retval = heapPtr;
		if ((size%Configuration.WORD_SIZE) != 0)
			throw new InternalCompilerException("Can only allocate in chunks of " + Configuration.WORD_SIZE + " bytes!");
		heapPtr += size;
		
		for(long i=retval; i<heapPtr; i++)
			mem[(int) i] = 0;

		return retval;
	}
	
	public long read(long addr) {
		if ((addr%Configuration.WORD_SIZE) != 0)
			throw new InternalCompilerException("Unaligned memory access!");
		return mem[(int) (addr/Configuration.WORD_SIZE)];
	}

	public void store(long addr, long value) {
		if ((addr%Configuration.WORD_SIZE) != 0)
			throw new InternalCompilerException("Unaligned memory access!");
		mem[(int) (addr/Configuration.WORD_SIZE)] = value;
	}

	private IRNode getParent(IRNode node) {
		return getParent(node, compUnit);
	}

	private IRNode findLabel(String name) {
		if (!nameToIndex.containsKey(name))
			throw new InternalCompilerException("Could not find label '"+name+"'!");
		return indexToInstr.get(nameToIndex.get(name));
	}

	public long call(String name, long[] params) {
		return call(name, params, mem.length*Configuration.WORD_SIZE);
	}


	public long call(String name, long[] params, long sp) {
		/* Catch standard library calls */
		final int ws = Configuration.WORD_SIZE;
		try {
			if (name.equals("_I_alloc_i")) {
				return malloc(params[0]);
			} else if (name.equals("_Iprint_pai")) {
				long ptr = params[0], size = read(ptr-ws);
				for (long i=0; i<size; ++i)
					System.out.print((char) read(ptr + i*ws));
				return 0;
			} else if (name.equals("_Iprintln_pai")) {
				long ptr = params[0], size = read(ptr-ws);
				for (long i=0; i<size; ++i)
					System.out.print((char) read(ptr + i*ws));
				System.out.println();
				return 0;
			} else if (name.equals("_Ireadln_ai")) {
				String line = inReader.readLine();
				long ptr = malloc((line.length()+1)*ws);
				store(ptr, line.length());
				for (int i=0; i<line.length(); ++i)
					store(ptr+(i+1)*ws, line.charAt(i));
				return ptr+ws;
			} else if (name.equals("_IunparseInt_aii")) {
				String line = String.valueOf(params[0]);
				long ptr = malloc((line.length()+1)*ws);
				store(ptr, line.length());
				for (int i=0; i<line.length(); ++i)
					store(ptr+(i+1)*ws, line.charAt(i));
				return ptr+ws;
			} else if (name.equals("_IparseInt_t2ibai")) {
				StringBuffer buf = new StringBuffer();
				long ptr = params[1], size = read(ptr-ws);
				for (int i=0; i<size; ++i)
					buf.append((char) read(ptr + i*ws));
				long retPtr = params[0];
				int result = 0, success = 1;
				try {
					result = Integer.parseInt(buf.toString());
				} catch (NumberFormatException e) {
					success = 0;
				}
				store(retPtr,   result);
				store(retPtr+ws, success);
				return retPtr;
			} else if (name.equals("_Igetchar_i")) {
				return inReader.read();
			} else if (name.equals("_Ieof_b")) {
				return inReader.ready() ? 0 : 1;
			} else if (name.equals("_I_outOfBounds_p")) {
				throw new Trap("Out of bounds!");
			} else if (name.equals("_Iassert_pb")) {
				if (params[0] != 1)
					throw new Trap("Assertion error!");
				return 0;
			}
		} catch (IOException e) {
			throw new Trap("I/O Exception in simulator");
		}

		IRFuncDecl fDecl = compUnit.getFunction(name);
		if (fDecl == null)
			throw new InternalCompilerException(
					"Tried to call an unknown function: '" + name + "'");
		ExecutionFrame frame = new ExecutionFrame(fDecl);
		int parametersOnStack = 0;

		boolean hasCalls = fDecl.containsCalls();
		int rvCount = hasCalls ? fDecl.computeMaximumCallResults() : 0;
		if (rvCount == 1)
			rvCount = 0;

		for (int i=params.length-1; i>=Configuration.PARAMETER_REGISTERS.length; --i) {
			sp -= ws;
			mem[(int) sp/ws] = params[i];
			parametersOnStack++;
		}
		sp -= 2*ws;

		frame.put(RegAlloc.FP_NAME, sp);
		sp -= rvCount*ws;
		frame.put(RegAlloc.CALL_RV_SCRATCH, sp);		
		
		for(int i=0; i<Math.min(Configuration.PARAMETER_REGISTERS.length, params.length); ++i)
			frame.put(RegAlloc.ABSTRACT_REG_PREFIX + i, params[i]);
		
		while (frame.advance())
			;

		sp += ws*(2+parametersOnStack+rvCount);
		return frame.get(RegAlloc.RV_NAME);
	}

	private void leave(ExecutionFrame frame) {
		if (frame.ip instanceof IRExp) {
			exprStack.pop();
		} else if (frame.ip instanceof IRName) {
			String name = ((IRName) frame.ip).name();
			if (compUnit.getFunction(name) != null)
				exprStack.pushName(nameToIndex.get(name), name);
			else if (libraryFunctions.contains(name))
				exprStack.pushName(-1, name);
			else if (nameToIndex.containsKey(name))
				exprStack.pushName(nameToIndex.get(name), name);
			else
				throw new InternalCompilerException("Invalid destination in NAME: '"+name+"'");
		} else if (frame.ip instanceof IRJump) {
			frame.setIP(indexToInstr.get(exprStack.popValue()));
		} else if (frame.ip instanceof IRMem) {
			long addr = exprStack.popValue();
			if ((addr % Configuration.WORD_SIZE) != 0)
				throw new InternalCompilerException("Unaligned memory access: " 
						+ addr + " (word size=" + Configuration.WORD_SIZE + ")");
 			addr /= Configuration.WORD_SIZE;
			exprStack.pushAddr(mem[(int) addr], addr);
		} else if (frame.ip instanceof IRTemp) {
			String tempName = ((IRTemp) frame.ip).name();
			exprStack.pushTemp(frame.get(tempName), tempName);
		} else if (frame.ip instanceof IRCall) {
			int paramCount = ((IRCall) frame.ip).params().size();
			long params[] = new long[paramCount];
			for (int i=paramCount-1; i>=0; --i)
				params[i] = exprStack.popValue();
			ExprStack.StackItem target = exprStack.pop();
			String targetName = target.name;
			if (target.type != ExprStack.StackItem.NAME) {
				if (indexToInstr.containsKey(target.value)) {
					IRNode node = indexToInstr.get(target.value);
					if (node instanceof IRFuncDecl)
						targetName = ((IRFuncDecl) node).name();
					else
						throw new InternalCompilerException("Call to a non-function instruction!");
				} else {
					throw new InternalCompilerException(
							"Invalid function call '"+frame.ip+"' (target '" + target.value
									+ "' is unknown)!");
				}
			}

			long retVal = call(targetName, params, frame.get(RegAlloc.FP_NAME));
			exprStack.pushValue(retVal);
		} else if (frame.ip instanceof IRBinOp) {
			long r = exprStack.popValue();
			long l = exprStack.popValue();
			long result;
			switch (((IRBinOp) frame.ip).opType()) {
				case ADD:    result = l + r; break;
				case SUB:    result = l - r; break;
				case MUL:    result = l * r; break;
				case DIV:
					if (r == 0)
						throw new Trap("Division by zero!");
					result = l / r; 
				break;
				case MOD:
					if (r == 0)
						throw new Trap("Division by zero!");
					result = l % r;
				break;
				case LOGAND:
				case AND:    result = l & r; break;
				case LOGOR:
				case OR:     result = l | r; break;
				case XOR:    result = l ^ r; break;
				case LSHIFT: result = l << r; break;
				case RSHIFT: result = l >>> r; break;
				case AR_LSHIFT: result = l << r; break;
				case AR_RSHIFT: result = l >> r; break;
				case EQ:     result = (l == r) ? 1 : 0; break;
				case NEQ:    result = (l != r) ? 1 : 0; break;
				case LT:     result = (l < r)  ? 1 : 0; break;
				case GT:     result = (l > r)  ?  1 : 0; break;
				case LE:     result = (l <= r) ? 1 : 0; break;
				case GE:     result = (l >= r) ? 1 : 0; break;

				case ULT:
					result = ((l < r) ^ ((l < 0) != (r < 0))) ? 1 : 0;
					break;

				case UGT:
					result = ((l > r) ^ ((l < 0) != (r < 0))) ? 1 : 0;
					break;

				case ULE:
					result = ((l == r) || ((l < r) ^ ((l < 0) != (r < 0)))) ? 1 : 0;
					break;

				case UGE:
					result = ((l == r) || ((l > r) ^ ((l < 0) != (r < 0)))) ? 1 : 0;
					break;

				default:
					throw new InternalCompilerException("Invalid binary operation");
			}
			exprStack.pushValue(result);
		} else if (frame.ip instanceof IRConst) {
			exprStack.pushValue(((IRConst) frame.ip).value());
		} else if (frame.ip instanceof IRMove) {
			long r = exprStack.popValue();
			ExprStack.StackItem stackItem = exprStack.pop();
			if (stackItem.type == ExprStack.StackItem.MEM) {
				if (debugLevel > 0)
					System.out.println("mem["+stackItem.addr*Configuration.WORD_SIZE+"]=" +r);
				mem[(int) stackItem.addr] = r;
			} else if (stackItem.type == ExprStack.StackItem.TEMP) {
				if (debugLevel > 0)
					System.out.println("temp["+stackItem.temp+"]=" +r);
				frame.put(stackItem.temp, r);
			} else {
				throw new InternalCompilerException("Invalid MOVE!");
			}
		} else if (frame.ip instanceof IRCJump) {
			IRCJump irCJump = (IRCJump) frame.ip;
			long top = exprStack.popValue();
			String label;
			if (top == 0)
				label = irCJump.falseLabel();
			else if (top == 1)
				label = irCJump.trueLabel();
			else
				throw new InternalCompilerException("Invalid value in CJUMP - expected 0/1, got " + top);
			if (label != null)
				frame.setIP(findLabel(label));
		}
	}
}
