package xic.visitors;

import xic.ErrorSink;
import xic.ast.ASTNode;
import xic.ir.IRBinOp;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import edu.cornell.cs.cs4120.xi.CompilationException;

/* 
 * An IR Level constant folding optimizer.  
 * 
 * We fold over IR in the hopes of having more opportunities
 * to reduce constants after the IRGeneration.  Can 
 * potentially create dead code if condition jump branches
 * are replaced with jumps.  We rely on a separate optimization
 * pass for dead-code elimination.
 * 
 */
public class IRConstantFolder extends ErrorHandlingVisitor {
	public IRConstantFolder (ErrorSink errorSink) {
		super(errorSink);
	}
	
	protected NodeVisitor enterNode(ASTNode parent, ASTNode n) {
		return this;
	}

	protected ASTNode leaveNode(ASTNode parent, ASTNode n, ASTNode nNew,
			NodeVisitor vNew) {
		return leaveIRNode((IRNode) nNew);
	}
	
	protected IRNode leaveIRNode(IRNode node) {
		if (node instanceof IRBinOp)
			return FoldBinOp((IRBinOp) node);
		else if (node instanceof IRCJump)
			return FoldCJump((IRCJump) node);
		else if (node instanceof IRMove)
			return FoldMove((IRMove) node);
		else 
			return node;
	}

	private IRNode FoldCJump(IRCJump node) {
		if(node.expr() instanceof IRConst) {
			IRConst condition = (IRConst)node.expr();
			if(condition.value() == 0) 
				return new IRJump(new IRName(node.falseLabel()));
			else 
				return new IRJump(new IRName(node.trueLabel()));		
		}
		return node;
	}

	private IRNode FoldBinOp(IRBinOp node) {
		IRExpr left = node.left();
		IRExpr right = node.right();

		if (left instanceof IRConst &&
				right instanceof IRConst) {
			long leftVal = ((IRConst)left).value();
			long rightVal = ((IRConst)right).value();
			return new IRConst(IRBinOp.eval(node.opType(), leftVal, rightVal));
		} else if (left instanceof IRConst ||
				   right instanceof IRConst) {
			IRConst constant = (IRConst)((left instanceof IRConst) ? left : right);
			IRExpr dynamic = (constant == right) ? left : right;

			switch(node.opType()) {
			case ADD:
				if(constant.value() == 0)
					return dynamic;
				else 
					return node;					
			case AND:
				if(constant.value() == 0)
					return new IRConst(0);
				else 
					return node;
			case LOGAND:
				if(constant.value() == 0)
					return new IRConst(0);
				else 
					return dynamic;
			case DIV:
				if(constant == right) {
					if(constant.value() == 1)
						return dynamic;
					else if(constant.value() == 0)
						throw new CompilationException("Divide by zero",constant.position());
					else if (isPowerOfTwo(constant.value()))
						return new IRBinOp(IRBinOp.OpType.AR_RSHIFT, 
								dynamic, new IRConst(log2(constant.value())));
				}
				break;
			case MOD:
				if(constant == right)
					if(constant.value() == 1)
						return new IRConst(0);	
					else if(constant.value() == 0)
						throw new CompilationException("Modulo by zero",constant.position());
					else 
						return node;
				break;
			case MUL:
				if(constant.value() == 0) {
					return new IRConst(0);
				} else if(constant.value() == 1) {
						return dynamic;
				} else {
					if (isPowerOfTwo(constant.value()))
						return new IRBinOp(IRBinOp.OpType.AR_LSHIFT, 
								dynamic, new IRConst(log2(constant.value())));
					else if(dynamic instanceof IRBinOp){
						IRBinOp binop = (IRBinOp) dynamic;
						if(binop.opType() == IRBinOp.OpType.ADD) {
							return new IRBinOp(IRBinOp.OpType.ADD,
									new IRBinOp(IRBinOp.OpType.MUL, constant,
											binop.left()), 
									new IRBinOp(IRBinOp.OpType.MUL, constant, 
											binop.right()));
						}
						else return node;
					}
					else
						return node;
				}
			case LOGOR:
				if(constant.value() == 1)
					return new IRConst(1);
				else 
					return dynamic;
			case SUB:
				if(constant == right)
					if(constant.value() == 0)
						return dynamic;
					else 
						return node;
				break;
			case XOR:
				if(constant.value() == 0)
					return dynamic;
				else
					return node;
			default:
				return node;
			}
		}
		if (node.opType() == IRBinOp.OpType.ADD
				&& ((left instanceof IRTemp && right instanceof IRBinOp)
				|| (right instanceof IRTemp && left instanceof IRBinOp))) {
			IRTemp temp = (left instanceof IRTemp) ? (IRTemp)left : (IRTemp)right;
			IRBinOp irBinOp = (temp == left) ? (IRBinOp) right : (IRBinOp) left;
			if(irBinOp.opType() == IRBinOp.OpType.MUL) {
				if(irBinOp.left() instanceof IRTemp || irBinOp.right() instanceof IRTemp) {
					IRTemp tempPrime = (irBinOp.left() instanceof IRTemp)?(IRTemp)irBinOp.left():(IRTemp)irBinOp.right();
					IRExpr subExpr = (tempPrime == irBinOp.left()) ? irBinOp.right():irBinOp.left();				
					if(tempPrime.name().equals(temp.name())) {
						return FoldBinOp(new IRBinOp(IRBinOp.OpType.MUL, temp,
								(IRExpr)leaveIRNode(new IRBinOp(IRBinOp.OpType.ADD,
										new IRConst(1), subExpr))));
					}
				}
			}
		}
		return node;
	}

	
	private int log2(long value) {
		int r = 0;
		while ((value >> r) != 0) {
			r++;
		}
		return r-1;
	}

	private boolean isPowerOfTwo(long value) {
		return (value & (value-1)) == 0;
	}

	public IRStmt FoldMove(IRMove move) {
		IRExpr src = move.expr(), target = move.target();

		if (target instanceof IRTemp && src instanceof IRTemp) {
			/* Optimize away unnecessary MOVEs */
			if (((IRTemp) target).name().equals(((IRTemp) src).name()))
				return new IRSeq();
		}

		return move;
	}
}

