package xic.visitors;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.TreeMap;

import xic.Configuration;
import xic.InternalCompilerException;
import xic.ast.ArrayType;
import xic.ast.Assignment;
import xic.ast.BinaryExpr;
import xic.ast.Block;
import xic.ast.BoolLiteral;
import xic.ast.Break;
import xic.ast.Call;
import xic.ast.CharLiteral;
import xic.ast.ClassDecl;
import xic.ast.ClassType;
import xic.ast.CompUnit;
import xic.ast.DataSource;
import xic.ast.Decl;
import xic.ast.DotExpr;
import xic.ast.Expr;
import xic.ast.ExprStmt;
import xic.ast.FuncDecl;
import xic.ast.FuncType;
import xic.ast.Identifier;
import xic.ast.If;
import xic.ast.IntLiteral;
import xic.ast.Length;
import xic.ast.Literal;
import xic.ast.LocalVarDecl;
import xic.ast.New;
import xic.ast.Null;
import xic.ast.NullType;
import xic.ast.PrimitiveType;
import xic.ast.Return;
import xic.ast.Stmt;
import xic.ast.StringLiteral;
import xic.ast.This;
import xic.ast.TupleDecl;
import xic.ast.TupleExpr;
import xic.ast.TupleType;
import xic.ast.TypeNode;
import xic.ast.UnaryExpr;
import xic.ast.UnitType;
import xic.ast.VarDecl;
import xic.ast.While;
import xic.codegen.GraphColoringRegAlloc;
import xic.codegen.RegAlloc;
import xic.ir.IRBinOp;
import xic.ir.IRBinOp.OpType;
import xic.ir.IRMem.MemType;
import xic.ir.IRCJump;
import xic.ir.IRCall;
import xic.ir.IRCompUnit;
import xic.ir.IRConst;
import xic.ir.IRCtor;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRESeq;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRFuncDecl;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.IRZData;
import xic.ir.NameGenerator;
import xic.ir.NameMangler;
import xic.parser.ParserImpl;
import xic.parser.ParserImpl.Language;
import edu.cornell.cs.cs4120.xi.CompilationException;

public class IRGenerator {
	public static final String MALLOC_NAME        = "_I_alloc_i";
	public static final String CALL_ABORT         = "call_abort";
	public static final String CALL_STRDUP        = "strdup";
	public static final String BOUNDS_ERRFN       = "_I_outOfBounds_p";
	private static final int MAX_PEEL_NODES       = 200;

	private NameGenerator nameGen;
	private Map<String, FuncDecl> functions;
	private Map<String, String> stringTable;
	private String currentFunction;
	private String epilogueName;
	private boolean needsAbort;
	private boolean peelLoops;
	private Stack<Boolean> inPeeledLoop;
	private Stack<String> escapeLabelStack;
	private int nPeeled = 0;
	private ClassType currentClass;
	private ParserImpl.Language language;

	public IRGenerator(NameGenerator nameGen, boolean peelLoops, ParserImpl.Language language) {
		this.nameGen = nameGen;
		this.peelLoops = peelLoops;
		this.escapeLabelStack = new Stack<String>();
		this.inPeeledLoop = new Stack<Boolean>();
		this.language = language;
	}

	public IRCompUnit translate(CompUnit compUnit, List<Decl> imports) {
		IRCompUnit irCompUnit = new IRCompUnit(compUnit.name());

		/* Create a dictionary of all functions */
		functions = new TreeMap<String, FuncDecl>();
		needsAbort = false;
		stringTable = new TreeMap<String, String>();
		inPeeledLoop.clear();
		inPeeledLoop.add(false);

		/* Build a dictionary of all known functions -- this is needed in several places */
		for (Decl decl: compUnit.globalStmtBlock().decls()) {
			if (decl instanceof ClassDecl) {
				ClassDecl cd = (ClassDecl) decl;
				for (Decl decl2: cd.declBlock().decls()) {
					if (decl2 instanceof FuncDecl)
						functions.put(cd.name() + "." + ((FuncDecl) decl2).name(), (FuncDecl) decl2);
				}
			} else if (decl instanceof FuncDecl) {
				functions.put(((FuncDecl) decl).name(), (FuncDecl) decl);
			}
		}
		for (Decl decl : imports) {
			if (decl instanceof ClassDecl) {
				ClassDecl cd = (ClassDecl) decl;
				for (Decl decl2: cd.declBlock().decls()) {
					if (decl2 instanceof FuncDecl)
						functions.put(cd.name() + "." + ((FuncDecl) decl2).name(), (FuncDecl) decl2);
				}
			} else if (decl instanceof FuncDecl) {
				functions.put(((FuncDecl) decl).name(), (FuncDecl) decl);
			}
		}

		/* Translate the functions and classes one by one */
		for (Decl decl: compUnit.globalStmtBlock().decls()) {
			if (decl instanceof FuncDecl) {
				irCompUnit.appendFunc(translateFunction(null, (FuncDecl) decl));
			} else if (decl instanceof ClassDecl) {
				translateClass(irCompUnit, (ClassDecl) decl);
			}
		}

		if (stringTable.size() > 0) {
			for (String str : stringTable.keySet()) {
				IRDataSeq dSeq = new IRDataSeq(stringTable.get(str), true);
				dSeq.appendConst(new IRConst(str.length()));
				for (int i = 0; i < str.length(); ++i)
					dSeq.appendConst(new IRConst(str.charAt(i)));
				irCompUnit.appendStmt(dSeq);
			}

			/* Generate an internal function for duplicating strings. It expects
			   a pointer to the size (-1th) element of the array */
			IRSeq seq = new IRSeq();
			String arrayTemp = "p0", 
				   lengthTemp = nameGen.nextTemp(),
				   mallocTemp = nameGen.nextTemp();
			String loopLabel = nameGen.nextLabel();
			IRExpr source;

			if (Configuration.PARAMETER_REGISTERS.length > 0) {
				String name = GraphColoringRegAlloc.ABSTRACT_REG_PREFIX + "0";
				seq.appendStmt(new IRDecl(name));
				source = new IRTemp(name);
			} else {
				source = new IRMem(new IRBinOp(IRBinOp.OpType.ADD,
						new IRTemp(RegAlloc.FP_NAME), new IRConst(2*Configuration.WORD_SIZE)),
						MemType.IMMUTABLE);
			}
			seq.appendStmt(new IRMove(new IRTemp(arrayTemp), source));
			
			seq.appendStmt(new IRMove(
					new IRTemp(lengthTemp),
					new IRMem(new IRTemp(arrayTemp))
			));
			IRCall mallocCall = new IRCall(new IRName(MALLOC_NAME), 1);
			mallocCall.appendParam(new IRBinOp(IRBinOp.OpType.ADD,
					new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(lengthTemp),
					new IRConst(Configuration.WORD_SIZE)),
					new IRConst(Configuration.WORD_SIZE)
			));
			seq.appendStmt(new IRMove(new IRTemp(mallocTemp), mallocCall));
			seq.appendStmt(new IRLabel(loopLabel));
			seq.appendStmt(new IRMove(
				new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(mallocTemp), 
						new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(lengthTemp), new IRConst(Configuration.WORD_SIZE)))),
				new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(arrayTemp), 
						new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(lengthTemp), new IRConst(Configuration.WORD_SIZE))))
			));

			seq.appendStmt(new IRMove(new IRTemp(lengthTemp), new IRBinOp(IRBinOp.OpType.SUB, 
					new IRTemp(lengthTemp), new IRConst(1))));
			IRLabel endLabel = new IRLabel(nameGen.nextLabel());
			seq.appendStmt(new IRCJump(
					new IRBinOp(OpType.GE, new IRTemp(lengthTemp), new IRConst(0)),
					loopLabel, endLabel.name()));
			seq.appendStmt(endLabel);
			seq.appendStmt(new IRMove(new IRTemp(RegAlloc.RV_NAME), new IRBinOp(IRBinOp.OpType.ADD, 
					new IRTemp(mallocTemp), new IRConst(Configuration.WORD_SIZE))));
			irCompUnit.appendFunc(new IRFuncDecl(CALL_STRDUP, seq, false));
		}

		if (needsAbort){
			IRSeq seq = new IRSeq();
			seq.appendStmt(new IRLabel(CALL_ABORT));
			seq.appendStmt(new IRExp(new IRCall(new IRName(BOUNDS_ERRFN), 0)));
			irCompUnit.appendStmt(seq);
		}
		
		for (Decl decl : compUnit.globalStmtBlock().decls()) {
			if (!(decl instanceof VarDecl))
				continue;
			VarDecl vd = (VarDecl) decl;
			for (TupleDecl tuple : vd.tupleDecl().subTupleDecls()) {
				String mangled = NameMangler.translateGlobalVarName(tuple.name(), tuple.getType());
				String mangledInit = NameMangler.translateGlobalVarInit(tuple.name(), tuple.getType());
				
				if (vd.hasInit()) {
					IRDataSeq data = new IRDataSeq(mangled, false);
					data.appendConst((IRConst) E(vd.init()));
					irCompUnit.appendStmt(data);
				} else {
					IRZData data = new IRZData(mangled, tuple.getType().size()
							* Configuration.WORD_SIZE);
					irCompUnit.appendStmt(data);
				}
				
				IRSeq seq = new IRSeq();
				List<String> lengthTemps = new ArrayList<String>();
				collectLengthExprs(tuple.getType(), lengthTemps, seq);
				initializeTemps(tuple.getType(), tuple.getDataSource(), lengthTemps.iterator(), seq);
				if (seq.stmts().size() > 0) {
					/* This global variable needs an initialization routine */
					irCompUnit.appendFunc(new IRFuncDecl(mangledInit, seq, true));
					irCompUnit.appendStmt(new IRCtor(mangledInit));
				}
			}
		}

		return irCompUnit;
	}

	private void translateClass(IRCompUnit irCompUnit, ClassDecl cd) {
		Map<String, FuncDecl> localFuncs = new HashMap<String, FuncDecl>();

		ClassType ct = (ClassType) cd.getType();
		currentClass = ct;
		for (Decl decl : cd.declBlock().decls()) {
			if (decl instanceof FuncDecl) {
				localFuncs.put(((FuncDecl) decl).name(), (FuncDecl) decl);
				irCompUnit.appendFunc(translateFunction(cd, (FuncDecl) decl));
			}
		}
		currentClass = null;

		String initName = ct.getStaticInitializationIdentifier();
		String sizeName = ct.getClassSizeIdentifier();
		String vtName = ct.getDispatchTableIdentifier();

		IRSeq seq = new IRSeq();
		IRFuncDecl func = new IRFuncDecl(initName, seq,true);
		Map<Integer, String> revDt = new HashMap<Integer, String>();
		int maxIndex = 0, objectSize = ct.getSizeSelf();

		/* Reverse the dispatch table */
		for (String name : ct.getDispatchTable().keySet()) {
			int dtIndex = ct.getDispatchTable().get(name);
			revDt.put(dtIndex, name);
			if (dtIndex > maxIndex)
				maxIndex = dtIndex;
		}

		if (!cd.isDerived()) {
			/* Create everything in a read-only part of the data segment */
			IRDataSeq sizeDseq = new IRDataSeq(sizeName, true);
			IRDataSeq vtDseq = new IRDataSeq(vtName, true);
			sizeDseq.appendConst(new IRConst((objectSize+1)*Configuration.WORD_SIZE));
			vtDseq.appendConst(new IRConst(0)); // reserved
			for (int i=1; i<=revDt.size(); ++i) {
				String name = revDt.get(i);
				name = NameMangler.getEncoding(localFuncs.get(name), cd);
				vtDseq.appendName(new IRName(name));
			}
			irCompUnit.appendStmt(sizeDseq);
			irCompUnit.appendStmt(vtDseq);
		} else {
			String parentInitName = ct.parent().getStaticInitializationIdentifier();
			String parentSizeName = ct.parent().getClassSizeIdentifier();
			String parentVtName = ct.parent().getDispatchTableIdentifier();

			irCompUnit.appendStmt(new IRZData(sizeName, Configuration.WORD_SIZE));
			irCompUnit.appendStmt(new IRZData(vtName, (revDt.size()+1+ct.inheritanceDepth()) * Configuration.WORD_SIZE));
			irCompUnit.appendStmt(new IRCtor(initName));
			/* Generate code for copying the parent's dispatch table */
			IRLabel endLabel = new IRLabel(nameGen.nextLabel());

			seq.appendStmt(new IRCJump(
				new IRBinOp(IRBinOp.OpType.EQ, new IRMem(new IRName(sizeName)), new IRConst(0)),
						null, endLabel.name()));
			seq.appendStmt(new IRExp(new IRCall(new IRName(parentInitName), 0)));
			seq.appendStmt(new IRMove(new IRMem(new IRName(sizeName)),
					new IRBinOp(IRBinOp.OpType.ADD, new IRMem(new IRName(parentSizeName)),
							new IRConst(objectSize*Configuration.WORD_SIZE))));

			String counter = nameGen.nextTemp(), startLabel = nameGen.nextLabel(), doneLabel = nameGen.nextLabel();
			seq.appendStmt(new IRMove(new IRTemp(counter), new IRConst(0)));
			seq.appendStmt(new IRLabel(startLabel));
			seq.appendStmt(new IRCJump(new IRBinOp(IRBinOp.OpType.LT,
					new IRTemp(counter),
					new IRConst(ct.parent().getDispatchTable().size()+1+ct.parent().inheritanceDepth())),
					null, doneLabel));

			seq.appendStmt(new IRMove(
					new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRName(vtName),
							new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(counter), new IRConst(Configuration.WORD_SIZE)))),
					new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRName(parentVtName),
							new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(counter), new IRConst(Configuration.WORD_SIZE))))
			));
			seq.appendStmt(new IRMove(new IRTemp(counter),
					new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(counter), new IRConst(1))
			));

			seq.appendStmt(new IRJump(new IRName(startLabel)));
			seq.appendStmt(new IRLabel(doneLabel));

			for (int i=0; i<=maxIndex; ++i) {
				String name = revDt.get(i);
				if (localFuncs.containsKey(name)) {
					name = NameMangler.getEncoding(localFuncs.get(name), cd);
					seq.appendStmt(new IRMove(new IRMem(
							new IRBinOp(IRBinOp.OpType.ADD, new IRName(vtName),new IRConst(i*Configuration.WORD_SIZE))),
							new IRName(name)));
				} else if (i == ct.parent().getDispatchTable().size()+1+ct.parent().inheritanceDepth()) {
					seq.appendStmt(new IRMove(new IRMem(
							new IRBinOp(IRBinOp.OpType.ADD, new IRName(vtName),new IRConst(i*Configuration.WORD_SIZE))),
							new IRConst(0)));
				}
			}
			
			seq.appendStmt(endLabel);
		}
		
		irCompUnit.appendFunc(func);
	}

	private IRFuncDecl translateFunction(ClassDecl cd, FuncDecl fDecl) {
		IRSeq body = new IRSeq();
		IRFuncDecl irfDecl = new IRFuncDecl(NameMangler.getEncoding(fDecl, cd),
												body,fDecl.isProcedure());
		boolean isMethod = (cd != null);
		int rvSize = fDecl.retType() != null ? fDecl.retType().size() : 0;

		if (isMethod)
			currentFunction = cd.name() + "." + fDecl.name();
		else
			currentFunction = fDecl.name();

		DataSource ds = fDecl.param().getDataSource();

		List<String> targetNames = new ArrayList<String>();
	
		if (rvSize > 1)
			targetNames.add(RegAlloc.RV_NAME);
		if (isMethod)
			targetNames.add(RegAlloc.THIS_NAME);

		
		for (int i=0; i<ds.size(); ++i)
			targetNames.add(ds.get(i).name());
		
		int regArgs = Math.min(targetNames.size(), Configuration.PARAMETER_REGISTERS.length);
		
		for (int i=0; i<regArgs; ++i)
			body.appendStmt(new IRDecl(RegAlloc.ABSTRACT_REG_PREFIX + i));

		for (int i=0; i<targetNames.size(); ++i) {
			IRExpr source;

			if (i < Configuration.PARAMETER_REGISTERS.length) {
				source = new IRTemp(RegAlloc.ABSTRACT_REG_PREFIX + i);
			} else {
				int fpOffset = (i-Configuration.PARAMETER_REGISTERS.length+2)
						* Configuration.WORD_SIZE + Configuration.SHADOW_ZONE_SIZE;
				source = new IRMem(new IRBinOp(IRBinOp.OpType.ADD,
						new IRTemp(RegAlloc.FP_NAME), new IRConst(fpOffset)), MemType.IMMUTABLE);
			}

			body.appendStmt(new IRMove(new IRTemp(targetNames.get(i)), source));
		}

		epilogueName = null;
		
		if (language == Language.IOTA_09) {
			/* Initialize entries of the return value */
			if (rvSize == 1) {
				if (language == Language.IOTA_09)
					body.appendStmt(new IRMove(new IRTemp(RegAlloc.RV_NAME), new IRConst(0)));
			} else if (rvSize > 1) {
				for (int i=0; i<rvSize; ++i) {
					body.appendStmt(new IRMove(new IRMem(new IRBinOp(
							IRBinOp.OpType.ADD, new IRTemp(RegAlloc.RV_NAME), new IRConst(i
								* Configuration.WORD_SIZE))), new IRConst(0)));
				}
			}
		}
			
		if (cd != null) {
			if (cd.isDerived())
				body.appendStmt(new IRMove(new IRTemp(RegAlloc.OFFSET_NAME), new IRMem(
					new IRName(((ClassType) cd.getType()).parent().getClassSizeIdentifier()))));
			else
				body.appendStmt(new IRMove(new IRTemp(RegAlloc.OFFSET_NAME), new IRConst(0)));
		}

		body.appendStmt((IRSeq) S(fDecl.body()));

		if (epilogueName != null)
			body.appendStmt(new IRLabel(epilogueName));
		
		return irfDecl;
	}

	public int getNumPeeled() {
		return nPeeled;
	}

	/**
	 * For convenience: similar to TypeNode.size(), but counts the unit 
	 * type as zero words instead of throwing an exception.
	 */
	private int typeSize(TypeNode node) {
		if (node instanceof TupleType) {
			int result = 0;
			for (TypeNode t : ((TupleType) node).elementTypes())
				result += typeSize(t);
			return result;
		} else if (node instanceof UnitType) {
			return 0;
		} else {
			return node.size();
		}
	}

	/**
	 * Generates IR to access a parameter/temporary as specified by a DataSource entry
	 */
	private IRExpr fromDataSource(DataSource ds, int index) {
		DataSource.Entry entry = ds.get(index);
		if (entry.isTemp() || entry.isParam()) {
			return new IRTemp(ds.get(index).name());
		} else if (entry.isAttribute()) {
			return new IRMem(new IRBinOp(OpType.ADD,
					new IRBinOp(OpType.ADD, 
							new IRTemp(RegAlloc.THIS_NAME), 
							new IRTemp(RegAlloc.OFFSET_NAME)),
							new IRConst(entry.offset()*Configuration.WORD_SIZE)));
		} else if (entry.isGlobal()) {
			if (entry.offset() == 0)
				return new IRMem(new IRName(entry.name()));
			else
				return new IRMem(new IRBinOp(IRBinOp.OpType.ADD, 
						new IRName(entry.name()),
						new IRConst(entry.offset() * Configuration.WORD_SIZE)));
		} else {
			throw new InternalCompilerException("Unknown data source type!");
		}
	}

	private IRExpr fromDataSource(DataSource ds) {
		if (ds.size() != 1)
			throw new InternalCompilerException("Size mismatch in fromDataSource() = " + ds.size());
		return fromDataSource(ds, 0);
	}

	// ***************************************************************
	// ***               Expression translation rules              ***
	// ***************************************************************

	public IRExpr E(Expr expr) {
		if (expr instanceof Literal)
			return ELiteral((Literal) expr);
		else if (expr instanceof BinaryExpr)
			return EBinaryExpr((BinaryExpr) expr);
		else if (expr instanceof UnaryExpr)
			return EUnaryExpr((UnaryExpr) expr);
		else if (expr instanceof Identifier)
			return EIdentifier((Identifier) expr);
		else if (expr instanceof Length)
			return ELength((Length) expr);
		else if (expr instanceof Call)
			return ECall((Call) expr);
		else if (expr instanceof TupleExpr && ((TupleExpr) expr).exprCount() == 1)
			return E(((TupleExpr) expr).subExprs().get(0));
		else if (expr instanceof This)
			return EThis((This) expr);
		else if (expr instanceof DotExpr)
			return EDotExpr((DotExpr) expr);
		else if (expr instanceof Null)
			return new IRConst(0);
		else if (expr instanceof New)
			return ENew((New) expr);
		else
			throw new InternalCompilerException("Encountered an unknown expression node - E" 
					+ expr.getClass().getSimpleName() + "() still needs to be implemented!");
	}

	public String ETemp(Expr expr, IRSeq seq) {
		IRExpr irExpr = E(expr);
		if (irExpr instanceof IRTemp)
			return ((IRTemp) irExpr).name();

		String tempName = nameGen.nextTemp();
		seq.appendStmt(new IRMove(new IRTemp(tempName), irExpr));
		return tempName;
	}


	private IRExpr EThis(This expr) {
		return new IRTemp(RegAlloc.THIS_NAME);
	}

	private IRExpr EDotExpr(DotExpr expr) {
		/* Function calls are handled in ECall - this is because
		   we need access to the 'this' pointer */

		ClassType targetClass = (ClassType) expr.target().getType();
		String member = expr.member();

		IRExpr offset = targetClass.isDerived() ? new IRMem(new IRName(
				targetClass.parent().getClassSizeIdentifier()), IRMem.MemType.IMMUTABLE)
				: new IRConst(0);

		if (targetClass == currentClass)
			offset = new IRTemp(RegAlloc.OFFSET_NAME);

		List<Integer> fieldOffsets = targetClass.getFieldTable().get(member);

		if (fieldOffsets.size() == 1)
			return new IRMem(new IRBinOp(OpType.ADD, new IRBinOp(OpType.ADD,
					E(expr.target()), offset), new IRConst(fieldOffsets.get(0)
					* Configuration.WORD_SIZE)));
		else
			throw new InternalCompilerException(
					"Encountered an unexpected dot expression with size >1");
	}

	private IRExpr ELiteral(Literal expr) {
		if (expr instanceof IntLiteral) {
			return new IRConst(((IntLiteral) expr).value());
		} else if (expr instanceof CharLiteral) {
			return new IRConst(((CharLiteral) expr).value());
		} else if (expr instanceof BoolLiteral) {
			return new IRConst(((BoolLiteral) expr).value() ? 1 : 0);
		} else if (expr instanceof StringLiteral) {
			String stringValue = ((StringLiteral) expr).value();
			String dataSegmentLabel = null;
			if (stringTable.containsKey(stringValue)) {
				dataSegmentLabel = stringTable.get(stringValue);
			} else {
				dataSegmentLabel = nameGen.nextLabel();
				stringTable.put(stringValue, dataSegmentLabel);
			}

			IRCall call = new IRCall(new IRName(CALL_STRDUP), 1);
			call.appendParam(new IRName(dataSegmentLabel));
			return call;
		} else {
			throw new InternalCompilerException("Encountered an unsupported literal node");
		}
	}

	private IRExpr EIdentifier(Identifier expr) {
		if (expr.getDataSource() == null)
			throw new InternalCompilerException("Identifier '" + expr.name() + "' does not have a data source!", expr.position());
		return fromDataSource(expr.getDataSource());
	}

	private IRExpr ELength(Length expr) {
		TypeNode type = expr.expr().getType();
		if (type instanceof ArrayType) {
			return new IRMem(new IRBinOp(IRBinOp.OpType.SUB, E(expr.expr()),
					new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE);
		} else if (type instanceof TupleType) {
			return new IRConst(((TupleType) type).elementTypes().size());
		} else {
			return new IRConst(1);
		}
	}
	
	private IRExpr EUnaryExpr(UnaryExpr expr) {
		switch(expr.opType()) {
		case MINUS:
			return new IRBinOp(OpType.SUB,
					new IRConst(0),E(expr.expr()));
		case NOT:
			return new IRBinOp(OpType.XOR,
					new IRConst(1),E(expr.expr()));
		default:
			throw new InternalCompilerException("Unknown unary optype");
		}
	}

	private IRExpr EBinaryExpr(BinaryExpr expr) {
		IRBinOp.OpType opType;
		
		if (expr.opType() == BinaryExpr.OpType.PLUS && expr.getType() instanceof ArrayType) {
			/* Array addition */
			Expr left = expr.left();
			Expr right = expr.right();
			DataSource leftDS = new DataSource(), rightDS = new DataSource();
			IRSeq seq = new IRSeq();

			if (!(left.getType() instanceof ArrayType)) {
				leftDS = TConv(left, expr.getType(), seq);
			} else {
				/* save array reference to a temp */
				leftDS.addTempName(ETemp(left, seq));
			}

			if (!(right.getType() instanceof ArrayType)) {
				rightDS = TConv(right, expr.getType(), seq);
			} else {
				/* save array reference to a temp */
				rightDS.addTempName(ETemp(right, seq));
			}
			int size = ((ArrayType) expr.getType()).elementType().size();
			
			String lengthLeft = nameGen.nextTemp(), lengthRight = nameGen.nextTemp(),
				combinedLength = nameGen.nextTemp(), tempName = nameGen.nextTemp();
			seq.appendStmt(new IRMove(new IRTemp(lengthLeft),
					new IRMem(new IRBinOp(OpType.SUB,
						fromDataSource(leftDS),
						new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE)));

			seq.appendStmt(new IRMove(new IRTemp(lengthRight),
					new IRMem(new IRBinOp(OpType.SUB,
						fromDataSource(rightDS),
						new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE)));
			
			seq.appendStmt(new IRMove(new IRTemp(combinedLength),
					new IRBinOp(OpType.ADD, new IRTemp(lengthLeft), new IRTemp(lengthRight))));

			IRCall call = new IRCall(new IRName(MALLOC_NAME), 1);
			call.appendParam(new IRBinOp(OpType.ADD,
					new IRConst(Configuration.WORD_SIZE),
					new IRBinOp(OpType.MUL,
					new IRTemp(combinedLength),
					new IRConst(size * Configuration.WORD_SIZE))));
	
			seq.appendStmt(new IRMove(new IRTemp(tempName), call));
			seq.appendStmt(new IRMove(new IRMem(new IRTemp(tempName)), 
					new IRTemp(combinedLength)));
			seq.appendStmt(new IRMove(new IRTemp(tempName), new IRBinOp(OpType.ADD, 
					new IRTemp(tempName), new IRConst(Configuration.WORD_SIZE))));
	
			String copyLeftLoop = nameGen.nextLabel(),
				copyLeftDone = nameGen.nextLabel(),
				copyRightLoop = nameGen.nextLabel(),
				copyRightDone = nameGen.nextLabel(),
				ctr = nameGen.nextTemp();
			
			seq.appendStmt(new IRMove(new IRTemp(ctr), new IRConst(0)));
			seq.appendStmt(new IRLabel(copyLeftLoop));
			seq.appendStmt(new IRCJump(
					 new IRBinOp(OpType.LT, new IRTemp(ctr), 
							 new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(lengthLeft), new IRConst(size))),
					 null, copyLeftDone));
			seq.appendStmt(new IRMove(
					new IRMem(new IRBinOp(OpType.ADD, new IRTemp(tempName),
						new IRBinOp(OpType.AR_LSHIFT, new IRTemp(ctr), new IRConst(Configuration.LOG_WORD_SIZE)))),
					new IRMem(new IRBinOp(OpType.ADD, fromDataSource(leftDS),
							new IRBinOp(OpType.AR_LSHIFT, new IRTemp(ctr), new IRConst(Configuration.LOG_WORD_SIZE))))));
			seq.appendStmt(new IRMove(new IRTemp(ctr), new IRBinOp(OpType.ADD, 
					new IRTemp(ctr), new IRConst(1))));
			seq.appendStmt(new IRJump(new IRName(copyLeftLoop)));
			seq.appendStmt(new IRLabel(copyLeftDone));
			
			seq.appendStmt(new IRMove(new IRTemp(ctr), new IRConst(0)));
			seq.appendStmt(new IRLabel(copyRightLoop));
			seq.appendStmt(new IRCJump(
					 new IRBinOp(OpType.LT, new IRTemp(ctr), 
							 new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(lengthRight), new IRConst(size))),
					 null, copyRightDone));
			seq.appendStmt(new IRMove(
					new IRMem(new IRBinOp(OpType.ADD, new IRTemp(tempName),
						new IRBinOp(OpType.AR_LSHIFT, new IRBinOp(OpType.ADD, new IRTemp(ctr), new IRTemp(lengthLeft)), new IRConst(Configuration.LOG_WORD_SIZE)))),
					new IRMem(new IRBinOp(OpType.ADD, fromDataSource(rightDS),
							new IRBinOp(OpType.AR_LSHIFT, new IRTemp(ctr), new IRConst(Configuration.LOG_WORD_SIZE))))));
			seq.appendStmt(new IRMove(new IRTemp(ctr), new IRBinOp(OpType.ADD, 
					new IRTemp(ctr), new IRConst(1))));
			seq.appendStmt(new IRJump(new IRName(copyRightLoop)));
			seq.appendStmt(new IRLabel(copyRightDone));

			return new IRESeq(seq, new IRTemp(tempName));
		}
			

		if ((expr.opType() == BinaryExpr.OpType.OR || expr.opType() == BinaryExpr.OpType.AND)
				&& expr.left().getType().isIdenticalType(PrimitiveType.BOOLEAN_TYPE)) {
			String tempName = nameGen.nextTemp(), endLabel = nameGen.nextLabel();
			String tLabel = nameGen.nextLabel(), fLabel = nameGen.nextLabel();
			IRSeq seq = new IRSeq();
			seq.appendStmt(C(expr, tLabel, fLabel));
			seq.appendStmt(new IRLabel(tLabel));
			seq.appendStmt(new IRMove(new IRTemp(tempName), new IRConst(1)));
			seq.appendStmt(new IRJump(new IRName(endLabel)));
			seq.appendStmt(new IRLabel(fLabel));
			seq.appendStmt(new IRMove(new IRTemp(tempName), new IRConst(0)));
			seq.appendStmt(new IRLabel(endLabel));
			return new IRESeq(seq, new IRTemp(tempName));
		}

		switch (expr.opType()) {
			case EQUAL:
			case NOT_EQUAL:
				boolean eq = expr.opType() == BinaryExpr.OpType.EQUAL;
				opType = expr.opType() == BinaryExpr.OpType.EQUAL ? IRBinOp.OpType.EQ
						: IRBinOp.OpType.NEQ;
				TypeNode lType = expr.left().getType(), rType = expr.right().getType();

				/* Turning a tuple into an array will always result in a negative 
				   comparison, so don't bother.. */
				if (lType.isSubtypeOf(rType) && needsTypeConversion(lType, rType))
					return new IRConst(eq ? 0 : 1);
				if (rType.isSubtypeOf(lType) && needsTypeConversion(rType, lType))
					return new IRConst(eq ? 0 : 1);

				if (lType.size() > 1 && rType.size() > 1) {
					/* Now that there are no array conversions, type checking guarantees 
					   that the tuple sizes match */
					String tempName = nameGen.nextTemp();
					String trueLabel = nameGen.nextLabel(), falseLabel = nameGen.nextLabel();
					IRSeq seq = new IRSeq();
					BinaryExpr newExpr = new BinaryExpr(null, BinaryExpr.OpType.EQUAL, expr.left(), expr.right());
					seq.appendStmt(new IRMove(new IRTemp(tempName), new IRConst(eq ? 0 : 1)));
					seq.appendStmt((IRSeq) C(newExpr, trueLabel, falseLabel));
					seq.appendStmt(new IRLabel(trueLabel));
					seq.appendStmt(new IRMove(new IRTemp(tempName), new IRConst(eq ? 1 : 0)));
					seq.appendStmt(new IRLabel(falseLabel));
					return new IRESeq(seq, new IRTemp(tempName));
				}
				/* array<->array & scalar comparisons are handled by the code below */
				break;
			case PLUS:   opType = IRBinOp.OpType.ADD; break;
			case MINUS:  opType = IRBinOp.OpType.SUB; break;
			case TIMES:  opType = IRBinOp.OpType.MUL; break;
			case DIVIDE: opType = IRBinOp.OpType.DIV; break;
			case MODULO: opType = IRBinOp.OpType.MOD; break;
			case LEQ:    opType = IRBinOp.OpType.LE;  break;
			case GEQ:    opType = IRBinOp.OpType.GE;  break;
			case GT:     opType = IRBinOp.OpType.GT;  break;
			case LT:     opType = IRBinOp.OpType.LT;  break;
			
			/* we differentiate between bitwise ops and logical ops
			 * to support constant folding at the IR level.
			 */
			case OR:     
				if(expr.left().getType().isIdenticalType(PrimitiveType.BOOLEAN_TYPE))
					opType = IRBinOp.OpType.LOGOR;  
				else 
					opType = IRBinOp.OpType.OR;  
				break;
			case AND:    
				if(expr.left().getType().isIdenticalType(PrimitiveType.BOOLEAN_TYPE))
					opType = IRBinOp.OpType.LOGAND;  
				else 
					opType = IRBinOp.OpType.AND;  
				break;
			default:
				throw new InternalCompilerException(
					"Encountered an unknown binary operation type!");

		}
		return new IRBinOp(opType, E(expr.left()), E(expr.right()));
	}

	/* there is likely some unnecessary code duplication between here and TCall */
	private IRExpr ECall(Call call) {
		TypeNode callTargetType = call.target().getType();
		TupleExpr param = call.param();

		if (callTargetType instanceof FuncType) {
			/* Function call */
			TypeNode sourceType = param.getType();
			TypeNode targetType;
			IRCall irCall;
			IRSeq seq = null;

			if(call.target() instanceof Identifier) {
				FuncDecl funcDecl = functions.get(((Identifier) call.target()).name());
				targetType = funcDecl.param().getType();
				String targetName = NameMangler.getEncoding(funcDecl, null);
				irCall = new IRCall(new IRName(targetName),
						funcDecl.retType() == null ? 0 : funcDecl.retType().size());
			} else {
				DotExpr dotExpr = (DotExpr)call.target();
				ClassType targetClass = (ClassType)dotExpr.target().getType();
				
				String funcName = dotExpr.member();
				int funcOffset = targetClass.getDispatchTable().get(funcName);

				seq = new IRSeq();
				String objectTemp = ETemp(dotExpr.target(), seq);

				ClassType implClass = ((ClassType) dotExpr.target().getType())
						.findImplementingClass(funcName);
				FuncDecl funcDecl = functions.get(implClass.name() + "." + funcName);
				targetType = funcDecl.param().getType();

				IRMem funcAddr = new IRMem(new IRBinOp(OpType.ADD,
						new IRMem(new IRTemp(objectTemp), IRMem.MemType.IMMUTABLE),
						new IRConst(funcOffset*Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE);	

				irCall = new IRCall(funcAddr, funcDecl.retType() == null ? 0 : funcDecl.retType().size());
				irCall.appendParam(new IRTemp(objectTemp));
			}

			if (sourceType.size() > 1 || needsTypeConversion(sourceType, targetType)) {
				if(seq == null)
					seq = new IRSeq();
				DataSource ds = TConv(param, targetType, seq);

				for (int i=0; i<ds.size(); ++i)
					irCall.appendParam(fromDataSource(ds, i));
				return new IRESeq(seq, irCall);
			} else if (sourceType.size() == 1) {
				irCall.appendParam(E(param.subExprs().get(0)));
			}
			if(seq == null)
				return irCall;
			else
				return new IRESeq(seq, irCall);

		} else { 
			IRSeq seq = new IRSeq();

			DataSource arrayDS = new DataSource();

			if (callTargetType instanceof TupleType) {
				/* turn into an array of the appropriate type */
				boolean isSingle = true;
				for (TypeNode subType: ((TupleType) callTargetType).elementTypes()) {
					if (subType.size()!=1) {
						isSingle = false;
						break;
					}
				}

				if (isSingle) {
					DataSource dSource = call.target().getDataSource();
					String result = nameGen.nextTemp();
					String outLabel = nameGen.nextLabel();
					String nextLabel = nameGen.nextLabel();
					
					 if (!(call.target() instanceof Identifier))
						 seq.appendStmt(T(call.target(), dSource));

					/* save the index to a temp since we need to reuse it */
					String index = ETemp(call.param(), seq);

					/* need to generate a call_abort label */
					this.needsAbort = true;

					/*
					 * compare index to array length, jump to CALL_ABORT if out
					 * of range
					 */
					for (int i = 0; i < dSource.size(); i++) {
						seq.appendStmt(new IRCJump(new IRBinOp(
								IRBinOp.OpType.EQ, new IRTemp(index),
								new IRConst(i)), null, nextLabel));
						seq.appendStmt(new IRMove(new IRTemp(result),
								fromDataSource(dSource, i)));
						seq.appendStmt(new IRJump(new IRName(outLabel)));
						seq.appendStmt(new IRLabel(nextLabel));
						nextLabel = nameGen.nextLabel();
					}
					seq.appendStmt(new IRJump(new IRName(CALL_ABORT)));

					seq.appendStmt(new IRLabel(outLabel));
					return new IRESeq(seq, new IRTemp(result));
				} else {
					arrayDS = TConv(call.target(),
							new ArrayType(call.getType()), seq);
					if (arrayDS.size() != 1)
						throw new InternalCompilerException(
								"Size mismatch in ECall()");
				}
			} else {
				/* save array reference to a temp */
				arrayDS.addTempName(ETemp(call.target(), seq));
			}
			/* save the index to a temp since we need to use it twice */
			String index = ETemp(param, seq);

			/* compare index to array length, jump to CALL_ABORT if out of range */
			seq.appendStmt(new IRCJump(
					new IRBinOp(OpType.ULT, new IRTemp(index),
							new IRMem(new IRBinOp(OpType.SUB,
										fromDataSource(arrayDS),
										new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE)),
					null,
					CALL_ABORT));

			/* need to generate a call_abort label */
			this.needsAbort = true;

			/* now that seq contains bounds checking,
			 * return the ESEQ node that ends with the
			 * array value.
			 */
			return new IRESeq(seq, new IRMem(new IRBinOp(OpType.ADD, fromDataSource(arrayDS),
					new IRBinOp(OpType.AR_LSHIFT, new IRTemp(index), new IRConst(Configuration.LOG_WORD_SIZE)))));

		}
	}

	private IRExpr ENew(New n) {
		ClassType ct = (ClassType) n.getType();
		IRCall mallocCall = new IRCall(new IRName(MALLOC_NAME), 1);

		mallocCall.appendParam( 
				new IRMem(new IRName(ct.getClassSizeIdentifier()), IRMem.MemType.IMMUTABLE));
		String tempName = nameGen.nextTemp();

		IRSeq seq = new IRSeq();
		seq.appendStmt(new IRMove(new IRTemp(tempName), mallocCall));
		seq.appendStmt(new IRMove(new IRMem(new IRTemp(tempName)), new IRName(
				ct.getDispatchTableIdentifier())));

		return new IRESeq(seq, new IRTemp(tempName));
	}

	// ***************************************************************
	// ***               Statement translation rules               ***
	// ***************************************************************

	public IRStmt S(Stmt stmt) {
		if (stmt instanceof Assignment)
			return SAssignment(stmt);
		else if (stmt instanceof If)
			return SIf((If) stmt);
		else if (stmt instanceof LocalVarDecl)
			return SLocalVarDecl((LocalVarDecl) stmt);
		else if (stmt instanceof While)
			return SWhile((While) stmt);
		else if (stmt instanceof Break)
			return SBreak((Break) stmt);
		else if (stmt instanceof Block)
			return SBlock((Block) stmt);
		else if (stmt instanceof Return)
			return SReturn((Return) stmt);
		else if (stmt instanceof ExprStmt)
			return SExprStmt((ExprStmt) stmt);
		else
			throw new InternalCompilerException("Encountered an unknown statement node - S" 
					+ stmt.getClass().getSimpleName() + "() still needs to be implemented!");
	}
	
	private IRStmt SBreak(Break stmt) {
		String escapeLabel = escapeLabelStack.peek();
		return new IRJump(new IRName(escapeLabel));
	}

	private IRStmt SExprStmt(ExprStmt stmt) {
		
		if (stmt.expr().getType() instanceof UnitType ||
				stmt.expr().getType().size() == 1)
			return new IRExp(E(stmt.expr()));
		else
			return T((TupleExpr) stmt.expr(), stmt.expr().getDataSource());
	}

	private IRStmt SLocalVarDecl(LocalVarDecl varDecl) {
		DataSource targetDS = varDecl.tupleDecl().getDataSource();
		IRSeq seq = new IRSeq();	

		if (varDecl.hasInit()) {
			TypeNode sourceType = varDecl.init().getType();
			TypeNode targetType = varDecl.tupleDecl().getType();

			if (sourceType.size() > 1 || needsTypeConversion(sourceType, targetType)) {
				DataSource source = TConv(varDecl.init(), targetType, seq);
				seq.appendStmt(D(varDecl.tupleDecl(), source));
			} else {
				seq.appendStmt(new IRMove(fromDataSource(targetDS), E(varDecl.init())));
			}

			return seq;
		} else {
			TypeNode targetType = varDecl.tupleDecl().getType();

			for(int i=0; i<targetType.size(); i++)
				/* Initialize bools to false, ints to 0, and arrays to null */
				seq.appendStmt(new IRMove(fromDataSource(targetDS,i), new IRConst(0)));			
			
			seq.appendStmt(D(varDecl.tupleDecl(), null));
			return seq;
		}
	}
	
	private IRStmt SReturn(Return ret) {
		if (epilogueName == null)
			epilogueName = nameGen.nextLabel();
		if(ret.hasExpr()) {
			TypeNode retType = functions.get(currentFunction).retType();
			IRSeq seq = new IRSeq();

			if (retType.size() > 1) {
				/* Returning a tuple - write to a special location 
				   passed in as the first parameter */
				DataSource ds = TConv(ret.expr(), retType, seq);

				/* Fill the entries of the return value */
				for (int i=0; i<ds.size(); ++i) {
					seq.appendStmt(new IRMove(new IRMem(new IRBinOp(
							IRBinOp.OpType.ADD, new IRTemp(RegAlloc.RV_NAME), new IRConst(i
									* Configuration.WORD_SIZE))), fromDataSource(ds, i)));
				}
			} else {
				int retSize = ret.expr().getType().size();
				if (retSize > 1 || needsTypeConversion(ret.expr().getType(), retType)) {
					DataSource ds = TConv(ret.expr(), retType, seq);
					seq.appendStmt(new IRMove(new IRTemp(RegAlloc.RV_NAME), fromDataSource(ds)));
				} else {
					seq.appendStmt(new IRMove(new IRTemp(RegAlloc.RV_NAME), E(ret.expr())));
				}
			}
			seq.appendStmt(new IRJump(new IRName(epilogueName)));
			return seq;
		} else {
			return new IRJump(new IRName(epilogueName));
		}
	}

	/**
	 * Takes a list of temps and possibly 
	 *  - turns parts into arrays
	 *  - removes some, if the target type contains the unit type somewhere
	 * @param ds
	 *      Input data source
	 * @param inputType
	 *      Input type of a tuple expression
	 * @param outputType
	 *      Type of the target expression
	 * @param seq
	 *      The resulting conversion code will be appended to this IRSeq
	 * @return
	 *      A new data source
	 */
	private DataSource typeConversion(DataSource ds, TypeNode inputType,
			TypeNode outputType, IRSeq seq) {
		DataSource result = new DataSource();
		if (inputType.size() != ds.size())
			throw new InternalCompilerException("Invalid input to arrayConversion(): " +
					ds.size() + " input temps, whereas the input type has size " +
					inputType.size());

		if (outputType instanceof UnitType) {
			/* Nothing */
		} else if (outputType instanceof TupleType) {
			int pos = 0, inputPos = 0;
			for (TypeNode oType : ((TupleType) outputType).elementTypes()) {
				TypeNode iType = ((TupleType) inputType).elementTypes().get(pos++);
				if (!(oType instanceof UnitType))
					result.add(typeConversion(ds.subList(inputPos, inputPos + iType.size()), iType, oType, seq));
				inputPos += iType.size();
			}
		} else if (!(inputType instanceof ArrayType) && outputType instanceof ArrayType) {
			if (!(inputType instanceof TupleType))
				inputType = new TupleType(inputType);  // Simplifies some of the below

			/* Generate IR which allocates and initializes an array */
			String tempName = nameGen.nextTemp();
			TupleType tType = (TupleType) inputType;
			ArrayType aType = (ArrayType) outputType;
			int length = tType.elementTypes().size();

			DataSource updatedDS = new DataSource();
			int pos = 0;
			for (TypeNode type : tType.elementTypes()) {
				updatedDS.add(typeConversion(ds.subList(pos, pos + type.size()), type, aType.elementType(), seq));
				pos += type.size();
			}

			IRCall call = new IRCall(new IRName(MALLOC_NAME), 1);
			call.appendParam(new IRConst((length*aType.elementType().size()+1)*Configuration.WORD_SIZE));
			seq.appendStmt(new IRMove(new IRTemp(tempName), call));
			seq.appendStmt(new IRMove(new IRMem(new IRTemp(tempName)), 
					new IRConst(((TupleType) inputType).elementTypes().size())));

			pos = 0;
			for (int i=0; i<length; ++i) {
				for (int j=0; j<aType.elementType().size(); ++j) {
					seq.appendStmt(
							new IRMove(new IRMem(new IRBinOp(IRBinOp.OpType.ADD, 
								new IRTemp(tempName), new IRConst((i*aType.elementType().size()+j+1)*Configuration.WORD_SIZE))),
								fromDataSource(updatedDS, pos++)
							)
					);
				}
			}

			seq.appendStmt(new IRMove(new IRTemp(tempName), new IRBinOp(IRBinOp.OpType.ADD, 
					new IRTemp(tempName), new IRConst(Configuration.WORD_SIZE))));

			result.addTempName(tempName);
		} else {
			if (outputType.size() != 1 || inputType.size() != 1 || ds.size() != 1)
				throw new InternalCompilerException("Internal error in array conversion");
			result.add(ds);
		}

		return result;
	}

	/**
	 * Check if a type conversion would be required - parameters are
	 * the same as above. This is used to simplify the tuple comparison
	 * routine in EBinaryExpr(). Essentially, if a tuple->array conversion
	 * would be required, any EQUALS comparison will return false.
	 */
	private boolean needsTypeConversion(TypeNode inputType, TypeNode outputType) {
		boolean result = false;

		if (outputType instanceof TupleType) {
			int pos = 0;
			for (TypeNode oType : ((TupleType) outputType).elementTypes()) {
				TypeNode iType = ((TupleType) inputType).elementTypes().get(pos++);
				result |= needsTypeConversion(iType, oType);
			}
		/* never need to convert null */
		} else if (inputType instanceof NullType || outputType instanceof NullType) {
			return false;
		} else if (!(inputType instanceof ArrayType) && outputType instanceof ArrayType) {
			result = true;
		} else if (outputType instanceof UnitType) {
			result = true;
		} else {
			if (outputType.size() != 1 || inputType.size() != 1)
				throw new InternalCompilerException("Internal error in needsArrayConversion()");
		}
		return result;
	}

	private IRStmt SBlock(Block block) {
		IRSeq seq = new IRSeq();
		for (Stmt stmt : block.stmts())
			seq.appendStmt(S(stmt));
		return seq;
	}

	private IRStmt SWhile(While whileStmt) {
		IRSeq newSeq = new IRSeq();
		String headLabel = nameGen.nextLabel(), 
		       trueLabel = nameGen.nextLabel(),
		       endLabel  = nameGen.nextLabel();
		escapeLabelStack.push(endLabel);
		
		IRStmt loopBody = S(whileStmt.body());

		/* Don't peel loops when we are already in a peeled loop */
		if (peelLoops && !inPeeledLoop.peek()
				&& loopBody.nodeCount() < MAX_PEEL_NODES) {
			inPeeledLoop.push(true);
			String peelLabel = nameGen.nextLabel();
			newSeq.appendStmt(C(whileStmt.cond(), peelLabel, endLabel));
			newSeq.appendStmt(new IRLabel(peelLabel));
			newSeq.appendStmt(S(whileStmt.body()));
			inPeeledLoop.pop();
			nPeeled++;
		}

		newSeq.appendStmt(new IRLabel(headLabel));
		newSeq.appendStmt(C(whileStmt.cond(), trueLabel, endLabel));
		newSeq.appendStmt(new IRLabel(trueLabel));
		newSeq.appendStmt(loopBody);
		newSeq.appendStmt(new IRJump(new IRName(headLabel)));
		newSeq.appendStmt(new IRLabel(endLabel));

		escapeLabelStack.pop();
		return newSeq;
	}

	private IRStmt SIf(If ifStmt) {
	    String trueLabel  = nameGen.nextLabel(),
	           falseLabel = nameGen.nextLabel(), 
	           doneLabel = null;
		IRSeq newSeq = new IRSeq();
		newSeq.appendStmt(C(ifStmt.guard(), trueLabel, falseLabel));
		newSeq.appendStmt(new IRLabel(trueLabel));
		newSeq.appendStmt(S(ifStmt.consequent()));

		if (ifStmt.hasAlternative()) {
			doneLabel = nameGen.nextLabel();
			newSeq.appendStmt(new IRJump(new IRName(doneLabel)));
		}

		newSeq.appendStmt(new IRLabel(falseLabel));
		if (ifStmt.hasAlternative()) {
			newSeq.appendStmt(S(ifStmt.alternative()));
			newSeq.appendStmt(new IRLabel(doneLabel));
		}
		return newSeq;
	}

	private IRStmt SAssignment(Stmt stmt) {
		/*
		 * Identifiers, Array
		 * S[e1 = e2] = MOVE(E(e1), E(e2));
		 * Tuple : careful about (x,y)=(y,x);
		 * 
		 */
		Expr target = ((Assignment)stmt).lvalue();
		Expr source = ((Assignment)stmt).rvalue();
		boolean isTupleAssignment = target instanceof Call
			&& ((Call) target).target().getType() instanceof TupleType;

		if (source.getType().size() > 1
				|| needsTypeConversion(source.getType(), target.getType())
				|| isTupleAssignment) {
			IRSeq seq = new IRSeq();
			/* Evaluate rvalue and store to temps to handle
			 * non-commutative assignments like
			 * (x,y) = (y,x) 
			 * 
			 * We generate a temp for each rvalue item.
			 * A copy propagation optimization pass would 
			 * remove unnecessary copies here.
			 */
			DataSource ds = TConv(source, target.getType(), seq);
			seq.appendStmt(A(target, ds));
			return seq;
		} else {
			return new IRMove(E(target), E(source));			
		}
	}

	// ***************************************************************
	// ***             Control flow translation rules              ***
	// ***************************************************************

	public IRStmt C(Expr expr, String t, String f) {
		if (expr instanceof TupleExpr && ((TupleExpr) expr).getType().size() == 1)
			expr = ((TupleExpr) expr).peel();
		if (expr instanceof BoolLiteral) {
			BoolLiteral boolLit = (BoolLiteral) expr;
			if(boolLit.value() == true)
				return new IRJump(new IRName(t));
			else
				return new IRJump(new IRName(f));	
		} else if (expr instanceof UnaryExpr) {
			UnaryExpr e = (UnaryExpr) expr;
			if (e.opType() != UnaryExpr.OpType.NOT)
				throw new InternalCompilerException("Unexpected unary "
						+"operation in C() transformation");
			return C(e.expr(), f, t);
		} else if (expr instanceof BinaryExpr) {
			BinaryExpr binExpr = (BinaryExpr)expr;
			if (binExpr.left().getType().size() == 1) {
				switch(binExpr.opType()) {
					case OR: {
						/* short-circuit OR */
						IRSeq seq = new IRSeq();
						String label = nameGen.nextLabel();
						seq.appendStmt(C(binExpr.left(), t, label));
						seq.appendStmt(new IRLabel(label));
						seq.appendStmt(C(binExpr.right(), t, f));
						return seq;
					}
					case AND: {		
						/* short-circuit AND */	
						IRSeq seq = new IRSeq();
						String label = nameGen.nextLabel();
						seq.appendStmt(C(binExpr.left(), label, f));
						seq.appendStmt(new IRLabel(label));
						seq.appendStmt(C(binExpr.right(), t, f));
						return seq;
					}
					default:	
						return new IRCJump(E(expr), t, f);
				}
			} else {
				/* Called from EBinaryExpr() - only tuples of matching
				   sizes will make it to here. */
				TupleExpr leftExpr = (TupleExpr) binExpr.left();
				TupleExpr rightExpr = (TupleExpr) binExpr.right();
				IRSeq seq = new IRSeq();
				Expr firstExpr = new BinaryExpr(null, binExpr.opType(), leftExpr.peel(), rightExpr.peel());
				Expr remainderExpr = new BinaryExpr(null, binExpr.opType(), leftExpr.remainder(), rightExpr.remainder());
				String nextLabel = nameGen.nextLabel();
				switch(binExpr.opType()) {
					case EQUAL: {
						IRStmt firstComp = C(firstExpr, nextLabel, f);
						seq.appendStmt(firstComp);
						seq.appendStmt(new IRLabel(nextLabel));
						seq.appendStmt(C(remainderExpr, t, f));
						break;
					}
					/* NOT_EQUAL implemented by negation in EBinaryExpr */
					default:
						throw new InternalCompilerException("Invalid operator in tuple binary expression");
				}
				return seq;
			}
		} else {
			return new IRCJump(E(expr), t, f);
		}
	}
	
	// ***************************************************************
	// ***                Tuple translation rule                   ***
	// ***************************************************************

	/**
	 * Beefed-up version of T. Takes an expr/tupleExpr, evaluates it,
	 * possibly performs type conversion, and returns a list of temps
	 * filled with data in the requested type.
	 */
	public DataSource TConv(Expr expr, TypeNode targetType, IRSeq seq) {
		DataSource ds = new DataSource();
		/* if expr is a tuple type, then we need to reuse the array generated before */

		if (expr.getType().size() == 1) {
			ds.addTempName(ETemp(expr, seq));
		} else {
			ds = expr.getDataSource();
			seq.appendStmt(T(expr, ds));
		}
		
		return typeConversion(ds, expr.getType(), targetType, seq);
	}

	/**
	 * Evaluate a tuple-valued expression and put it into a list of 
	 * designated TEMPs
	 */
	public IRStmt T(Expr e, DataSource dests) {
		IRSeq seq = new IRSeq();
		if (dests.size() != e.getType().size())
			throw new InternalCompilerException("Internal error: tuple size mismatch in T():" +
					" got " + dests.size() + " TEMPs, expression size is " + e.getType().size() +
					" data words");

		if (e.getType().size() == 1) {
			/* Handle directly */
			IRMove move = new IRMove(fromDataSource(dests), E(e));
			if (!move.expr().equalsTree(move.target()))
				seq.appendStmt(move);
			return seq;
		} if (e instanceof Call) {
			/* Function call or array indexing operation */
			return TCall((Call) e, dests);
		} if (e instanceof DotExpr) {
			return TDotExpr((DotExpr) e, dests);
		} if (e instanceof Identifier) {
			/* Identifier - copy from parameters / TEMPs to TEMPs */
			DataSource ds = e.getDataSource();
			if (dests.size() != ds.size())
				throw new InternalCompilerException("Internal error: tuple size mismatch in T():" +
						" got " + dests.size() + " TEMPs, data source size is " + ds.size() +
						" data words");

			for (int i=0; i<dests.size(); ++i) {
				IRMove move = new IRMove(fromDataSource(dests, i), fromDataSource(ds, i));
				if (!move.expr().equalsTree(move.target()))
					seq.appendStmt(move);
			}

			return seq;
		} else if (e.getType().size() > 1) {
			/* This is a tuple - peel off one expression and recurse */
			Expr head = ((TupleExpr )e).peel(), remainder = ((TupleExpr) e).remainder();
			int headSize = head.getType().size();
			seq.appendStmt(T(head, dests.subList(0, headSize)));
			seq.appendStmt(T(remainder, dests.subList(headSize, dests.size())));
			return seq;
		} else {
			return null;
		}
	}

	private IRStmt TDotExpr(DotExpr expr, DataSource dests) {
		ClassType targetClass = (ClassType) expr.target().getType();
		String member = expr.member();

		IRExpr offset = targetClass.isDerived() ? new IRMem(new IRName(
				targetClass.parent().getClassSizeIdentifier()), IRMem.MemType.IMMUTABLE)
				: new IRConst(0);

		if (targetClass == currentClass)
			offset = new IRTemp(RegAlloc.OFFSET_NAME);

		List<Integer> fieldOffsets = targetClass.getFieldTable().get(member);
		if (fieldOffsets.size() != dests.size())
			throw new InternalCompilerException("Size mismatch in TDotExpr!");

		IRSeq seq = new IRSeq();

		String tempName = nameGen.nextTemp();

		seq.appendStmt(new IRMove(new IRTemp(tempName), new IRBinOp(OpType.ADD,
				E(expr.target()), offset)));

		for (int i = 0; i < dests.size(); ++i) {
			IRMove move = new IRMove(fromDataSource(dests, i), new IRMem(
					new IRBinOp(OpType.ADD, new IRTemp(tempName), new IRConst(
							fieldOffsets.get(i) * Configuration.WORD_SIZE))));
			seq.appendStmt(move);
		}
		return seq;
	}

	private IRStmt TCall(Call call, DataSource dests) {
		TypeNode callTargetType = call.target().getType();
		TupleExpr param = call.param();
		IRSeq seq = new IRSeq();
		int resultSize = call.getType().size();
		if (resultSize == 1)
			throw new InternalCompilerException("TCall() invoked on a single-word expression!");

		TypeNode sourceType = param.getType();
		TypeNode targetType;
		IRCall irCall;

		if (callTargetType instanceof FuncType) {
			if(call.target() instanceof Identifier) {
				FuncDecl funcDecl = functions.get(((Identifier) call.target()).name());
				targetType = funcDecl.param().getType();
				String targetName = NameMangler.getEncoding(funcDecl, null);
				irCall = new IRCall(new IRName(targetName), funcDecl.retType() == null ? 0 : funcDecl.retType().size());
			} else {
				DotExpr dotExpr = (DotExpr)call.target();
				ClassType targetClass = (ClassType)dotExpr.target().getType();
				
				String funcName = dotExpr.member();
				int funcOffset = targetClass.getDispatchTable().get(funcName);
				
				seq = new IRSeq();
				String objectTemp = ETemp(dotExpr.target(), seq);

				ClassType implClass = ((ClassType) dotExpr.target().getType())
						.findImplementingClass(funcName);
				FuncDecl funcDecl = functions.get(implClass.name()+"."+funcName);
				targetType = funcDecl.param().getType();

				IRMem funcAddr = new IRMem(new IRBinOp(OpType.ADD,
											new IRMem(new IRTemp(objectTemp), IRMem.MemType.IMMUTABLE),
											new IRConst(funcOffset*Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE);	
				
				irCall = new IRCall(funcAddr, funcDecl.retType() == null ? 0 : funcDecl.retType().size());
				irCall.appendParam(new IRTemp(objectTemp));
			}

			/* Generate code for the function call */
			String rvTemp = nameGen.nextTemp();

			seq.appendStmt(new IRMove(new IRTemp(rvTemp), new IRTemp(RegAlloc.CALL_RV_SCRATCH)));
			irCall.insertParam(0, new IRTemp(rvTemp));

			if (sourceType.size() > 1 || needsTypeConversion(sourceType, targetType)) {
				DataSource ds = TConv(param, targetType, seq);
				for (int i=0; i<ds.size(); ++i)
					irCall.appendParam(fromDataSource(ds, i));
			} else if (sourceType.size() == 1) {
				irCall.appendParam(E(param.subExprs().get(0)));
			}
			seq.appendStmt(new IRExp(irCall));

			/* Save the results */
			for (int i=0; i<resultSize; ++i) {
				seq.appendStmt(new IRMove(fromDataSource(dests, i), 
						new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(RegAlloc.CALL_RV_SCRATCH),
								new IRConst(i*Configuration.WORD_SIZE)))));
			}

			return seq;
		} else if (callTargetType instanceof TupleType || callTargetType instanceof ArrayType) {
			if (callTargetType instanceof TupleType) {
				/* turn into an array of the appropriate type */
				boolean isHomogeneous = true;

				for (TypeNode subType: ((TupleType) callTargetType).elementTypes()) {
					if (subType.size()!=((TupleType) callTargetType).elementTypes().get(0).size()) {
						isHomogeneous = false;
						break;
					}
				}

				if (isHomogeneous) {
					DataSource dSource = call.target().getDataSource();
					
					if (!(call.target() instanceof Identifier))
						seq.appendStmt(T(call.target(), call.target().getDataSource()));
					String outLabel = nameGen.nextLabel();
					String nextLabel = nameGen.nextLabel();

					/* save the index to a temp since we need to reuse it */
					String index = ETemp(call.param(), seq);

					/* need to generate a call_abort label */
					this.needsAbort = true;

					/*
					 * compare index to array length, jump to CALL_ABORT if out
					 * of range
					 */
					for (int i = 0; i < dSource.size()/resultSize; i++) {
						seq.appendStmt(new IRCJump(new IRBinOp(
								IRBinOp.OpType.EQ, new IRTemp(index),
								new IRConst(i)), null, nextLabel));
						for (int j=0; j<resultSize; ++j)
							seq.appendStmt(new IRMove(fromDataSource(dests, j),
									fromDataSource(dSource, i*resultSize+j)));
						seq.appendStmt(new IRJump(new IRName(outLabel)));
						seq.appendStmt(new IRLabel(nextLabel));
						nextLabel = nameGen.nextLabel();
					}
					seq.appendStmt(new IRJump(new IRName(CALL_ABORT)));

					seq.appendStmt(new IRLabel(outLabel));
				}
				return seq;
			}

			DataSource arrayDS = null;
			if (callTargetType instanceof TupleType) {
				/* Tuple indexing */
				arrayDS = TConv(call.target(), new ArrayType(call.getType()), seq);
				if (arrayDS.size() != 1)
					throw new InternalCompilerException("Size mismatch in TCall()");
			} else {
				arrayDS = new DataSource();
				arrayDS.addTempName(ETemp(call.target(), seq));
			}

			/* save the index to a temp since we need to use it twice */
			String indexName = ETemp(param, seq);

			/* compare index to array length, jump to CALL_ABORT if out of range */
			seq.appendStmt(new IRCJump(
					new IRBinOp(OpType.ULT, new IRTemp(indexName),
							new IRMem(new IRBinOp(OpType.SUB,
										fromDataSource(arrayDS),
										new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE)),
					null,
					CALL_ABORT));

			/* need to generate a call_abort label */
			this.needsAbort = true;

			/* Compute the offset */
			IRBinOp offsetExpr = new IRBinOp(OpType.ADD, fromDataSource(arrayDS),
					new IRBinOp(OpType.MUL, new IRTemp(indexName), 
							new IRConst(resultSize*Configuration.WORD_SIZE)));

			String tempName = nameGen.nextTemp();
			seq.appendStmt(new IRMove(new IRTemp(tempName), offsetExpr));

			/* Copy the values into their designated TEMPS */
			for (int i=0; i<dests.size(); ++i) {
				seq.appendStmt(new IRMove(fromDataSource(dests, i),
						new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(tempName),
								new IRConst(i * Configuration.WORD_SIZE)))));
			}
		}
		return seq;
	}

	// ***************************************************************
	// ***                 D translation rule                      ***
	// ***************************************************************

	/**
	 * Take a list of TEMPs and use them to initialize the left-hand side
	 * of a tuple declaration. When 'dest' is set to NULL, we still possibly
	 * need to allocate memory for arrays which have a length expressions.
	 */
	public IRSeq D(TupleDecl d, DataSource dests) {
		IRSeq seq = new IRSeq();

		if (!d.isLeaf()) {
			if (d.subTupleDecls().size() >= 1) {
				/* Peel off one element and recurse */
				TupleDecl head = d.peel(), remainder = d.remainder();
				int headSize = typeSize(head.getType());
				seq.appendStmt(D(head, dests == null ? null : dests.subList(0, headSize)));
				seq.appendStmt(D(remainder, dests == null ? null : dests.subList(headSize, dests.size())));
			}
		} else if (!d.isUnderscore()) {
			if (dests != null) {
				/* Move into the TEMPs corresponding to a variable */
				if (dests.size() != d.getDataSource().size())
					throw new InternalCompilerException("Internal error: size mismatch in D()");
				for (int i=0; i<dests.size(); ++i)
					seq.appendStmt(new IRMove(fromDataSource(d.getDataSource(), i), fromDataSource(dests, i)));
			} else {
				List<String> lengthTemps = new ArrayList<String>();
				collectLengthExprs(d.getType(), lengthTemps, seq);
				initializeTemps(d.getType(), d.getDataSource(), lengthTemps.iterator(), seq);
			}
		}
		return seq;
	}


	/* Collect array length expressions in a tuple declaration and store their results into TEMPs */
	private void collectLengthExprs(TypeNode type, List<String> lengthTemps, IRSeq seq) {
		if (type instanceof ArrayType && ((ArrayType) type).hasLengthExpr()) {
			String newTemp = nameGen.nextTemp();
			seq.appendStmt(new IRMove(new IRTemp(newTemp), E(((ArrayType) type).lengthExpr())));
			lengthTemps.add(newTemp);
			collectLengthExprs(((ArrayType) type).elementType(), lengthTemps, seq);
		} else if (type instanceof TupleType) {
			for (TypeNode elementType : ((TupleType) type).elementTypes())
				collectLengthExprs(elementType, lengthTemps, seq);
		}
	}

	/* Walk through a tuple declaration and call M() on all arrays that need to be initialized */
	private void initializeTemps(TypeNode type, DataSource temps, Iterator<String> lengthTemps, IRSeq seq) {
		if (type instanceof ArrayType) {
			seq.appendStmt(M(type, fromDataSource(temps), lengthTemps));
		} else if (type instanceof TupleType) {
			int pos = 0;
			for (TypeNode eType : ((TupleType) type).elementTypes()) {
				initializeTemps(eType, temps.subList(pos, pos+typeSize(eType)), lengthTemps, seq);
				pos += typeSize(eType);
			}
		}
	}

	/**
	 * The M translation generates array allocation code, in particular in
	 * the presence of nesting. This makes it possible to handle code such
	 * as
	 *       a:(int, (int, int[2])[3]);
	 * 
	 */
	public IRStmt M(TypeNode type, IRExpr target, Iterator<String> lengthTemps) {
		IRSeq seq = new IRSeq();
		if (type instanceof TupleType) {
			int currentOffset =0;
			boolean generatedCode = false;
			String tempName = nameGen.nextTemp();
			seq.appendStmt(new IRMove(new IRTemp(tempName), target));
			for (TypeNode elementType : ((TupleType) type).elementTypes()) {
				IRStmt mCode = M(elementType, new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(tempName),
						new IRConst(currentOffset)), lengthTemps);
				seq.appendStmt(mCode);
				if (mCode != null)
					generatedCode = true;
				currentOffset += elementType.size() * Configuration.WORD_SIZE;
			}
			if (!generatedCode)
				return null;
		} else if (type instanceof ArrayType && ((ArrayType) type).hasLengthExpr()) {
			String lengthName = lengthTemps.next();
			ArrayType at = (ArrayType) type; 
			String tempName = nameGen.nextTemp();

			IRCall call = new IRCall(new IRName(MALLOC_NAME), 1);
			call.appendParam(new IRBinOp(OpType.ADD,
					new IRConst(Configuration.WORD_SIZE),
					new IRBinOp(OpType.MUL,
					new IRTemp(lengthName),
					new IRConst(at.elementType().size()
							* Configuration.WORD_SIZE))));

			seq.appendStmt(new IRMove(new IRTemp(tempName), new IRBinOp(IRBinOp.OpType.ADD, 
					call, new IRConst(Configuration.WORD_SIZE))));

			seq.appendStmt(new IRMove(new IRMem(new IRBinOp(IRBinOp.OpType.SUB,
					new IRTemp(tempName), new IRConst(Configuration.WORD_SIZE))), new IRTemp(lengthName)));
			if (target instanceof IRTemp || target instanceof IRMem)
				seq.appendStmt(new IRMove(target, new IRTemp(tempName)));
			else
				seq.appendStmt(new IRMove(new IRMem(target), new IRTemp(tempName)));

			seq.appendStmt(MInitArray(at, tempName, lengthName, lengthTemps));
		} else {
			return null;
		}
		return seq;
	}

	/* Allocate an array and call the initialization routine on the array elements */
	private IRStmt MInitArray(ArrayType type, String arrayAddrName, String lengthName, Iterator<String> lengthTemps) {
		IRSeq seq = new IRSeq();
		String counterName = nameGen.nextTemp();
		String loopStartLabel = nameGen.nextLabel();
		String loopBodyLabel = nameGen.nextLabel();
		String loopEndLabel = nameGen.nextLabel();
		seq.appendStmt(new IRMove(new IRTemp(counterName), new IRConst(0)));
		seq.appendStmt(new IRLabel(loopStartLabel));
		seq.appendStmt(new IRCJump(new IRBinOp(IRBinOp.OpType.EQ, new IRTemp(
				counterName), new IRTemp(lengthName)), loopEndLabel,
				loopBodyLabel));
		seq.appendStmt(new IRLabel(loopBodyLabel));

		IRExpr newTarget = new IRBinOp(IRBinOp.OpType.ADD, new IRTemp(
				arrayAddrName), new IRBinOp(IRBinOp.OpType.MUL, new IRTemp(
				counterName), new IRConst(type.elementType().size()
				* Configuration.WORD_SIZE)));

		IRStmt mCode = M(type.elementType(), newTarget, lengthTemps);
		if (mCode == null) {
			/* Array elements need no initialization - do not generate the loop code */
			return null;
		}
		seq.appendStmt(mCode);
		seq.appendStmt(new IRMove(new IRTemp(counterName), new IRBinOp(
				IRBinOp.OpType.ADD, new IRTemp(counterName), new IRConst(1))));
		seq.appendStmt(new IRJump(new IRName(loopStartLabel)));
		seq.appendStmt(new IRLabel(loopEndLabel));
		return seq;
	}

	// ***************************************************************
	// ***                     A translation rule                  ***
	// ***     A is the dual of T: it takes a list of sources      *** 
	// ***     and moves them into an lvalue expression.           ***
	// ***************************************************************

	public IRStmt A(Expr target, DataSource sourceDS) {
		TypeNode targetType = target.getType();
		IRSeq seq = new IRSeq();
		boolean isTupleAssignment = target instanceof Call
				&& ((Call) target).target().getType() instanceof TupleType;

		if(targetType.size() == 1 && !isTupleAssignment) {
			/* Word-valued expression -> handle directly */
			return new IRMove(E(target), fromDataSource(sourceDS));
		} else if (target instanceof Identifier) {
			DataSource targetDS = target.getDataSource();
			if (sourceDS.size() != targetDS.size())
				throw new InternalCompilerException("Size mismatch in A()!");
			for (int i=0; i<sourceDS.size(); ++i)
				seq.appendStmt(new IRMove(fromDataSource(targetDS, i), fromDataSource(sourceDS, i)));
		} else if (target instanceof DotExpr) {
			/* Target is a dot expression */
			DotExpr expr = (DotExpr) target;
			ClassType targetClass = (ClassType) expr.target().getType();
			String member = expr.member();

			IRExpr offset = targetClass.isDerived() ? new IRMem(new IRName(
					targetClass.parent().getClassSizeIdentifier()))
					: new IRConst(0);

			if (targetClass == currentClass)
				offset = new IRTemp(RegAlloc.OFFSET_NAME);

			List<Integer> fieldOffsets = targetClass.getFieldTable().get(member);
			if (fieldOffsets.size() != sourceDS.size())
				throw new InternalCompilerException("Size mismatch in TDotExpr!");

			if (sourceDS.size() == 1) {
				/* Avoid creating an extra temp */
				seq.appendStmt(new IRMove(new IRMem(new IRBinOp(OpType.ADD,
						new IRBinOp(OpType.ADD, E(expr.target()), offset),
						new IRConst(fieldOffsets.get(0)
								* Configuration.WORD_SIZE))), fromDataSource(sourceDS)));
			} else {
				String tempName = nameGen.nextTemp();
	
				seq.appendStmt(new IRMove(new IRTemp(tempName), new IRBinOp(OpType.ADD,
						E(expr.target()), offset)));
	
				for (int i = 0; i < sourceDS.size(); ++i) {
					IRMove move = new IRMove(new IRMem(
							new IRBinOp(OpType.ADD, new IRTemp(tempName), new IRConst(
									fieldOffsets.get(i) * Configuration.WORD_SIZE))),
									fromDataSource(sourceDS, i));
					seq.appendStmt(move);
				}
			}
			return seq;
		} else if (target instanceof Call) {
			/* Target is an array access or tuple indexing operation */
			Call callExpr = (Call) target;
			int resultSize = callExpr.getType().size();
			if (sourceDS.size() != resultSize)
				throw new InternalCompilerException("Size mismatch in A()!");

			if (callExpr.target().getType() instanceof ArrayType) {
				String array = ETemp(callExpr.target(), seq);
				String index = ETemp(callExpr.param(), seq);

				/* compare index to array length, jump to CALL_ABORT if out of range */
				seq.appendStmt(new IRCJump(
						new IRBinOp(OpType.ULT, new IRTemp(index),
								new IRMem(new IRBinOp(OpType.SUB,
											new IRTemp(array),
											new IRConst(Configuration.WORD_SIZE)), IRMem.MemType.IMMUTABLE)),
						null,
						CALL_ABORT));

				/* need to generate a call_abort label */
				this.needsAbort = true;

				IRBinOp offsetExpr = new IRBinOp(OpType.ADD, new IRTemp(array), 
						new IRBinOp(OpType. MUL, new IRTemp(index), 
								new IRConst(resultSize * Configuration.WORD_SIZE)));

				if (resultSize > 1) {
					/* Assign each local word to the proper place in the array */
					String tempName = nameGen.nextTemp();
					seq.appendStmt(new IRMove(new IRTemp(tempName), offsetExpr));

					for(int i = 0; i<resultSize; ++i)
						seq.appendStmt(new IRMove(new IRMem(new IRBinOp(OpType.ADD,
								new IRTemp(tempName), new IRConst(i * Configuration.WORD_SIZE))), 
								fromDataSource(sourceDS, i)));
				} else {
					seq.appendStmt(new IRMove(new IRMem(offsetExpr), fromDataSource(sourceDS)));
				}
			} else if (callExpr.target().getType() instanceof TupleType) {
				String outLabel = nameGen.nextLabel();
				String nextLabel = nameGen.nextLabel();

				DataSource ds = null;
				/* Do we need to evaluate the target first? */
				if (callExpr.target() instanceof Call)
					ds = TConv(callExpr.target(), callExpr.target().getType(), seq);
				else 
					ds = callExpr.target().getDataSource();

				/* save the index to a temp since we need to reuse it */
				String index = ETemp(callExpr.param(), seq);

				/* need to generate a call_abort label */
				this.needsAbort = true;

				/*
				 * compare index to array length, jump to CALL_ABORT if out
				 * of range
				 */
				for (int i = 0; i < ds.size()/resultSize; i++) {
					seq.appendStmt(new IRCJump(new IRBinOp(
							IRBinOp.OpType.EQ, new IRTemp(index),
							new IRConst(i)), null, nextLabel));

					for (int j =0; j<resultSize; ++j)
						seq.appendStmt(new IRMove(fromDataSource(ds, i*resultSize+j), fromDataSource(sourceDS, j)));

					seq.appendStmt(new IRJump(new IRName(outLabel)));
					seq.appendStmt(new IRLabel(nextLabel));
					nextLabel = nameGen.nextLabel();
				}
				seq.appendStmt(new IRJump(new IRName(CALL_ABORT)));

				seq.appendStmt(new IRLabel(outLabel));
				
				/* Move the results back */
				if (callExpr.target() instanceof Call)
					seq.appendStmt(A(callExpr.target(), ds));
			}
		} else if(target instanceof TupleExpr) {
			int pos = 0;
			for (Expr expr : ((TupleExpr) target).subExprs()) {
				seq.appendStmt(A(expr, sourceDS.subList(pos, pos + expr.getType().size())));
				pos += expr.getType().size();
			}
			if (pos != sourceDS.size())
				throw new InternalCompilerException("Size mismatch in A()!");
		} else {
			throw new CompilationException("A: unexpected node: " 
					+ target.label(), null);
		}
		return seq;
	}
}
