package xic.visitors;

import xic.InternalCompilerException;
import xic.ast.ASTNode;
import xic.ir.IRBinOp;
import xic.ir.IRCJump;
import xic.ir.IRCall;
import xic.ir.IRCompUnit;
import xic.ir.IRConst;
import xic.ir.IRCtor;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRESeq;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRFuncDecl;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.IRZData;
import xic.ir.NameGenerator;

/**
 * Implements the lowering transformation using a visitor pattern
 */
public class IRLowering extends NodeVisitor {
	private NameGenerator nameGen;

	public IRLowering (NameGenerator namegenerator) {
		nameGen = namegenerator;
	}

	protected ASTNode leave(ASTNode parent, ASTNode n, ASTNode nNew,
			NodeVisitor vNew) {
		if (nNew instanceof IRStmt)
			return LStmt((IRStmt) nNew);
		else if (nNew instanceof IRExpr)
			return LExpr((IRExpr) nNew);
		else if (nNew instanceof IRCompUnit)
			return nNew;
		else if (nNew instanceof IRFuncDecl)
			return nNew;
		throw new InternalCompilerException("Encountered an unknown IR node - L"
				+ nNew.getClass().getSimpleName()
				+ "() still needs to be implemented!");
	}

	private IRStmt LStmt(IRStmt irstmt) {
		if (irstmt instanceof IRSeq)
			return LSeq((IRSeq) irstmt);
		else if (irstmt instanceof IRExp)
			return LExp((IRExp) irstmt);
		else if (irstmt instanceof IRJump)
			return LJump((IRJump) irstmt);
		else if (irstmt instanceof IRCJump)
			return LCJump((IRCJump) irstmt);
		else if (irstmt instanceof IRLabel)
			return LLabel((IRLabel) irstmt);
		else if (irstmt instanceof IRMove)
			return LMove((IRMove) irstmt);
		else if (irstmt instanceof IRDataSeq)
			return LDataSeq((IRDataSeq) irstmt);
		else if (irstmt instanceof IRDecl)
			return LDecl((IRDecl) irstmt);
		else if (irstmt instanceof IRZData)
			return LZData((IRZData) irstmt);
		else if (irstmt instanceof IRCtor)
			return LCtor((IRCtor) irstmt);
		else
			throw new InternalCompilerException("Encountered an unknown IR Statement node - L" 
					+ irstmt.getClass().getSimpleName() + "() still needs to be implemented!");
	}
	
	private IRNode LExpr(IRExpr irexpr) {
		if (irexpr instanceof IRConst)
			return LConst((IRConst) irexpr);
		else if (irexpr instanceof IRName)
			return LName((IRName) irexpr);
		else if (irexpr instanceof IRTemp)
			return LTemp((IRTemp) irexpr);
		else if (irexpr instanceof IRMem)
			return LMem((IRMem) irexpr);
		else if (irexpr instanceof IRESeq)
			return LESeq((IRESeq) irexpr);
		else if (irexpr instanceof IRCall)
			return LCall((IRCall) irexpr);
		else if (irexpr instanceof IRBinOp)
			return LBinOp((IRBinOp) irexpr);
		throw new InternalCompilerException("Encountered an unknown IR Expression node - L" 
				+ irexpr.getClass().getSimpleName() + "() still needs to be implemented!");
	}

	// ***************************************************************
	// ***                Lowering IR statements                   ***
	// ***************************************************************

	/**
	 * Lowering SEQ(s1, s2, ..., sn)
	 * L(e) = L(s1); L(s2); ...; L(sn) 
	 */
	private IRStmt LSeq(IRSeq irseq) {
		/* Nothing to do here - handled by code in IRSeq */
		return irseq;
	}
	
	private IRDataSeq LDataSeq(IRDataSeq dseq) {
		/* Nothing to do here */
		return dseq;
	}

	/**
	 * Lowering EXP(e)
	 * L(e) = s+; e' => L(EXP(e)) = s+
	 */
	private IRStmt LExp(IRExp irexp) {
		return ((IRESeq) irexp.expr()).stmt();
	}

	/**
	 * Lowering JUMP(e)
	 * L(e) = s+; e' => L(JUMP(e)) = s+; JUMP(e')
	 */
	private IRStmt LJump(IRJump irjump) {
		IRSeq newJumpSeq = new IRSeq();
		IRESeq target = (IRESeq) irjump.target();
		newJumpSeq.appendStmt(target.stmt());
		newJumpSeq.appendStmt(new IRJump(target.expr()));
		return newJumpSeq;
	}

	/**
	 * Lowering CJUMP(e, t, f)
	 * L(e) = s+; e' => L(CJUMP(e)) = s+; CJUMP(e', t, f)
	 */
	private IRStmt LCJump(IRCJump ircjump) {
		IRSeq newCJumpSeq = new IRSeq();
		IRESeq expr = (IRESeq)ircjump.expr();
		newCJumpSeq.appendStmt(expr.stmt());
		newCJumpSeq.appendStmt(new IRCJump(expr.expr(), ircjump.trueLabel(), ircjump.falseLabel()));
		return newCJumpSeq;
	}

	/**
	 * Lowering LABEL(l)
	 * L(LABEL(l)) = LABEL(l)
	 */
	private IRStmt LLabel(IRLabel irlabel) {
		return irlabel;
	}

	/**
	 * Lowering DECL(l)
	 * L(DECL(l)) = DECL(l)
	 */
	private IRStmt LDecl(IRDecl irdecl) {
		return irdecl;
	}

	/**
	 * Lowering L(MOVE(dest, expr))
	 * L(expr) = s+; expr'    L(dest) = s'+; dest'
	 * => L(MOVE(dest, expr)) = s+; MOVE(TEMP(t), e'); s'+; MOVE(dest', t)
	 */
	private IRStmt LMove(IRMove irmove) {
		IRSeq newMoveSeq = new IRSeq();
		IRESeq lvalue = (IRESeq) (irmove.target());
		IRESeq rvalue = (IRESeq) (irmove.expr());
		if (lvalue.stmt().commutesWith(rvalue.expr())) {
			newMoveSeq.appendStmt(rvalue.stmt());
			newMoveSeq.appendStmt(lvalue.stmt());
			newMoveSeq.appendStmt(new IRMove(lvalue.expr(), rvalue.expr()));
		} else {
			String tempName = nameGen.nextTemp();
			newMoveSeq.appendStmt(rvalue.stmt());
			newMoveSeq.appendStmt(new IRMove(new IRTemp(tempName), rvalue.expr()));
			newMoveSeq.appendStmt(lvalue.stmt());
			newMoveSeq.appendStmt(new IRMove(lvalue.expr(), new IRTemp(tempName)));
		}
		return newMoveSeq;
	}
	
	private IRZData LZData(IRZData irzdata) {
		return irzdata;
	}
	
	private IRCtor LCtor(IRCtor ctor) {
		return ctor;
	}

	// ***************************************************************
	// ***                Lowering IR expressions                  ***
	// ***************************************************************

	/**
	 * Lowering L(CONST(i))
	 * L(CONST(i)) = (empty seq of statements); CONST(i)
	 */
	private IRNode LConst(IRConst irconst) {
		IRSeq emptyStmt = new IRSeq();
		return new IRESeq(emptyStmt, irconst);
	}

	/**
	 * Lowering L(NAME(i))
	 * L(NAME(i)) = (empty seq of statements); NAME(i)
	 */
	private IRNode LName(IRName irname) {
		return new IRESeq(new IRSeq(), irname);
	}

	/**
	 * Lowering L(TEMP(i))
	 * L(TEMP(i)) = (empty seq of statements); TEMP(i)
	 */
	private IRNode LTemp(IRTemp irtemp) {
		return new IRESeq(new IRSeq(), irtemp);
	}

	/**
	 * Lowering L(MEM(e))
	 * L(e) = s+; e'  =>  L(MEM(e)) = s+; MEM(e')
	 */
	private IRNode LMem(IRMem irmem) {
		IRESeq eseq = (IRESeq) irmem.expr();
		return new IRESeq(eseq.stmt(), new IRMem(eseq.expr(), irmem.memType()));
	}

	/**
	 * Lowering L(ESEQ(s, e))
	 * L(s) = s+    L(e) = s'+; e' 
	 *  =>  L(ESEQ(s, e)) = s+; s'+; e'
	 */
	private IRNode LESeq(IRESeq ireseq) {
		IRESeq eseq2 = (IRESeq) ireseq.expr();
		IRSeq seq = new IRSeq();
		seq.appendStmt(ireseq.stmt());
		seq.appendStmt(eseq2.stmt());
		return new IRESeq(seq, eseq2.expr());
	}

	/**
	 * Lowering L(OP(e1, e2))
	 * L(e1) = s1+;e1'   L(e) = s2+; e2' 
	 *  =>  L(OP(e1, e2)) = s1+; s2+; OP(e1', e2')
	 */
	private IRNode LBinOp(IRBinOp irbinop) {
		IRESeq left = (IRESeq) irbinop.left();
		IRESeq right = (IRESeq) irbinop.right();
		IRSeq seq = new IRSeq();
		
		if (right.stmt().commutesWith(left.expr())) {
			seq.appendStmt(left.stmt());
			seq.appendStmt(right.stmt());

			return new IRESeq(seq, new IRBinOp(irbinop.opType(),
					left.expr(), right.expr()));
		} else {
			String tempName = nameGen.nextTemp();
			seq.appendStmt(left.stmt());
			seq.appendStmt(new IRMove(new IRTemp(tempName), left.expr()));
			seq.appendStmt(right.stmt());

			return new IRESeq(seq, new IRBinOp(irbinop.opType(),
					new IRTemp(tempName), right.expr()));			
		}
	}

	/**
	 * Lowering L(CALL(e0, e1, ..., en))
	 * L(ei) = si+; ei' (i=0..n)  
	 *  => L(CALL(e0, e1, e2,..., en)) = s0+; 
	 *                                     MOVE(TEMP(t0), e0');
	 *                                   s1+; 
	 *                                     MOVE(TEMP(t1), e1')
	 *                                   ...
	 *                                   sn+; 
	 *                                     MOVE(TEMP(tn), en');
	 *                                   MOVE(TEMP(t), CALL(t0, t1, ..., tn));
	 *                                   TEMP(t)
	 */
	private IRNode LCall(IRCall ircall) {
		IRSeq newSeq = new IRSeq();
		IRESeq target = (IRESeq) ircall.target();
		String valueTemp = nameGen.nextTemp();
		String[] tempNames = new String[ircall.params().size()];
		boolean[] needsTemp = new boolean[ircall.params().size()];

		/* should only have SEQ node */
		if (((IRSeq)target.stmt()).stmts().size() > 1)
			throw new InternalCompilerException("Call target has side-effects!");
		
				
		for (int i=0; i<needsTemp.length; ++i) {
			IRESeq peseq = (IRESeq) ircall.params().get(i);
			boolean hasSE = ((IRSeq) peseq.stmt()).stmts().size() > 0;
			if (hasSE) {
				for (int j=0; j<=i; ++j)
					needsTemp[j] = true;
			} else {
				needsTemp[i] = false;
			}
		}

		int idx=0;
		for (IRExpr param : ircall.params()) {
			IRESeq peseq = (IRESeq) param;

			if (!needsTemp[idx]) {
				idx++;
				continue;
			}

			newSeq.appendStmt(peseq.stmt());
			tempNames[idx] = nameGen.nextTemp();
			newSeq.appendStmt(new IRMove(new IRTemp(tempNames[idx]), peseq.expr()));
			idx++;
		}

		IRCall call = new IRCall(target.expr(), ircall.returnSize());
		idx=0;
		for (IRExpr param : ircall.params()) {
			IRESeq peseq = (IRESeq) param;
			if (!needsTemp[idx])
				call.appendParam(peseq.expr());
			else
				call.appendParam(new IRTemp(tempNames[idx]));
			idx++;
		}

		newSeq.appendStmt(new IRMove(new IRTemp(valueTemp), call));
		return new IRESeq(newSeq, new IRTemp(valueTemp));
	}
}
