package xic;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.util.List;

import org.junit.Before;
import org.junit.Test;

import xic.Configuration.Architecture;
import xic.Configuration.AssemblySyntax;
import xic.Configuration.Target;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRMem;
import xic.ir.IRTemp;


public class CodeGenTest {
	@Before
	public void initialize() {
		Configuration.setArchitecture(Architecture.I386);
		Configuration.setTarget(Target.LINUX);
		Configuration.setSyntax(AssemblySyntax.INTEL_SYNTAX);
	}

	public void compareStrings(String expected, String actual) {
		String n_expected = expected.replaceAll("\\s+", "").trim();
		String n_actual = actual.replaceAll("\\s+","").trim();
		assertEquals(n_expected, n_actual);
	}

	@Test
	public void testIncDecBin() {
		/* This is usually translated to INCs/DECs when possible */
		String assembly = ExecutableTest.getAssembly(
			"a(j:int) = {" +
			"  i:int=0;" +
			"  i = i+1;" +
			"  i = 1+i;" +
			"  j = j+1;" +
			"  i = i-1;" +
			"  i = 1-i;" + // DEC doesn't work here, since "-" is not commutative
			"}", 
			new String[] {  "-O", "--no_ra", "--asmformat", "intel"  }
		);

		/* Test that this is the case */
		compareStrings(
			".intel_syntax noprefix" +
			".text" +
			".globl _Ia_pi" +
			"_Ia_pi:" +
			"   mov _j, dword ptr [FP+8]" +
			"   mov _i, 0"  +
			"	inc _i" +
			"	inc _i" +
			"	inc _j" +
			"	dec _i" +
			"   mov t0, 1" +
			"   sub t0, _i" +
			"   mov _i, t0 ",
			assembly
		);

		/* Now prioritize the BinOpMatcher */
		assembly = ExecutableTest.getAssembly(
			"a(j:int) = {" +
			"  i:int;" +
			"  i = i+1;" +
			"  i = 1+i;" +
			"  j = j+1;" +
			"  i = i-1;" +
			"  i = 1-i;" +
			"}", 
			new String[] {   "-O", "--no_ra" , "--asmformat", "intel", "--boost", "BinOpMatcher" }
		);

		compareStrings(
			".intel_syntax noprefix" +
			".text" +
			".globl _Ia_pi" +
			"_Ia_pi:" +
			"   mov _j, dword ptr [FP+8]" +
			"   mov _i, 0"  +
			"	add _i, 1" +
			"	add _i, 1" +
			"	add _j, 1" +
			"	sub _i, 1" +
			"   mov t0, 1" +
			"   sub t0, _i" +
			"   mov _i, t0",
			assembly
		);
	}

	@Test
	public void testUnaryOp() {
		String assembly = ExecutableTest.getAssembly(
			"a(j:int, c:bool) = {" +
			"  i:int;" +
			"  b:bool;" +
			"  i = -i;" +
			"  j = -j;" +
			"  b = !b;" +
			"  c = !c;" +
			"  i = -j;" +
			"}", 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);
		
		// Logical NOT should not be translated to the NOT 
		// instruction (it flips each bit) - stick with XOR <arg>, 1

		compareStrings(
			".intel_syntax noprefix" +
			".text" +
			".globl _Ia_pib" +
			"_Ia_pib:" +
			"   mov _j, dword ptr [FP+8]" +
			"   mov _c, dword ptr [FP+12]" +
			"   mov _i, 0"  +
			"   mov _b, 0"  +
			"	neg _i" +
			"	neg _j" +
			"	xor _b, 1" +
			"	xor _c, 1" +
			"	mov t0, _j" +
			"	neg t0" +
			"	mov _i, t0",
			assembly
		);
	}

	@Test
	public void testAddrModeFinder() {
		IRMem mem = new IRMem(new IRBinOp(IRBinOp.OpType.ADD, new IRConst(23), 
			new IRBinOp(IRBinOp.OpType.ADD,
				new IRBinOp(IRBinOp.OpType.AR_LSHIFT, new IRTemp("t1"), new IRConst(3)),
				new IRTemp("t2")
			)
		));
		Tiler tiler = new Tiler();
		List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);
		for (AddrMode mode: modes) {
			if (mode.getScale() == 8 && mode.getDisplacement() == 23 &&
					mode.hasBase() && mode.hasIndex() &&
					mode.estimateCost() == 0)
				return; // success
			
		}
		fail("Unable to find the addressing mode");
	}

	@Test
	public void testAddrModeFinder2() {
		IRMem mem = new IRMem(new IRBinOp(IRBinOp.OpType.ADD,  
				new IRBinOp(IRBinOp.OpType.MUL, new IRTemp("t1"), new IRConst(5)),
				new IRConst(15)
		));
		Tiler tiler = new Tiler();
		List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);
		for (AddrMode mode: modes) {
			if (mode.getScale() == 4 && mode.getDisplacement() == 15 && 
					mode.hasBase() && mode.hasIndex() &&
					mode.estimateCost() == 0)
				return; // success
			
		}
		fail("Unable to find the addressing mode");
	}

	@Test
	public void testAddrModeFinder3() {
		IRBinOp binOp = new IRBinOp(IRBinOp.OpType.ADD,
				new IRBinOp(IRBinOp.OpType.MUL, new IRTemp("t1"),
						new IRConst(Configuration.WORD_SIZE)),
						new IRConst(Configuration.WORD_SIZE));
		Tiler tiler = new Tiler();
		List<AddrMode> modes = AddrModeHelper.findLeaAddressingModes(tiler, binOp);
		for (AddrMode mode: modes) {
			if (mode.getScale() == 4 && mode.getDisplacement() == 4 &&
					!mode.hasBase() && mode.hasIndex() &&
					mode.estimateCost() == 0)
				return; // success
			
		}
		fail("Unable to find the addressing mode");
	}

	@Test
	public void testBinOpComparison() {
		// BinOps involving booleans
		String assembly = ExecutableTest.getAssembly(
			"a(b:bool) = {" +
			"  k:bool;" +
			"  b = b|true;" +
			"  b = k|b;" +
			"  b = (4 == 5);" +
			"}", 
			new String[] {  "-O", "--no_ra" , "--boost", "BinOpMatcher", "--asmformat", "intel" }
		);

		compareStrings(
			".intel_syntax noprefix" +
			".text" +
			".globl _Ia_pb" +
			"_Ia_pb:" +
			"   mov _b, dword ptr [FP+8]" +
			"   mov _k, 0"  +
			"   test _b, _b" +
			"   jnz .L1 " +
			".L1:" +
			"	mov t1, 1" +
			"	mov _b, t1" +
			"	mov t3, 1" +
			"	xor t3, _k" +
			"	test t3, t3" +
			"	jnz .L7" +
			".L5:" +
			"	mov t2, 1" +
			".L4:"+
			"	mov _b, t2"+
			"	mov t4, 4"+
			"	xor t5, t5" +
			"	cmp t4, 5" +
			"	sete t5" +
			"	mov _b, t5" +
			"	jmp .E0" +
			".L6:" +
			"	mov t2, 0" +
			"	jmp .L4" +
			".L7:" +
			"	test _b, _b" +
			"	jnz .L5" +
			"	jmp .L6" +
			".E0:",assembly
		);
	}

	@Test
	public void testLEA() {
		// BinOps involving booleans
		String assembly = ExecutableTest.getAssembly(
			"a() = {" +
			"  i:int;" +
			"  j:int;" +
			"  i = 21+j + 4*i;" +
			"  j = 9*i + 3;" +   /// :)
			"}", 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings( 
			".intel_syntax noprefix" +
			".text" +
			".globl _Ia_p" +
			"_Ia_p:" +
			"   mov _i, 0"  +
			"   mov _j, 0"  +
			"   lea _i, dword ptr [_j + _i*4 + 21]" +
			"   lea _j, dword ptr [_i + _i*8 + 3]",
			assembly
		);
	}
	
	@Test
	public void testCallwithTuple() {
		String assembly = ExecutableTest.getAssembly(
			"a(x:(int,int,bool)) = {" +
			"  	(i:int, j:int, k:bool) = x;" +
			"}" , 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings(
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_pt3iib"+
				"_Ia_pt3iib:"+
				"   mov _x$0, dword ptr [FP+8]" +
				"   mov _x$1, dword ptr [FP+12]" +
				"   mov _x$2, dword ptr [FP+16]"  +
				"   mov _i, _x$0" +
				"   mov _j, _x$1" +
				"   mov _k, _x$2",
				assembly
				);
	}
	
	@Test
	public void testCallwithNestedTuple() {
		String assembly = ExecutableTest.getAssembly(
			"a(x:(int,(int,bool),bool,(int,int))) = {" +
			"  	(i:int, (j:int,_), k:bool, (_,l:int)) = x;" +
			"}" , 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings(
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_pt4it2ibbt2ii"+
				"_Ia_pt4it2ibbt2ii:"+
				"   mov _x$0, dword ptr [FP+8]" +
				"   mov _x$1, dword ptr [FP+12]" +
				"   mov _x$2, dword ptr [FP+16]" +
				"   mov _x$3, dword ptr [FP+20]" +
				"   mov _x$4, dword ptr [FP+24]" +
				"   mov _x$5, dword ptr [FP+28]" +
				"   mov _i, _x$0"+
				"   mov _j, _x$1"+
				"   mov _k, _x$3"+
				"   mov _l, _x$5",
				assembly
				);
	}
	
	@Test
	public void testCallwithArray() {
		String assembly = ExecutableTest.getAssembly(
				"a(x:int[]) = {" +
				"  	i:int = x 1;" +
				"}",
				new String[] { "-O", "--no_ra", "--asmformat", "intel" }
			);

			compareStrings(assembly,
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_pai"+
				"_Ia_pai:"+
				"   mov _x, dword ptr [FP+8]"+
				"   mov t1, 1"+
				"   cmp t1, dword ptr [_x-4]"+
				"   jae call_abort"+
				"   mov _i, dword ptr [_x+t1*4]"+
				"call_abort:"+
				"   call _I_outOfBounds_p"
			);
	}
	@Test
	public void testIf() {
		String assembly = ExecutableTest.getAssembly(
				"a(x:int[]) = {" +
				"  	i:int = x 1;" +
				"	if (i > 4)" +
				"		x i = 3;" +
				"	else " +
				"		x i = 0;" +
				"}",
				new String[] { "-O", "--no_ra", "--asmformat", "intel" }
			);

			compareStrings(
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_pai"+
				"_Ia_pai:"+
				"   mov _x, dword ptr [FP+8]"+
				"   mov t1, 1"+
				"   cmp t1, dword ptr [_x-4]"+
				"   jae call_abort"+
				"   mov _i, dword ptr [_x+t1*4]" +
				"   cmp _i, 4" +
				"   jle .L1" +
				"   cmp _i, dword ptr [_x-4]"+
				"   jae call_abort"+
				"   mov dword ptr [_x+_i*4], 3" +
				".L2:" +
				"   jmp .E0"+
				".L1:" +
				"   cmp _i, dword ptr [_x-4]"+
				"   jae call_abort"+
				"   mov dword ptr [_x+_i*4], 0" +
				"   jmp .L2"+
				".E0:" +
				"call_abort:"+
				"   call _I_outOfBounds_p",
				assembly);
	}

	@Test
	public void testTupleParam() {
		String assembly = ExecutableTest.getAssembly(
			"a(x:(int,int), y:(int,int)) = {" +
			"  	(i:int, j:int) = x;" +
			"  	(k:int, l:int) = y;" +
			"}" +
			"b() = {" +
			"	a((1,2),(3,4));" +
			"}", 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings(
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_pt2iit2ii" +
				"_Ia_pt2iit2ii:"+
				"   mov _x$0, dword ptr [FP+8]"+
				"   mov _x$1, dword ptr [FP+12]"+
				"   mov _y$0, dword ptr [FP+16]"+
				"   mov _y$1, dword ptr [FP+20]"+
				"   mov _i, _x$0"+
				"   mov _j, _x$1"+
				"   mov _k, _y$0"+
				"   mov _l, _y$1" +
				".globl _Ib_p" +
				"_Ib_p:" +
				"   mov t0,1" +
				"   mov t1,2" +
				"   mov t2,3" +
				"   mov t3,4" +
				"   mov dword ptr[ARG_SCRATCH+12], t3" +
				"   mov dword ptr[ARG_SCRATCH+8], t2" +
				"   mov dword ptr[ARG_SCRATCH+4], t1" +
				"   mov dword ptr[ARG_SCRATCH], t0" +
				"   call _Ia_pt2iit2ii",
				assembly
			);
	}

	@Test
	public void testStringTable() {
		// Two matching strings are allocated to a single string
		// table entry. Before use, a local copy is made from the
		// data segment using the internal function strdup
		String assembly = ExecutableTest.getAssembly(
			"a() = {" +
			"	x:int[] = \"Hello\";" +
			"}" +
			"b() = {" +
			"	x:int[] = \"Hello\";" +
			"}", 
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings(
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_p" +
				"_Ia_p:" +
				"	lea t5, dword ptr[.L0]" +
				"   mov dword ptr[ARG_SCRATCH], t5" +
				"	call strdup" +
				"	mov t6, eax" +
				"	mov t2, t6" +
				"	mov _x, t2" +
				"" +
				".globl _Ib_p" +
				"_Ib_p:" +
				"	lea t7, dword ptr [.L0]" +
				"   mov dword ptr [ARG_SCRATCH], t7" +
				"	call strdup" +
				"	mov t8, eax" +
				"	mov t3, t8" +
				"	mov _x, t3" +
				"" +
				"strdup:" +
				"	mov p0, dword ptr [FP+8]" +
				"	mov t0, dword ptr [p0]" +
				"	lea t9, dword ptr [t0*4+4]" +
				"	mov dword ptr [ARG_SCRATCH], t9" +
				"	call _I_alloc_i" +
				"	mov t10, eax" +
				"	mov t4, t10" +
				"	mov t1, t4" +
				".L1:" +
				"	mov t11, dword ptr [p0+t0*4]" +
				"	mov dword ptr [t1+t0*4], t11" +
				"	dec t0" +
				"	cmp t0, 0" +
				"	jge .L1" +
				"	mov t12,t1" +
				"   add t12,4" +
				"   mov RV,t12" +
				"	mov eax, RV" +
				"" +
				"	.section .rodata" +
				"	.align 4" +
				".L0:" +
				"	.long 5" +
				"	.long 72" +
				"	.long 101" +
				"	.long 108" +
				"	.long 108" +
				"	.long 111" +
				"	.text",
			assembly);
	}
	
	@Test
	public void testTupleIndexing() {
		String assembly = ExecutableTest.getAssembly(
			  "b( ) = {" +
			  "	  x:(int,int)= (1,2);" +
			  "	  x 1 = 3;" +
			  "	  y:int = x 1;" +
			  "}",
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);

		compareStrings (
			".intel_syntax noprefix" +
			".text" +
			".globl _Ib_p" +
			"_Ib_p:" +
			"	mov t0, 1" +
			"	mov t1, 2" +
			"	mov _x$0, t0" +
			"	mov _x$1, t1" +
			"	mov t4, 3" +
			"	mov t5, 1" +
			"	cmp t5, 0" +
			"	jne .L1" +
			"	mov _x$0, t4" +
			".L0:" +
			"	mov t7, 1" +
			"	cmp t7, 0" +
			"	jne .L5" +
			"	mov t6, _x$0" +
			".L4:" +
			"	mov _y, t6" +
			"	jmp .E0" +
			".L1:" +
			"	cmp t5, 1" +
			"	jne call_abort" +
			"	mov _x$1, t4" +
			"	jmp .L0" +
			".L5:" +
			"	cmp t7, 1" +
			"	jne call_abort" +
			"	mov t6, _x$1" +
			"	jmp .L4" +
			".E0:" +
			"call_abort:" +
			"	call _I_outOfBounds_p", assembly
			);
	}
	
	@Test
	public void testGlobalVar() {
		String assembly = ExecutableTest.getAssembly(
			  "y:int =1;" +
			  "b( ) = {" +
			  "	  y = 2; " +
			  "}",
			new String[] { "-O", "--no_ra", "--asmformat", "intel" }
		);
		compareStrings (
			".intel_syntax noprefix" +
			".text" +
			".globl _Ib_p" +
			"_Ib_p:" +
			"	mov dword ptr [_I_g_y_i], 2" +
			"	.section .data" +
			"	.align 4" +
			".globl _I_g_y_i" +
			"_I_g_y_i:" +
			"	.long 1" +
			"	.text",assembly
			);
	}
	
	@Test
	public void testCodeOrderingProblem() {
		String assembly = ExecutableTest.getAssembly(
			"a() = {" +
			"    if (false) { } else { }"+
			"    if (true) { }"+
			"}", new String[] {"--asmformat", "intel"});
		compareStrings (
				".intel_syntax noprefix" +
				".text" +
				".globl _Ia_p" +
				"_Ia_p:" +
				"    ret",
				assembly);

	}
}
