package xic;

import static org.junit.Assert.assertEquals;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;

import org.junit.Before;
import org.junit.Test;

import polyglot.util.OptimalCodeWriter;
import xic.Configuration.Architecture;
import xic.ast.CompUnit;
import xic.ast.ErrorReporter;
import xic.ir.IRCompUnit;
import xic.ir.IRPrinter;
import xic.parser.IotaParser;
import xic.parser.ParserImpl.Language;
import xic.parser.TokenType;
import xic.parser.XiParser;
import xic.visitors.IRConstantFolder;
import xic.visitors.IRGenerator;
import xic.visitors.IRLowering;
import xic.visitors.TypeChecker;
import edu.cornell.cs.cs4120.xi.CompilationException;

public class IRGenTest {
	@Before
	public void initialize() {
		Configuration.setArchitecture(Architecture.I386);
	}

	public static IRCompUnit compile(String input) {
		try {
			ErrorSink errorSink = new ErrorSink();
			IotaParser p = new IotaParser(input, TokenType.SOURCE_IDENTIFIER);
			CompUnit compUnit = p.doParse(errorSink);
	
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}
	
			TypeChecker typeChecker = new TypeChecker(Language.IOTA_09, errorSink, new ArrayList<String>());

			CompUnit typeCheckedUnit = typeChecker.run(compUnit);
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}
			
			IRGenerator generator = new IRGenerator(typeChecker.getNameGenerator(), false, Language.IOTA_09);
			return generator.translate(typeCheckedUnit, typeChecker.getImports());
		} catch (IOException e) {
			throw new CompilationException("I/O error", null);
		}
	}
	public static IRCompUnit compileAndLower(String input) {
		try {
			ErrorSink errorSink = new ErrorSink();
			XiParser p = new XiParser(input, TokenType.SOURCE_IDENTIFIER);
			CompUnit compUnit = p.doParse(errorSink);
	
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}
	
			TypeChecker typeChecker = new TypeChecker(Language.IOTA_09, errorSink, new ArrayList<String>());

			CompUnit typeCheckedUnit = typeChecker.run(compUnit);
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}
			
			IRGenerator generator = new IRGenerator(typeChecker.getNameGenerator(), false, Language.IOTA_09);
			IRCompUnit irCompUnit = generator.translate(typeCheckedUnit, typeChecker.getImports());
			
			IRConstantFolder constantFolder = new IRConstantFolder(errorSink);
			irCompUnit = (IRCompUnit) constantFolder.visit(irCompUnit);
				
		
			IRLowering loweringVisitor = new IRLowering(typeChecker.getNameGenerator());
			irCompUnit = (IRCompUnit) loweringVisitor.visit(irCompUnit);
				
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input,errorSink);
				throw new CompilationException("Error optimizing string", null);
			}
			return irCompUnit;
		} catch (IOException e) {
			throw new CompilationException("I/O error", null);
		}
	}
	
	public static String getIRString(String input) {		
		ByteArrayOutputStream output = new ByteArrayOutputStream();
		OutputStreamWriter writer = new OutputStreamWriter(output);
		IRPrinter printer = new IRPrinter(
					new OptimalCodeWriter(writer, 80));
		printer.print(compile(input));
		return output.toString();
	}

	public void compareIRStrings(String expectedIR, String actualIR) {
		String n_expected = expectedIR.replaceAll("\\s+", "").trim();
		String n_actual = actualIR.replaceAll("\\s+","").trim();
		if (!n_expected.equals(n_actual))
			assertEquals(n_expected, n_expected);
	}

	@Test
	public void testSimpleFunction() {
		// Tests function return
		String source = 
			"a(i:int):int = {\n" +
				"return 1;\n"+
			"}";
		String expectedIR = "COMPUNIT(" +
		"	FUNC _Ia_ii(" +
		"		    SEQ(" +
		"              MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
		"			   MOVE(TEMP(RV), CONST(0))," +
		"		       MOVE(TEMP(RV), CONST(1))," +
		"		       JUMP(NAME(.L0))," +
		"		       LABEL(.L0)" +
		"		    )" +
		"		 )" +
		"		) ";
		compareIRStrings(expectedIR,getIRString(source));
	}

	@Test
	public void testSimpleProcedure() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
				"return;\n"+
			"}";
		String expectedIR = "COMPUNIT(" +
		"	FUNC _Ia_pi(" +
		"		    SEQ(" +
		"              MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
		"		       JUMP(NAME(.L0))," +
		"		       LABEL(.L0)" +
		"		    )" +
		"		 )" +
		"		)";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testTupleDecl() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
				"(x:int, y:(int, int), (z:int,u:int)) = (1,(2,3),(4,5));\n"+
			"}";
		
		String expectedIR = "COMPUNIT("+
		"	FUNC _Ia_pi("+
		"			SEQ("+
		"               MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
	    "               MOVE(TEMP(t4), CONST(1)),"+ 
	    "           	MOVE(TEMP(t0), CONST(2)),"+ 
	    "           	MOVE(TEMP(t1), CONST(3)),"+ 
	    "           	MOVE(TEMP(t2), CONST(4)),"+ 
	    "           	MOVE(TEMP(t3), CONST(5)),"+
        "    			MOVE(TEMP(_x), TEMP(t4)),"+ 
        "    			MOVE(TEMP(_y$0), TEMP(t0)),"+ 
        "    			MOVE(TEMP(_y$1), TEMP(t1)),"+ 
        "    			MOVE(TEMP(_z), TEMP(t2)),"+ 
        "    			MOVE(TEMP(_u), TEMP(t3))"+
        "		)"+
        "	 )"+
        " )";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testTupleDeclWithUnderscore() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
				"(_, y:(int, int), (_,u:int)) = (1,(2,3),(4,5));\n"+
			"}";
		
		String expectedIR = "COMPUNIT("+
		"	FUNC _Ia_pi("+
		"			SEQ("+
		"               MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
	    "           	MOVE(TEMP(t4), CONST(1)),"+ 
	    "           	MOVE(TEMP(t0), CONST(2)),"+ 
	    "           	MOVE(TEMP(t1), CONST(3)),"+ 
	    "           	MOVE(TEMP(t2), CONST(4)),"+ 
	    "           	MOVE(TEMP(t3), CONST(5)),"+
        "    			MOVE(TEMP(_y$0), TEMP(t0)),"+ 
        "    			MOVE(TEMP(_y$1), TEMP(t1)),"+ 
        "    			MOVE(TEMP(_u), TEMP(t3))"+
        "			)"+
        "	)"+
        " )";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testArrayAccess() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
				"x:int[] = (1,2,3,4);\n"+
				"y:int = x 2;\n"+
			"}";
		String expectedIR = "COMPUNIT("
				+ "	FUNC _Ia_pi("
				+ "		    SEQ("
				+ "           MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8)))),"
				+ "           MOVE(TEMP(t0), CONST(1)),"
				+ "           MOVE(TEMP(t1), CONST(2)),"
				+ "           MOVE(TEMP(t2), CONST(3)),"
				+ "           MOVE(TEMP(t3), CONST(4)),"
				+ "           MOVE(TEMP(t5), CALL(NAME(_I_alloc_i), CONST(20))),"
				+ "           MOVE(MEM(TEMP(t5)), CONST(4)),"
				+ "           MOVE(MEM(ADD(TEMP(t5), CONST(4))), TEMP(t0)),"
				+ "           MOVE(MEM(ADD(TEMP(t5), CONST(8))), TEMP(t1)),"
				+ "           MOVE(MEM(ADD(TEMP(t5), CONST(12))), TEMP(t2)),"
				+ "           MOVE(MEM(ADD(TEMP(t5), CONST(16))), TEMP(t3)),"
				+ "           MOVE(TEMP(t5), ADD(TEMP(t5), CONST(4))),"
				+ "           MOVE(TEMP(_x), TEMP(t5)), "
				+ "           MOVE(TEMP(_y), " 
				+ "				ESEQ( SEQ( "
				+ "					MOVE(TEMP(t6), CONST(2)), " 
				+ "					CJUMP(ULT(TEMP(t6), MEM_I(SUB(TEMP(_x), CONST(4)))), <none>, call_abort)),"
				+ "					MEM(ADD(TEMP(_x), AR_LSHIFT(TEMP(t6), CONST(2))))))"
				+ "		    )" 
				+ "		 ),"
				+ "SEQ( LABEL(call_abort), EXP(CALL(NAME(_I_outOfBounds_p))))"
				+ "\t\n)";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testAssignments() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
				"x:int = 1;"+
				"y:int = 2;"+
				"(x,y)=(y,x);"+
			"}";
		String expectedIR = "COMPUNIT("+
		"		FUNC _Ia_pi("+
		"			SEQ("+
		"               MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
		"    			MOVE(TEMP(_x), CONST(1)),"+
        "				MOVE(TEMP(_y), CONST(2)),"+
        "       		MOVE(TEMP(t2), TEMP(_y)),"+ 
        "       		MOVE(TEMP(t3), TEMP(_x)),"+
        "       		MOVE(TEMP(_x), TEMP(t2)),"+ 
        "       		MOVE(TEMP(_y), TEMP(t3))"+
        "			)"+
        "		)"+
        " )";
		
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testIf() {
		String source = 
			"a(i:int) = {\n" +
			"    x:int = 0;"+
			"    if (true) x=1;\n"+
			"    x=2;\n"+
			"}";
		String expectedIR = "COMPUNIT("+
		"		FUNC _Ia_pi("+
		"			SEQ("+
		"               MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
        "    			MOVE(TEMP(_x), CONST(0)),"+
        "    			JUMP(NAME(.L0)),"+ 
        "    			LABEL(.L0),"+ 
        "       		MOVE(TEMP(_x), CONST(1)),"+
        "	    		LABEL(.L1),"+
        "				MOVE(TEMP(_x), CONST(2))"+ 
        "			)"+
        "		)"+
        " )";
		compareIRStrings(expectedIR,getIRString(source));
	}

	@Test
	public void testExpCall() {
		// Tests the EXP node
		String source = 
			"a():int = {\n" +
			"    return 1;"+
			"}"+
			"b() = {\n" +
			"    _ = a();"+
			"}";
		String expectedIR = "COMPUNIT("+
			"   FUNC _Ia_i("+
			"		SEQ(" +
			"		  MOVE(TEMP(RV), CONST(0))," +
			"         MOVE(TEMP(RV), CONST(1)),"+ 
			"         JUMP(NAME(.L0)), "+
			"         LABEL(.L0)"+
			"      )"+
			"   ), "+
			"   FUNC _Ib_p("+
			"      SEQ("+
			"         MOVE(TEMP(t0), CALL(NAME(_Ia_i)))"+ 
			"      )"+
			"   )"+
			")";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	
	@Test
	public void testShortCircuit() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
			"    x:int = 0;"+
			"    if (false&false&true) x=1;\n"+
			"    if (true|true|false) x=2;\n"+
			"	 while(false&true) x=3;\n"+
			"}";
		String expectedIR = "COMPUNIT("+
		"		FUNC _Ia_pi("+
		"         SEQ(" +
		"           MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
        "    		MOVE(TEMP(_x), CONST(0)),"+
        "          	JUMP(NAME(.L1)),"+ 
        "          	LABEL(.L3),"+ 
        "          	JUMP(NAME(.L1)),"+
        "       	LABEL(.L2),"+ 
        "       	JUMP(NAME(.L0)),"+
        "    		LABEL(.L0),"+ 
        "       	MOVE(TEMP(_x), CONST(1)),"+
        "    		LABEL(.L1),"+
        "          	JUMP(NAME(.L4)),"+ 
        "          	LABEL(.L7),"+ 
        "          	JUMP(NAME(.L4)),"+
        "       	LABEL(.L6),"+ 
        "       	JUMP(NAME(.L5)),"+
        "    		LABEL(.L4),"+ 
        "       	MOVE(TEMP(_x), CONST(2)),"+
        "    		LABEL(.L5),"+
        "   		LABEL(.L8),"+ 
        "       	JUMP(NAME(.L10)),"+ 
        "       	LABEL(.L11),"+ 
        "       	JUMP(NAME(.L9)),"+
        "    		LABEL(.L9),"+ 
        "       	MOVE(TEMP(_x), CONST(3)),"+
        "    		JUMP(NAME(.L8)),"+ 
        "    		LABEL(.L10)"+
        "		)"+
        "	)"+
        " )";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testWhile() {
		// Tests a procedure return
		String source = 
			"a(i:int) = {\n" +
			"    x:int = 1;"+
			"    while (true) x=1;\n"+
			"}";
		String expectedIR = "COMPUNIT("+
		"		FUNC _Ia_pi("+
		"			SEQ("+
		"    			MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8))))," +
        "    			MOVE(TEMP(_x), CONST(1)),"+
        "    			LABEL(.L0),"+ 
        "    			JUMP(NAME(.L1)),"+ 
        "    			LABEL(.L1),"+ 
        "       		MOVE(TEMP(_x), CONST(1)),"+
        "    			JUMP(NAME(.L0)),"+ 
        "    			LABEL(.L2)"+ 
        "			)"+
        "		)"+
        " )";
		compareIRStrings(expectedIR,getIRString(source));
	}
	
	@Test
	public void testUnderscore() {
		String source = 
			  "a(i:int, _, (j:int, _)):int = {"
			+ "	  return i+j;"
			+ "}"
			+ "b() = {"
			+ "	  _ = a(1, 2, (3,4));"
			+ "}";

		String expectedIR =
			  "COMPUNIT("
			+ "   FUNC _Ia_ii$t2i$("
			+ "		 SEQ(" 
			+ "         MOVE(TEMP(_i), MEM_I(ADD(TEMP(FP), CONST(8)))),"
			+ "         MOVE(TEMP(_j), MEM_I(ADD(TEMP(FP), CONST(12)))),"
			+ "			MOVE(TEMP(RV), CONST(0))," 
			+ "         MOVE(TEMP(RV), ADD(TEMP(_i), TEMP(_j)))," 
			+ "         JUMP(NAME(.L0))," 
			+ "         LABEL(.L0)"
			+ "      )"
			+ "   ),"
			+ "   FUNC _Ib_p("
			+ "	    SEQ("
			+ "        MOVE(TEMP(t4), ESEQ("
			+ "           SEQ("
			+ "              MOVE(TEMP(t2), CONST(1))," 
			+ "              MOVE(TEMP(t3), CONST(2)),"
			+ "              MOVE(TEMP(t0), CONST(3))," 
			+ "              MOVE(TEMP(t1), CONST(4))"
			+ "           )," 
			+ "           CALL(NAME(_Ia_ii$t2i$), TEMP(t2), TEMP(t0))"
			+ "       ))"
			+ "     )"
			+ "   )"
			+ ")";
		compareIRStrings(expectedIR,getIRString(source));
	}
	@Test
	public void testSimpleClass() {
		String source = 
		"class ColoredPoint {" +
		"	x, y: int;" +
		"   getX() : int = {" +
		"      return x;" +
		"   }" +
		"   getY() : int = {" +
		"      return y;" +
		"   }" +
		"   addXY() : int = {" +
		"      return (this.getX()+this.getY());" +
		"   }" +
		"}";
		String expectedIR = 
			"COMPUNIT("
			+"   FUNC _I_ColoredPoint_addXY_i("
			+"      SEQ("
			+"         MOVE(TEMP(RV), CONST(0)), "
			+"         MOVE(TEMP(THIS), MEM(ADD(TEMP(FP), CONST(8)))), "
			+"         MOVE(TEMP(OFFSET), CONST(0)), "
			+"         MOVE(TEMP(RV), ADD(ESEQ("
			+"                               SEQ,"
			+"                               CALL(MEM_I(ADD(MEM_I(TEMP(THIS)), CONST(4))), TEMP(THIS))"
			+"         ), ESEQ(SEQ,"
			+"               CALL(MEM_I(ADD(MEM_I(TEMP(THIS)), CONST(8))), TEMP(THIS))"
			+"         ))), "
			+"         JUMP(NAME(.L2)), "
			+"         LABEL(.L2)"
			+"      )"
			+"   ), "
			+"   FUNC _I_ColoredPoint_getX_i("
			+"      SEQ("
			+"         MOVE(TEMP(RV), CONST(0)), "
			+"         MOVE(TEMP(THIS), MEM(ADD(TEMP(FP), CONST(8)))), "
			+"         MOVE(TEMP(OFFSET), CONST(0)), "
			+"         MOVE(TEMP(RV), MEM(ADD(ADD(TEMP(THIS), TEMP(OFFSET)), CONST(4)))), "
			+"         JUMP(NAME(.L0)), "
			+"         LABEL(.L0)"
			+"      )"
			+"   ), "
			+"   FUNC _I_ColoredPoint_getY_i("
			+"      SEQ("
			+"         MOVE(TEMP(RV), CONST(0)), "
			+"         MOVE(TEMP(THIS), MEM(ADD(TEMP(FP), CONST(8)))), "
			+"         MOVE(TEMP(OFFSET), CONST(0)), "
			+"         MOVE(TEMP(RV), MEM(ADD(ADD(TEMP(THIS), TEMP(OFFSET)), CONST(8)))), "
			+"         JUMP(NAME(.L1)), "
			+"         LABEL(.L1)"
			+"      )"
			+"   ), "
			+"   FUNC _I_init_ColoredPoint("
			+"      SEQ"
			+"      "
			+"   ), "
			+"   SEQ("
			+"      DSEQ_RO(\"_I_size_ColoredPoint\", CONST(12)), "
			+"      DSEQ_RO(\"_I_vt_ColoredPoint\", CONST(0), NAME(_I_ColoredPoint_getX_i), "
			+"        NAME(_I_ColoredPoint_getY_i), NAME(_I_ColoredPoint_addXY_i))"
			+"   )"
			+")";
		compareIRStrings(expectedIR,getIRString(source));
	}

	@Test
	public void testTupleIndexing() {
		String source = 
		"a ( ) = { " +
		"	x:(int, int)=(1,2);" +
		"	y:int = x 2;" +
		"}";
		String expectedIR =
			"COMPUNIT(" +
			"	FUNC _Ia_p(" +
			"		SEQ( " +
			"			MOVE(TEMP(t0), CONST(1)), " +
			"			MOVE(TEMP(t1), CONST(2))," +
			"			MOVE(TEMP(_x$0), TEMP(t0)), " +
			"			MOVE(TEMP(_x$1), TEMP(t1)), " +
			"			MOVE(TEMP(_y), ESEQ(" +
			"						SEQ(" +
			"							MOVE(TEMP(t4), CONST(2)), " +
			"							CJUMP(EQ(TEMP(t4), CONST(0)), <none>, .L1)," +
			"							MOVE(TEMP(t3), TEMP(_x$0)), " +
			"							JUMP(NAME(.L0)), " +
			"							LABEL(.L1), " +
			"							CJUMP(EQ(TEMP(t4), CONST(1)), <none>, .L2), " +
			"							MOVE(TEMP(t3), TEMP(_x$1)), " +
			"							JUMP(NAME(.L0)), " +
			"							LABEL(.L2), " +
			"							JUMP(NAME(call_abort)), " +
			"							LABEL(.L0)" +
			"						), " +
			"						TEMP(t3)" +
			"		)))" +
			"	), " +
			"	SEQ(" +
			"		LABEL(call_abort), " +
			"		EXP(CALL(NAME(_I_outOfBounds_p)))" +
			"	)" +
			")";
		compareIRStrings(expectedIR,getIRString(source));
	}
}
