package xic;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import xic.Configuration.Architecture;
import xic.analysis.CFGGenerator;
import xic.analysis.CodeOrdering;
import xic.ast.CompUnit;
import xic.ast.ErrorReporter;
import xic.ir.IRCompUnit;
import xic.ir.IRFuncDecl;
import xic.ir.IRSimulator;
import xic.parser.IotaParser;
import xic.parser.ParserImpl.Language;
import xic.parser.TokenType;
import xic.visitors.IRConstantFolder;
import xic.visitors.IRGenerator;
import xic.visitors.IRLowering;
import xic.visitors.TypeChecker;
import edu.cornell.cs.cs4120.xi.CompilationException;

@RunWith(Parameterized.class)
public class IRSimTest {
	private static int WS;

	@Parameters
	public static List<Boolean[]> parameters() {
		return Arrays.asList(new Boolean[][] {
			{ false, false, false, false }, /* No constant folding, no lowering */
			{ false, true, false, false },  /* Constant folding, no lowering */
			{ false, true, true, false },   /* Constant folding and lowering */
			{ false, true, true, true },     /* Constant folding, lowering and code layout */
			
			/* Repeat the tests once more for x86_64 */
			{ true, false, false, false }, /* No constant folding, no lowering */
			{ true, true, false, false },  /* Constant folding, no lowering */
			{ true, true, true, false },   /* Constant folding and lowering */
			{ true, true, true, true }     /* Constant folding, lowering and code layout */
		});
	}

	private boolean lower, optimize, codeLayout;

	public IRSimTest(boolean x86_64, boolean optimize, boolean lower, boolean codeLayout) {
		Configuration.setArchitecture(x86_64 ? Architecture.X86_64 : Architecture.I386);
		WS = Configuration.WORD_SIZE;
		this.optimize = optimize;
		this.lower = lower;
		this.codeLayout = codeLayout;
	}

	public IRCompUnit compile(String input) {
		try {
			ErrorSink errorSink = new ErrorSink();
			IotaParser p = new IotaParser(input, TokenType.SOURCE_IDENTIFIER);
			CompUnit compUnit = p.doParse(errorSink);

			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}
	
			ArrayList<String> searchPath = new ArrayList<String>();
			searchPath.add(ExecutableTest.getIotaIncludeDirectory());
			TypeChecker typeChecker = new TypeChecker(Language.IOTA_09, errorSink, searchPath);

			CompUnit typeCheckedUnit = typeChecker.run(compUnit);
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error compiling string", null);
			}

			IRGenerator generator = new IRGenerator(typeChecker.getNameGenerator(), optimize, Language.IOTA_09);
			IRCompUnit irCompUnit = generator.translate(typeCheckedUnit, typeChecker.getImports());
			
			if (optimize) {
				IRConstantFolder constantFolder = new IRConstantFolder(errorSink);
				irCompUnit = (IRCompUnit) constantFolder.visit(irCompUnit);
			}

			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error during optimization", null);
			}

			if (lower) {
				IRLowering loweringVisitor = new IRLowering(typeChecker.getNameGenerator());
				irCompUnit = (IRCompUnit) loweringVisitor.visit(irCompUnit);
			}
			
			if (codeLayout) {
				Map<String,Boolean> isProcHash = new HashMap<String,Boolean>();
				for(IRFuncDecl func: irCompUnit.functions().values())
					isProcHash.put(func.name(), func.isProcedure());
				
				irCompUnit = CodeOrdering.orderCode(
						irCompUnit.name(), 
						CFGGenerator.createCFG(irCompUnit),
						irCompUnit.extraSequence(),
						isProcHash
						);
			}

			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(input, errorSink);
				throw new CompilationException("Error during lowering", null);
			}

			return irCompUnit;
		} catch (IOException e) {
			throw new CompilationException("I/O error", null);
		}
	}

	@Test
	public void testSimpleFunction() {
		// Runs a simple arithmetic expression in the simulator
		String source = 
			"a(i:int, j:int):int = {\n" +
				"return i + (2 * j);\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(7, sim.call("_Ia_iii", new long[] { 1, 3 }));
	}

	@Test
	public void testFactorial() {
		// Runs a recursive function in the simulator
		String source = 
			"factorial(n:int):int = {" +
			"  if (n == 0)" +
			"    return 1;" +
			"  else" +
			"    return n*factorial(n-1);" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(120, sim.call("_Ifactorial_ii", new long[] { 5 }));
	}

	@Test
	public void testReturnTuple() {
		// Tests tuple return values
		String source = 
			"test(i:int):(int, int, int) = {" +
			"    return (5,i,13);" +
			"}"+
			"testcall(i:int):(int, int, int) = {"+
			"	 return test i;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(WS*3);
		sim.call("_Itestcall_t3iiii", new long[] { ptr, 7 });
		assertEquals(5, sim.read(ptr+0));
		assertEquals(7, sim.read(ptr+WS));
		assertEquals(13, sim.read(ptr+2*WS));
	}

	@Test
	public void testReturnArray() {
		// Tests tuple->array conversion (in return stmt)
		String source = 
			"test(i:int):int[] = {" +
			"    return (5,i,13);" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Itest_aii", new long[] { 7 });
		assertEquals(3, sim.read(ptr-WS));
		assertEquals(5, sim.read(ptr+0));
		assertEquals(7, sim.read(ptr+WS));
		assertEquals(13, sim.read(ptr+2*WS));
	}
	
	@Test
	public void testReturnArrayFromInt() {
		// Tests tuple->array conversion (in return stmt)
		String source = 
			"test(i:int):int[] = {" +
			"    return i;" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Itest_aii", new long[] { 7 });
		assertEquals(1, sim.read(ptr-WS));
		assertEquals(7, sim.read(ptr+0));
	}

	@Test
	public void testReturnString() {
		// Tests string array allocation (in return stmt)
		String source = 
			"test():int[] = {" +
			"    return \"abc\";" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Itest_ai", new long[] { });
		assertEquals(3, sim.read(ptr-WS), 3);
		assertEquals('a', sim.read(ptr+0));
		assertEquals('b', sim.read(ptr+WS));
		assertEquals('c', sim.read(ptr+2*WS));
	}
	
	@Test
	public void testReturnFancy() {
		// More complicated mixture of the above, with a larger array element type
		String source = 
			"test(i:int, j:int) : ((int, int)[], int) = {" +
			"    return (((5,i), (i*2,3)), j);" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(WS*3);
		long ptr2 = sim.call("_Itest_t2at2iiiii", new long[] { ptr, 7, 13 });
		assertEquals(ptr2, ptr);
		assertEquals(13, sim.read(ptr+WS));
		long ptr3 = sim.read(ptr);
		assertEquals(2, sim.read(ptr3-WS));
		assertEquals(5, sim.read(ptr3+0));
		assertEquals(7, sim.read(ptr3+WS));
		assertEquals(14, sim.read(ptr3+2*WS));
		assertEquals(3, sim.read(ptr3+3*WS));
	}
	
	@Test
	public void testTupleComparison() {
		// Tuple comparisons
		String source = 
			"test(i:int) : bool = {" +
			"    return ((1,2),3) == ((1,i), 3);" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(1, sim.call("_Itest_bi", new long[] { 2 }));
		assertEquals(0, sim.call("_Itest_bi", new long[] { 3 }));
	}
	
	@Test
	public void testTupleComparison2() {
		// Tuple comparisons
		String source = 
			"test(i:int) : bool = {" +
			"    return ((1,2),3) != ((1,i), 3);" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(sim.call("_Itest_bi", new long[] { 2 }), 0);
		assertEquals(sim.call("_Itest_bi", new long[] { 3 }), 1);
	}

	@Test
	public void testTupleIndexing() {
		String source = 
			"a() : int = {\n" +
				"x:(int, int, int, int) = (1,2,3,4);\n"+
				"return x 2;\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));

		assertEquals(3, sim.call("_Ia_i", new long[] { }));
	}

	@Test
	public void testTupleTupleIndexing() {
		String source = 
			"a() : (int, int) = {\n" +
				"x:((int, int), (int, int)) = ((1,2),(3,4));\n"+
				"return x 1;\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(2*WS);
		assertEquals(ptr, sim.call("_Ia_t2ii", new long[] { ptr }));
		assertEquals(3, sim.read(ptr));
		assertEquals(4, sim.read(ptr+WS));
	}

	@Test
	public void testTupleTupleIndexing2() {
		String source = 
			"a() : (int, int) = {\n" +
				"return ((1,2),(3,4)) 1;\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(2*WS);
		assertEquals(ptr, sim.call("_Ia_t2ii", new long[] { ptr }));
		assertEquals(3, sim.read(ptr));
		assertEquals(4, sim.read(ptr+WS));
	}

	@Test
	public void testTupleArrayIndexing() {
		String source = 
			"a() : (int, int) = {\n" +
				"x:(int, int)[] = ((1,2), (3,4));\n"+
				"return x 1;\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(2*WS);

		sim.call("_Ia_t2ii", new long[] { ptr });
		assertEquals(3, sim.read(ptr+0));
		assertEquals(4, sim.read(ptr+WS));
	}

	@Test
	public void testArrayIndexing() {
		String source = 
			"a() : int = {\n" +
				"x:int[] = (1,2,3,4);\n"+
				"return x 2;\n"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));

		assertEquals(3, sim.call("_Ia_i", new long[] { }));
	}

	@Test
	public void testTupleValuedFunctionCall() {
		String source = 
		"a() : (int, int) = {\n" +
			"return (1,2);\n"+
		"}" +
		"b() : (int, int) = {\n" +
			"return a();\n"+
		"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(2*WS);

		sim.call("_Ib_t2ii", new long[] { ptr });
		assertEquals(1, sim.read(ptr+0));
		assertEquals(2, sim.read(ptr+WS));
	}
	
	@Test
	public void testGCD() {
		String source = 
			"gcd(a: int, b: int): int = {\n" 
			+ "  while (a != 0) {\n"
			+ "    if (a<b) b = b-a;\n"
			+ "    else a = a - b;\n" 
			+ "  }\n"
			+ "  return b;\n" 
			+ "}";
		IRSimulator sim = new IRSimulator(compile(source));

		assertEquals(8, sim.call("_Igcd_iii", new long[] { 24, 32 }));
	}
	
	@Test
	public void testMin() {
		String source = 
			"min(x:int, y:int):int = if (x<y) return x; else return y;";
		IRSimulator sim = new IRSimulator(compile(source));

		assertEquals(6, sim.call("_Imin_iii", new long[] { 6, 8 }), 6);
	}

	@Test
	public void testSort () {
		String source =
			"sort(a: int[]) : int[] = {"
			+	"  i:int = 0;"
			+	"  n:int = length a;"
			+	"  while (i < n) {"
			+	"    j:int = i;"
			+	"    while (j > 0) {"
			+	"      if (a(j-1) > a(j)) { // parens are optional on a(j)\n"
			+	"         swap:int = a(j);"
			+	"         a(j) = a(j-1);"
			+	"         a(j-1) = swap;"
			+	"	   }"
			+	"      j = j-1;"
			+	"    }"
			+	"    i = i+1;"
			+	"  }"
			+	"  return a;"
			+	"}"
			+"callSort():int[] = {"
			+	"  a:(int,int,int,int,int,int) = (8,6,9,2,13,6);"
			+	"  return sort a;"
			+	"}";
		
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_IcallSort_ai", new long[] {});
		assertEquals(6, sim.read(ptr-WS));
		assertEquals(2, sim.read(ptr));
		assertEquals(6, sim.read(ptr+WS));
		assertEquals(6, sim.read(ptr+2*WS));
		assertEquals(8, sim.read(ptr+3*WS));
		assertEquals(9, sim.read(ptr+4*WS));
		assertEquals(13, sim.read(ptr+5*WS));
	}

	@Test
	public void testLength() {
		// Tuples can be used in place of arrays.
		// (b:int) and b:int have the same type, therefore this is ok
		String source =
			"test():int[] = {\n"
			+ "  r:int[3];"
			+ "  b:int = 3;"
			+ "  r = (length b, length r, length 1);"
			+ "  return r;"
			+ "}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Itest_ai", new long[] {});
		assertEquals(3, sim.read(ptr-WS));
		assertEquals(1, sim.read(ptr));
		assertEquals(3, sim.read(ptr+WS));
		assertEquals(1, sim.read(ptr+2*WS));
	}

	@Test
	public void testSelfAssignment() {
		// Assignment statements of type: 
		// array(i) = array(i);
		String source =
				"fun(x:int,y:int):int[] = {"
				+"    r:int[];"
				+"    (x,y) = (y,x);"
				+"    r = (x,y);"
				+"    return r;"
				+"}";
		
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Ifun_aiii", new long[] {2,3});
		assertEquals(2, sim.read(ptr-WS));
		assertEquals(3, sim.read(ptr));
		assertEquals(2, sim.read(ptr+WS));
	}

	@Test
	public void testUnaryExprs() {
		// Assignment statements of type: 
		// array(i) = array(i);
		String source =
				"fun(x:int,y:int):int[] = {"
				+"    r:int[];"
				+"    if(!false) \n"
				+"    	r = (-x,-y);\n" 
				+"    else \n" 
				+"      r = (x, y);\n"
				+"    return r;"
				+"}";
		
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Ifun_aiii", new long[] {2,3});
		assertEquals(2, sim.read(ptr-WS));
		assertEquals(-2, sim.read(ptr));
		assertEquals(-3, sim.read(ptr+WS));
	}

	@Test
	public void testArrayofArrays() {
		/// Tests memory allocation code as well as the
		/// order reversal of nested ArrayTypes
		String source =
			"fun():int[] = {"
			+"    r:int[2][3];"
			+"    i:int = 0;"
			+"    while (i<2) {"
			+"      j:int = 0;"
			+"      while (j<3) {"
			+"        r i j=i*2+j;"
			+"        j = j+1;"
			+"      }"
			+"      i = i+1;"
			+"    }"
			+"    r 1 2 = 7;"
			+"    return r 1;"
			+"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Ifun_ai", new long[] {});
		assertEquals(3, sim.read(ptr-WS));
		assertEquals(2, sim.read(ptr));
		assertEquals(3, sim.read(ptr+WS));
		assertEquals(7, sim.read(ptr+2*WS));
	}
	
	@Test
	public void testNastyInitialization() {
		String source = 
		     "a(i:int, j:int) : (int, (int, int[])[]) = {"
			+"  b:(int, (int, int[j])[i]);"
			+"  return b;"
			+"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.malloc(2*WS);
		assertEquals(ptr, sim.call("_Ia_t2iat2iaiii", new long[] { ptr, 2, 4 }));
		assertEquals(3*WS,  sim.read(ptr+WS)); // index of first created array
		assertEquals(2,     sim.read(2*WS));   // size-2 outer array
		long ptr1= sim.read(3*WS+2*WS*0+WS);
		long ptr2= sim.read(3*WS+2*WS*1+WS);
		assertEquals(4,   sim.read(ptr1-WS));
		assertEquals(4,   sim.read(ptr2-WS));
	}

	@Test
	public void testMatrixMultiplication() {
		// Randomly generated 2x4 * 4x3 -> 2x3 matrix multiplication with memory allocation
		// Also tests nested array conversion
		String source =
			"mmul(A:int[][], B:int[][]):int[][] = {"
			+"    (m:int, n:int, l:int) = (length A, length B 0, length A 0);"
			+"    C:int[m][n];"
			+"    i:int = 0; "
			+"    while (i<m) {"
			+"       j:int = 0;"
			+"       while (j<n) {"
			+"          k:int = 0;"
			+"          while (k<l) {"
			+"              C i j = C i j + A i k * B k j;"
			+"              k=k+1;"
			+"          }"
			+"          j=j+1;"
			+"       }"
			+"       i=i+1;"
			+"    }"
			+"    return C;"
			+"}"
			+ ""
			+"test():int[][] = {"
			+"   B:int[4][3];"
			+"   B 0 0 = 6;"
			+"   B 0 1 = 3;"
			+"   B 0 2 = 10;"
			+"   B 1 0 = 9;"
			+"   B 1 1 = 3;"
			+"   B 1 2 = 10;"
			+"   B 2 0 = 2;"
			+"   B 2 1 = 10;"
			+"   B 2 2 = 6;"
			+"   B 3 0 = 6;"
			+"   B 3 1 = 4;"
			+"   B 3 2 = 0;"
			+"   return mmul("
			+"    ("
			+"      (4, 6, 3, 0),"
			+"      (1, 5, 5, 1)"
			+"    ),B);"
			+"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Itest_aai", new long[] { });
		assertEquals(sim.read(ptr-WS), 2);
		long row0 = sim.read(ptr);
		long row1 = sim.read(ptr+WS);
		assertEquals(3, sim.read(row0-WS));
		assertEquals(84, sim.read(row0+0));
		assertEquals(60, sim.read(row0+WS));
		assertEquals(118, sim.read(row0+2*WS));
		assertEquals(3, sim.read(row1-WS));
		assertEquals(67, sim.read(row1+0));
		assertEquals(72, sim.read(row1+WS));
		assertEquals(90, sim.read(row1+2*WS));
	}

	@Test
	public void testUnderscore() {
		String source = "a(i:int, _, (j:int, _)):int = {"+
		"	return i+j;"+
		"}"+
		"b() = {"+
		"	_ = a(1, 2, (3,4));"+
		"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(14, sim.call("_Ia_ii$t2i$", new long[] { 5, 9 }));
	}

	@Test
	public void testFloydWarshall() {
		String source = 
		     "use io;" +
		     "min(a:int, b:int):int = {" +
		     "	if(a<b)" +
		     "		return a;" +
		     "	return b;" +
		     "}" +
		     "" +
		     "FloydWarshall(path:int[][], n:int) = {" +
		     "	k:int = 0;" +
		     "	i:int = 0;" +
		     "	j:int = 0;" +
		     "	while(k<=n-1) {" +
		     "		while(i<=n-1) {" +
		     "			while(j<=n-1) {" +
		     "				path i j = min(path i j, (path i k) + (path k j));" +
		     "				j = j+1;" +
		     "			}" +
		     "			i = i+1;" +
		     "		}" +
		     "		k = k+1;" +
		     "	}" +
		     "}" +
		     "" +
		     "execute():int[][] = {" +
		     "	n:int = 5;" +
		     "	path:int[][] = ((0, 1, -3, 2, -4)," +
		     "		(3, 0, -4, 1, -1)," +
		     "		(7, 4, 0, 5, 3)," +
		     "		(2, -1, -5, 0, -2)," +
		     "		(8, 5, 1, 6, 0));" +
		     "	FloydWarshall(path, n);" +
		     "	return path;" +
		     "}" +
		     "" +
		     "main(path:int[][]) = {" +
		     "	" +
		     "}";
		
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Iexecute_aai", new long[] { });
		assertEquals(sim.read(ptr-WS), 5);
		long row0 = sim.read(ptr);
		long row1 = sim.read(ptr+WS);
		long row2 = sim.read(ptr+2*WS);
		long row3 = sim.read(ptr+3*WS);
		long row4 = sim.read(ptr+4*WS);
		
		assertEquals(5, sim.read(row0-WS));//Test
		assertEquals(5, sim.read(row1-WS));//length
		assertEquals(5, sim.read(row2-WS));//of
		assertEquals(5, sim.read(row2-WS));//rows
		assertEquals(5, sim.read(row2-WS));
		
		//first row is (0, 1, -3, 2, -4)
		assertEquals(0, sim.read(row0+0));
		assertEquals(1, sim.read(row0+WS));
		assertEquals(-3, sim.read(row0+2*WS));
		assertEquals(2, sim.read(row0+3*WS));
		assertEquals(-4, sim.read(row0+4*WS));
		
		//second row is (3, 0, -4, 1, -1)
		assertEquals(3, sim.read(row1+0));
		assertEquals(0, sim.read(row1+WS));
		assertEquals(-4, sim.read(row1+2*WS));
		assertEquals(1, sim.read(row1+3*WS));
		assertEquals(-1, sim.read(row1+4*WS));
		
		//third row is (7, 4, 0, 5, 3)
		assertEquals(7, sim.read(row2+0));
		assertEquals(4, sim.read(row2+WS));
		assertEquals(0, sim.read(row2+2*WS));
		assertEquals(5, sim.read(row2+3*WS));
		assertEquals(3, sim.read(row2+4*WS));
		
		//fourth row is (2, -1, -5, 0, -2)
		assertEquals(2, sim.read(row3+0));
		assertEquals(-1, sim.read(row3+WS));
		assertEquals(-5, sim.read(row3+2*WS));
		assertEquals(0, sim.read(row3+3*WS));
		assertEquals(-2, sim.read(row3+4*WS));
		
		//fifth row is (8, 5, 1, 6, 0)
		assertEquals(8, sim.read(row4+0));
		assertEquals(5, sim.read(row4+WS));
		assertEquals(1, sim.read(row4+2*WS));
		assertEquals(6, sim.read(row4+3*WS));
		assertEquals(0, sim.read(row4+4*WS));
	}

	@Test
	public void testBellmanFord() {
		String source = 
			"use io;" +
			"BellmanFord(predecessor:int[], weight:int[][], edges:(int, int)[],  source:int):int[] = {" +
			"	n:int = 5;" +
			"	m:int = 6;" +
			"	distance:int[n];" +
			"" +
			"" +
			"	j:int=0;" +
			"	while(j < n) {" +
			"		if(j == source)" +
			"			distance j = 0;" +
			"		else" +
			"			distance j = 10000;" +
			"		predecessor(j) = -1;" +
			"		j = j+1;" +
			"	}" +
			"" +
			"	i:int = 0;" +
			"	j = 0;" +
			"	while(i<= n-2) {" +
			"		while(j<m) {" +
			"			u:int = edges(j) 0;" +
			"			v:int = edges(j) 1;" +
			"			if(distance(u) + (weight u v) < distance(v)) {" +
			"				distance(v) = distance(u) + (weight u v);" +
			"				predecessor(v) = u;" +
			"			}" +
			"			j = j+1;" +
			"		}" +
			"		i = i+1;" +	
			"	}" +
			"	" +
			"" +
			"	i = 0;" +
			"	while(i < m) { " +
			"		u:int = edges(i) 0;" +
			"		v:int = edges(i) 1;" +
			"		if(distance(u) + (weight u v) < distance(v))" +
			"			print(\"Error: graph contains a negative-weight cycle\");" +
			"		i = i+1;" +
			"	}" +
			"	return distance;\n" +
			"}" +
			"" +
			"execute():int[] = {" +
			"	weight:int[][] = (" +
			"		(0, 7, 10000, 10000, 1)," +
			"		(7, 0, 1, 10000, 8)," +
			"		(10000, 1, 0, 2, 10000)," +
			"		(10000, 10000, 2, 0, 2)," +
			"		(1, 8, 10000, 2, 0)" +
			"	);" +
			"	predecessor:int[length weight];" +
			"	source:int = 0;" +
			"	edges:(int,int)[] = (" +
			"		(0, 1)," +
			"		(0, 4)," +
			"		(1, 2)," +
			"		(1, 4)," +
			"		(2, 3)," +
			"		(3, 4)" +
			"	);" +
			"	return BellmanFord(predecessor, weight, edges, source);" +
			"}" +
			"go():bool = {" +
			"	a:int[]=execute();\n   " +
			"	b:int[]=(0,7,8,10,1);\n " +
			"	i:int = 0;\n" +
			"	r:bool = true;" +
			"	blen:int = length b;\n" +
			"	c:int = 123;" +
			"	while(i < blen) {" +
			"		if(a i != b i) {" +
			"			return false;\n" +
			"		}  \n" +
			"		i=i+1;\n" +
			"	}" +
			"	return true;\n" +
			"}";
		
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(1,sim.call("_Igo_b", new long[] { }));
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTestArrayBoundsChecking() {
		String source = 
			"a() : int = {\n" +
				"x:int[] = (1,2,3,4);\n" +
				"y:int[] = (0,0,0,0);"+
				"i:int = 0;" +
				"while(true) {" +
				"	i=i+1;\n" +
				"	y i = x i;\n" +
				"}"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(3, sim.call("_Ia_i", new long[] { }));
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTestDivideByZero() {
		String source = 
			"a(i:int) : int = {\n" +
			"    return 5/i;" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(3, sim.call("_Ia_ii", new long[] { 0 }));
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTestInvalidAlloc() {
		String source = 
			"a(i:int) = {\n" +
			"    j:int[i];" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ia_pi", new long[] { -1 });
	}
	
	@Test(expected = IRSimulator.Trap.class)
	public void negativeTupleBoundCheck () {
		String source = 
			"a(i:(int, int, int)):int = {\n" +
			"    j:int = i 5;" +
			"    return j;" +
			"}"+
			"b():int = {\n" +
			"	r:int = a (1,2,3);\n"+
			"   return r;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test
	public void tupleArrayAssign () {
		String source = 
			"b():int = {\n" +
			"	r:(int,int)[] = ((1,2),(3,4));\n"+
			"   r 1 = (5,6);\n" +
			"   return r 1 0;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTupleArrayAssign () {
		String source = 
			"b():int = {\n" +
			"	r:(int,int)[] = ((1,2),(3,4));\n"+
			"   r 2 = (5,6);\n" +
			"   return r 1 0;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTupleArrayIndex() {
		String source = 
			"b():int = {\n" +
			"	r:(int,int)[] = ((1,2),(3,4));\n"+
			"   r 2 = (5,6);\n" +
			"   return r 1 2;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTupleArrayIndex2() {
		String source = 
			"b():int = {\n" +
			"	r:(int,int)[] = ((1,2),(3,4));\n"+
			"   return r 2 1;"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeArrayNegativeIndex() {
		String source = 
			"b():int = {\n" +
			"	r:int[] = (1,2,3);\n"+
			"   return r(-1);"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}

	@Test(expected = IRSimulator.Trap.class)
	public void negativeTupleArrayTupleIndex() {
		String source = 
			"b():int = {\n" +
			"	r:int[][] = ((1,2,3),(2,3));\n"+
			"   return r 1 (2);"+
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		sim.call("_Ib_i", new long[] { });
	}
	
	@Test
	public void testPrim() {
		// O(n^2) version of Prim's algorithm

		String source = 
			"updateDistances(adj:int[][], d:int[], treeSucc:int[], target:int) = {" +
			"    n:int = length(adj);" +
			"    i:int = 0;" +
			"    while (i<n) {" +
			"        if (adj target i > 0 &" +
			"            d i > adj target i) {" +
			"            d i = adj target i;" +
			"            treeSucc i = target;" +
			"        }" +
			"        i = i+1;" +
			"    }" +
			"}" +
			"" +
			"prim(adj:int[][]) : int = {" +
			"    (cost:int, n:int) = (0, length(adj));" +
			"    contained:bool[n];" +
			"    treeSucc:int[n];" +
			"" +
			"    d:int[n];" +
			"    large_val:int = 100000;" +
			"" +
			"    i:int = 0;" +
			"    while (i < n) {" +
			"        d(i) = large_val;" +
			"        contained(i) = false;" +
			"        i = i+1;" +
			"    }" +
			"" +
			"    contained(0) = true;" +
			"    updateDistances(adj, d, treeSucc, 0);" +
			"" +
			"    i = 0;" +
			"    while (i<n-1) {" +
			"        (j:int, min:int) = (0, -1);" +
			"        while (j<n) {" +
			"            if (!contained j)" +
			"                if (min == -1 | d(j) < d(min))" +
			"                    min = j;" +
			"            j = j+1;" +
			"        }" +
			"" +
			"        contained(min) = true;" +
			"        updateDistances(adj, d, treeSucc, min);" +
			"        cost = cost + d min;" +
			"" +
			"        i = i+1;" +
			"    }" +
			"" +
			"" +
			"    return cost;" +
			"}" +
			"" +
			"test():int = {" +
			"    adj:int[7][7];" +
			"    i:int = 0;" +
			"    while (i<7) {" +
			"        j:int = 0;" +
			"        while (j<7) {" +
			"            adj i j = 0;" +
			"            j = j+1;" +
			"        }" +
			"        i = i+1;" +
			"    }" +
			"    adj 0 1 = 7;  adj 1 0 = 7;" +
			"    adj 0 3 = 5;  adj 3 0 = 5;" +
			"    adj 1 2 = 8;  adj 2 1 = 8;" +
			"    adj 1 3 = 9;  adj 3 1 = 9;" +
			"    adj 1 4 = 7;  adj 4 1 = 7;" +
			"    adj 2 4 = 5;  adj 4 2 = 5;" +
			"    adj 3 4 = 15; adj 4 3 = 15;" +
			"    adj 3 5 = 6;  adj 5 3 = 6;" +
			"    adj 4 5 = 8;  adj 5 4 = 8;" +
			"    adj 4 6 = 9;  adj 6 4 = 9;" +
			"    adj 5 6 = 11; adj 6 5 = 11;" +
			"    return prim adj;" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(39, sim.call("_Itest_i", new long[] { }));
	}
	
	@Test
	public void testItoa() {
		String source = 
			"itoa(i:int) : int[] = {" +
			"    string:int[0];" +
			"" +
			"    while (true) {" +
			"        newString:int[length string + 1];" +
			"        j:int = 0;" +
			"        while (j<length string) {" +
			"            newString(j+1) = string j;" +
			"            j = j+1;" +
			"        }" +
			"        string = newString;" +
			"        string 0 = '0' + (i % 10);" +
			"        i = i / 10;" +
			"" +
			"        if (i == 0) " +
			"            break;" +
			"    }" + 
			"    return string;" + 
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Iitoa_aii", new long[] { 153 });
		assertEquals(sim.read(ptr-WS), 3);
		assertEquals(sim.read(ptr), '1');
		assertEquals(sim.read(ptr+WS), '5');
		assertEquals(sim.read(ptr+2*WS), '3');
	}

	@Test
	public void tupleShortCircuit() {
		String source = 
			"b(i:int[],j:int, k:int):int = {\n" +
			"	i j = k;" +
			"	return k;"+
			"}"+
			"a():int[] = {\n"+
			"	r:int[]=(0,1,2);" +
			"	if ((b (r, 0, 1), b (r, 1, 2), b (r, 2, 3)) == (1,1,1))" +
			"		return r;" +
			"	else return r;" +
			"}";
			
		
		IRSimulator sim = new IRSimulator(compile(source));
		long ptr = sim.call("_Ia_ai", new long[] { });
		assertEquals(sim.read(ptr-WS), 3);
		assertEquals(sim.read(ptr),   1);
		assertEquals(sim.read(ptr+WS), 2);
		assertEquals(sim.read(ptr+2*WS), 2);
	}
	@Test
	public void uninitializedTest() {
		String source = 
			"g():(int,int) = {\n" +
			"}"+
			"f():bool = {\n"+
			"	a:int;" +
			"	b:int;" +
			"	c:bool;" +
			"	(d:int, e:int) = g();" +
			"	x:int[];\n" +
			"	(y:bool[], (z:int[2][], za:(int[],int)[]));" +
			"	if (c) " +
			"		return false;" +
			"	else if(a+b != 0)" +
			"		return false;" +
			"	else if(d+e != 0) " +
			"		return false;" +
			"	else if(x != null | y !=null | z == null | z 0 != null | za != null ) " +
			"		return false;" +
			"	else " +
			"		return true;" +
			"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(1,sim.call("_If_b", new long[] { }));	
	}

	public void tupleAssignmentTest() {
		String source = 
		"f():bool = {\n"+
			"words:(int[],int[],int[]) = (\"Please\", \"Thanks\", \"You're Welcome\");"+
			"words(1) = \"Thank you\";"+
			"if (words 1 5 == 's') " +
			"	return false;" +
			"else " +
			"	return true;" +
		"}";
		IRSimulator sim = new IRSimulator(compile(source));
		assertEquals(1,sim.call("_If_b", new long[] { }));	
	}
}
